sqlglot.dialects.tsql API documentation

Edit on GitHub

# sqlglot.dialects.tsql

View Source

       1from __future__ import annotations
       2
       3import datetime
       4import re
       5import typing as t
       6from functools import partial
       7
       8from sqlglot import exp, generator, parser, tokens, transforms
       9from sqlglot.dialects.dialect import (
      10    Dialect,
      11    NormalizationStrategy,
      12    any_value_to_max_sql,
      13    date_delta_sql,
      14    datestrtodate_sql,
      15    generatedasidentitycolumnconstraint_sql,
      16    max_or_greatest,
      17    min_or_least,
      18    build_date_delta,
      19    rename_func,
      20    trim_sql,
      21)
      22from sqlglot.helper import seq_get
      23from sqlglot.time import format_time
      24from sqlglot.tokens import TokenType
      25
      26if t.TYPE_CHECKING:
      27    from sqlglot._typing import E
      28
      29FULL_FORMAT_TIME_MAPPING = {
      30    "weekday": "%A",
      31    "dw": "%A",
      32    "w": "%A",
      33    "month": "%B",
      34    "mm": "%B",
      35    "m": "%B",
      36}
      37
      38DATE_DELTA_INTERVAL = {
      39    "year": "year",
      40    "yyyy": "year",
      41    "yy": "year",
      42    "quarter": "quarter",
      43    "qq": "quarter",
      44    "q": "quarter",
      45    "month": "month",
      46    "mm": "month",
      47    "m": "month",
      48    "week": "week",
      49    "ww": "week",
      50    "wk": "week",
      51    "day": "day",
      52    "dd": "day",
      53    "d": "day",
      54}
      55
      56
      57DATE_FMT_RE = re.compile("([dD]{1,2})|([mM]{1,2})|([yY]{1,4})|([hH]{1,2})|([sS]{1,2})")
      58
      59# N = Numeric, C=Currency
      60TRANSPILE_SAFE_NUMBER_FMT = {"N", "C"}
      61
      62DEFAULT_START_DATE = datetime.date(1900, 1, 1)
      63
      64BIT_TYPES = {exp.EQ, exp.NEQ, exp.Is, exp.In, exp.Select, exp.Alias}
      65
      66# Unsupported options:
      67# - OPTIMIZE FOR ( @variable_name { UNKNOWN | = <literal_constant> } [ , ...n ] )
      68# - TABLE HINT
      69OPTIONS: parser.OPTIONS_TYPE = {
      70    **dict.fromkeys(
      71        (
      72            "DISABLE_OPTIMIZED_PLAN_FORCING",
      73            "FAST",
      74            "IGNORE_NONCLUSTERED_COLUMNSTORE_INDEX",
      75            "LABEL",
      76            "MAXDOP",
      77            "MAXRECURSION",
      78            "MAX_GRANT_PERCENT",
      79            "MIN_GRANT_PERCENT",
      80            "NO_PERFORMANCE_SPOOL",
      81            "QUERYTRACEON",
      82            "RECOMPILE",
      83        ),
      84        tuple(),
      85    ),
      86    "CONCAT": ("UNION",),
      87    "DISABLE": ("EXTERNALPUSHDOWN", "SCALEOUTEXECUTION"),
      88    "EXPAND": ("VIEWS",),
      89    "FORCE": ("EXTERNALPUSHDOWN", "ORDER", "SCALEOUTEXECUTION"),
      90    "HASH": ("GROUP", "JOIN", "UNION"),
      91    "KEEP": ("PLAN",),
      92    "KEEPFIXED": ("PLAN",),
      93    "LOOP": ("JOIN",),
      94    "MERGE": ("JOIN", "UNION"),
      95    "OPTIMIZE": (("FOR", "UNKNOWN"),),
      96    "ORDER": ("GROUP",),
      97    "PARAMETERIZATION": ("FORCED", "SIMPLE"),
      98    "ROBUST": ("PLAN",),
      99    "USE": ("PLAN",),
     100}
     101
     102OPTIONS_THAT_REQUIRE_EQUAL = ("MAX_GRANT_PERCENT", "MIN_GRANT_PERCENT", "LABEL")
     103
     104
     105def _build_formatted_time(
     106    exp_class: t.Type[E], full_format_mapping: t.Optional[bool] = None
     107) -> t.Callable[[t.List], E]:
     108    def _builder(args: t.List) -> E:
     109        assert len(args) == 2
     110
     111        return exp_class(
     112            this=exp.cast(args[1], exp.DataType.Type.DATETIME),
     113            format=exp.Literal.string(
     114                format_time(
     115                    args[0].name.lower(),
     116                    (
     117                        {**TSQL.TIME_MAPPING, **FULL_FORMAT_TIME_MAPPING}
     118                        if full_format_mapping
     119                        else TSQL.TIME_MAPPING
     120                    ),
     121                )
     122            ),
     123        )
     124
     125    return _builder
     126
     127
     128def _build_format(args: t.List) -> exp.NumberToStr | exp.TimeToStr:
     129    this = seq_get(args, 0)
     130    fmt = seq_get(args, 1)
     131    culture = seq_get(args, 2)
     132
     133    number_fmt = fmt and (fmt.name in TRANSPILE_SAFE_NUMBER_FMT or not DATE_FMT_RE.search(fmt.name))
     134
     135    if number_fmt:
     136        return exp.NumberToStr(this=this, format=fmt, culture=culture)
     137
     138    if fmt:
     139        fmt = exp.Literal.string(
     140            format_time(fmt.name, TSQL.FORMAT_TIME_MAPPING)
     141            if len(fmt.name) == 1
     142            else format_time(fmt.name, TSQL.TIME_MAPPING)
     143        )
     144
     145    return exp.TimeToStr(this=this, format=fmt, culture=culture)
     146
     147
     148def _build_eomonth(args: t.List) -> exp.LastDay:
     149    date = exp.TsOrDsToDate(this=seq_get(args, 0))
     150    month_lag = seq_get(args, 1)
     151
     152    if month_lag is None:
     153        this: exp.Expression = date
     154    else:
     155        unit = DATE_DELTA_INTERVAL.get("month")
     156        this = exp.DateAdd(this=date, expression=month_lag, unit=unit and exp.var(unit))
     157
     158    return exp.LastDay(this=this)
     159
     160
     161def _build_hashbytes(args: t.List) -> exp.Expression:
     162    kind, data = args
     163    kind = kind.name.upper() if kind.is_string else ""
     164
     165    if kind == "MD5":
     166        args.pop(0)
     167        return exp.MD5(this=data)
     168    if kind in ("SHA", "SHA1"):
     169        args.pop(0)
     170        return exp.SHA(this=data)
     171    if kind == "SHA2_256":
     172        return exp.SHA2(this=data, length=exp.Literal.number(256))
     173    if kind == "SHA2_512":
     174        return exp.SHA2(this=data, length=exp.Literal.number(512))
     175
     176    return exp.func("HASHBYTES", *args)
     177
     178
     179DATEPART_ONLY_FORMATS = {"DW", "HOUR", "QUARTER"}
     180
     181
     182def _format_sql(self: TSQL.Generator, expression: exp.NumberToStr | exp.TimeToStr) -> str:
     183    fmt = expression.args["format"]
     184
     185    if not isinstance(expression, exp.NumberToStr):
     186        if fmt.is_string:
     187            mapped_fmt = format_time(fmt.name, TSQL.INVERSE_TIME_MAPPING)
     188
     189            name = (mapped_fmt or "").upper()
     190            if name in DATEPART_ONLY_FORMATS:
     191                return self.func("DATEPART", name, expression.this)
     192
     193            fmt_sql = self.sql(exp.Literal.string(mapped_fmt))
     194        else:
     195            fmt_sql = self.format_time(expression) or self.sql(fmt)
     196    else:
     197        fmt_sql = self.sql(fmt)
     198
     199    return self.func("FORMAT", expression.this, fmt_sql, expression.args.get("culture"))
     200
     201
     202def _string_agg_sql(self: TSQL.Generator, expression: exp.GroupConcat) -> str:
     203    this = expression.this
     204    distinct = expression.find(exp.Distinct)
     205    if distinct:
     206        # exp.Distinct can appear below an exp.Order or an exp.GroupConcat expression
     207        self.unsupported("T-SQL STRING_AGG doesn't support DISTINCT.")
     208        this = distinct.pop().expressions[0]
     209
     210    order = ""
     211    if isinstance(expression.this, exp.Order):
     212        if expression.this.this:
     213            this = expression.this.this.pop()
     214        # Order has a leading space
     215        order = f" WITHIN GROUP ({self.sql(expression.this)[1:]})"
     216
     217    separator = expression.args.get("separator") or exp.Literal.string(",")
     218    return f"STRING_AGG({self.format_args(this, separator)}){order}"
     219
     220
     221def _build_date_delta(
     222    exp_class: t.Type[E], unit_mapping: t.Optional[t.Dict[str, str]] = None
     223) -> t.Callable[[t.List], E]:
     224    def _builder(args: t.List) -> E:
     225        unit = seq_get(args, 0)
     226        if unit and unit_mapping:
     227            unit = exp.var(unit_mapping.get(unit.name.lower(), unit.name))
     228
     229        start_date = seq_get(args, 1)
     230        if start_date and start_date.is_number:
     231            # Numeric types are valid DATETIME values
     232            if start_date.is_int:
     233                adds = DEFAULT_START_DATE + datetime.timedelta(days=int(start_date.this))
     234                start_date = exp.Literal.string(adds.strftime("%F"))
     235            else:
     236                # We currently don't handle float values, i.e. they're not converted to equivalent DATETIMEs.
     237                # This is not a problem when generating T-SQL code, it is when transpiling to other dialects.
     238                return exp_class(this=seq_get(args, 2), expression=start_date, unit=unit)
     239
     240        return exp_class(
     241            this=exp.TimeStrToTime(this=seq_get(args, 2)),
     242            expression=exp.TimeStrToTime(this=start_date),
     243            unit=unit,
     244        )
     245
     246    return _builder
     247
     248
     249def qualify_derived_table_outputs(expression: exp.Expression) -> exp.Expression:
     250    """Ensures all (unnamed) output columns are aliased for CTEs and Subqueries."""
     251    alias = expression.args.get("alias")
     252
     253    if (
     254        isinstance(expression, (exp.CTE, exp.Subquery))
     255        and isinstance(alias, exp.TableAlias)
     256        and not alias.columns
     257    ):
     258        from sqlglot.optimizer.qualify_columns import qualify_outputs
     259
     260        # We keep track of the unaliased column projection indexes instead of the expressions
     261        # themselves, because the latter are going to be replaced by new nodes when the aliases
     262        # are added and hence we won't be able to reach these newly added Alias parents
     263        query = expression.this
     264        unaliased_column_indexes = (
     265            i for i, c in enumerate(query.selects) if isinstance(c, exp.Column) and not c.alias
     266        )
     267
     268        qualify_outputs(query)
     269
     270        # Preserve the quoting information of columns for newly added Alias nodes
     271        query_selects = query.selects
     272        for select_index in unaliased_column_indexes:
     273            alias = query_selects[select_index]
     274            column = alias.this
     275            if isinstance(column.this, exp.Identifier):
     276                alias.args["alias"].set("quoted", column.this.quoted)
     277
     278    return expression
     279
     280
     281# https://learn.microsoft.com/en-us/sql/t-sql/functions/datetimefromparts-transact-sql?view=sql-server-ver16#syntax
     282def _build_datetimefromparts(args: t.List) -> exp.TimestampFromParts:
     283    return exp.TimestampFromParts(
     284        year=seq_get(args, 0),
     285        month=seq_get(args, 1),
     286        day=seq_get(args, 2),
     287        hour=seq_get(args, 3),
     288        min=seq_get(args, 4),
     289        sec=seq_get(args, 5),
     290        milli=seq_get(args, 6),
     291    )
     292
     293
     294# https://learn.microsoft.com/en-us/sql/t-sql/functions/timefromparts-transact-sql?view=sql-server-ver16#syntax
     295def _build_timefromparts(args: t.List) -> exp.TimeFromParts:
     296    return exp.TimeFromParts(
     297        hour=seq_get(args, 0),
     298        min=seq_get(args, 1),
     299        sec=seq_get(args, 2),
     300        fractions=seq_get(args, 3),
     301        precision=seq_get(args, 4),
     302    )
     303
     304
     305def _build_with_arg_as_text(
     306    klass: t.Type[exp.Expression],
     307) -> t.Callable[[t.List[exp.Expression]], exp.Expression]:
     308    def _parse(args: t.List[exp.Expression]) -> exp.Expression:
     309        this = seq_get(args, 0)
     310
     311        if this and not this.is_string:
     312            this = exp.cast(this, exp.DataType.Type.TEXT)
     313
     314        expression = seq_get(args, 1)
     315        kwargs = {"this": this}
     316
     317        if expression:
     318            kwargs["expression"] = expression
     319
     320        return klass(**kwargs)
     321
     322    return _parse
     323
     324
     325def _build_json_query(args: t.List, dialect: Dialect) -> exp.JSONExtract:
     326    if len(args) == 1:
     327        # The default value for path is '$'. As a result, if you don't provide a
     328        # value for path, JSON_QUERY returns the input expression.
     329        args.append(exp.Literal.string("$"))
     330
     331    return parser.build_extract_json_with_path(exp.JSONExtract)(args, dialect)
     332
     333
     334def _json_extract_sql(
     335    self: TSQL.Generator, expression: exp.JSONExtract | exp.JSONExtractScalar
     336) -> str:
     337    json_query = self.func("JSON_QUERY", expression.this, expression.expression)
     338    json_value = self.func("JSON_VALUE", expression.this, expression.expression)
     339    return self.func("ISNULL", json_query, json_value)
     340
     341
     342class TSQL(Dialect):
     343    NORMALIZATION_STRATEGY = NormalizationStrategy.CASE_INSENSITIVE
     344    TIME_FORMAT = "'yyyy-mm-dd hh:mm:ss'"
     345    SUPPORTS_SEMI_ANTI_JOIN = False
     346    LOG_BASE_FIRST = False
     347    TYPED_DIVISION = True
     348    CONCAT_COALESCE = True
     349
     350    TIME_MAPPING = {
     351        "year": "%Y",
     352        "dayofyear": "%j",
     353        "day": "%d",
     354        "dy": "%d",
     355        "y": "%Y",
     356        "week": "%W",
     357        "ww": "%W",
     358        "wk": "%W",
     359        "hour": "%h",
     360        "hh": "%I",
     361        "minute": "%M",
     362        "mi": "%M",
     363        "n": "%M",
     364        "second": "%S",
     365        "ss": "%S",
     366        "s": "%-S",
     367        "millisecond": "%f",
     368        "ms": "%f",
     369        "weekday": "%W",
     370        "dw": "%W",
     371        "month": "%m",
     372        "mm": "%M",
     373        "m": "%-M",
     374        "Y": "%Y",
     375        "YYYY": "%Y",
     376        "YY": "%y",
     377        "MMMM": "%B",
     378        "MMM": "%b",
     379        "MM": "%m",
     380        "M": "%-m",
     381        "dddd": "%A",
     382        "dd": "%d",
     383        "d": "%-d",
     384        "HH": "%H",
     385        "H": "%-H",
     386        "h": "%-I",
     387        "S": "%f",
     388        "yyyy": "%Y",
     389        "yy": "%y",
     390    }
     391
     392    CONVERT_FORMAT_MAPPING = {
     393        "0": "%b %d %Y %-I:%M%p",
     394        "1": "%m/%d/%y",
     395        "2": "%y.%m.%d",
     396        "3": "%d/%m/%y",
     397        "4": "%d.%m.%y",
     398        "5": "%d-%m-%y",
     399        "6": "%d %b %y",
     400        "7": "%b %d, %y",
     401        "8": "%H:%M:%S",
     402        "9": "%b %d %Y %-I:%M:%S:%f%p",
     403        "10": "mm-dd-yy",
     404        "11": "yy/mm/dd",
     405        "12": "yymmdd",
     406        "13": "%d %b %Y %H:%M:ss:%f",
     407        "14": "%H:%M:%S:%f",
     408        "20": "%Y-%m-%d %H:%M:%S",
     409        "21": "%Y-%m-%d %H:%M:%S.%f",
     410        "22": "%m/%d/%y %-I:%M:%S %p",
     411        "23": "%Y-%m-%d",
     412        "24": "%H:%M:%S",
     413        "25": "%Y-%m-%d %H:%M:%S.%f",
     414        "100": "%b %d %Y %-I:%M%p",
     415        "101": "%m/%d/%Y",
     416        "102": "%Y.%m.%d",
     417        "103": "%d/%m/%Y",
     418        "104": "%d.%m.%Y",
     419        "105": "%d-%m-%Y",
     420        "106": "%d %b %Y",
     421        "107": "%b %d, %Y",
     422        "108": "%H:%M:%S",
     423        "109": "%b %d %Y %-I:%M:%S:%f%p",
     424        "110": "%m-%d-%Y",
     425        "111": "%Y/%m/%d",
     426        "112": "%Y%m%d",
     427        "113": "%d %b %Y %H:%M:%S:%f",
     428        "114": "%H:%M:%S:%f",
     429        "120": "%Y-%m-%d %H:%M:%S",
     430        "121": "%Y-%m-%d %H:%M:%S.%f",
     431    }
     432
     433    FORMAT_TIME_MAPPING = {
     434        "y": "%B %Y",
     435        "d": "%m/%d/%Y",
     436        "H": "%-H",
     437        "h": "%-I",
     438        "s": "%Y-%m-%d %H:%M:%S",
     439        "D": "%A,%B,%Y",
     440        "f": "%A,%B,%Y %-I:%M %p",
     441        "F": "%A,%B,%Y %-I:%M:%S %p",
     442        "g": "%m/%d/%Y %-I:%M %p",
     443        "G": "%m/%d/%Y %-I:%M:%S %p",
     444        "M": "%B %-d",
     445        "m": "%B %-d",
     446        "O": "%Y-%m-%dT%H:%M:%S",
     447        "u": "%Y-%M-%D %H:%M:%S%z",
     448        "U": "%A, %B %D, %Y %H:%M:%S%z",
     449        "T": "%-I:%M:%S %p",
     450        "t": "%-I:%M",
     451        "Y": "%a %Y",
     452    }
     453
     454    class Tokenizer(tokens.Tokenizer):
     455        IDENTIFIERS = [("[", "]"), '"']
     456        QUOTES = ["'", '"']
     457        HEX_STRINGS = [("0x", ""), ("0X", "")]
     458        VAR_SINGLE_TOKENS = {"@", "$", "#"}
     459
     460        KEYWORDS = {
     461            **tokens.Tokenizer.KEYWORDS,
     462            "CLUSTERED INDEX": TokenType.INDEX,
     463            "DATETIME2": TokenType.DATETIME,
     464            "DATETIMEOFFSET": TokenType.TIMESTAMPTZ,
     465            "DECLARE": TokenType.DECLARE,
     466            "EXEC": TokenType.COMMAND,
     467            "FOR SYSTEM_TIME": TokenType.TIMESTAMP_SNAPSHOT,
     468            "IMAGE": TokenType.IMAGE,
     469            "MONEY": TokenType.MONEY,
     470            "NONCLUSTERED INDEX": TokenType.INDEX,
     471            "NTEXT": TokenType.TEXT,
     472            "OPTION": TokenType.OPTION,
     473            "OUTPUT": TokenType.RETURNING,
     474            "PRINT": TokenType.COMMAND,
     475            "PROC": TokenType.PROCEDURE,
     476            "REAL": TokenType.FLOAT,
     477            "ROWVERSION": TokenType.ROWVERSION,
     478            "SMALLDATETIME": TokenType.DATETIME,
     479            "SMALLMONEY": TokenType.SMALLMONEY,
     480            "SQL_VARIANT": TokenType.VARIANT,
     481            "SYSTEM_USER": TokenType.CURRENT_USER,
     482            "TOP": TokenType.TOP,
     483            "TIMESTAMP": TokenType.ROWVERSION,
     484            "TINYINT": TokenType.UTINYINT,
     485            "UNIQUEIDENTIFIER": TokenType.UNIQUEIDENTIFIER,
     486            "UPDATE STATISTICS": TokenType.COMMAND,
     487            "XML": TokenType.XML,
     488        }
     489        KEYWORDS.pop("/*+")
     490
     491        COMMANDS = {*tokens.Tokenizer.COMMANDS, TokenType.END}
     492
     493    class Parser(parser.Parser):
     494        SET_REQUIRES_ASSIGNMENT_DELIMITER = False
     495        LOG_DEFAULTS_TO_LN = True
     496        ALTER_TABLE_ADD_REQUIRED_FOR_EACH_COLUMN = False
     497        STRING_ALIASES = True
     498        NO_PAREN_IF_COMMANDS = False
     499
     500        QUERY_MODIFIER_PARSERS = {
     501            **parser.Parser.QUERY_MODIFIER_PARSERS,
     502            TokenType.OPTION: lambda self: ("options", self._parse_options()),
     503        }
     504
     505        FUNCTIONS = {
     506            **parser.Parser.FUNCTIONS,
     507            "CHARINDEX": lambda args: exp.StrPosition(
     508                this=seq_get(args, 1),
     509                substr=seq_get(args, 0),
     510                position=seq_get(args, 2),
     511            ),
     512            "DATEADD": build_date_delta(exp.DateAdd, unit_mapping=DATE_DELTA_INTERVAL),
     513            "DATEDIFF": _build_date_delta(exp.DateDiff, unit_mapping=DATE_DELTA_INTERVAL),
     514            "DATENAME": _build_formatted_time(exp.TimeToStr, full_format_mapping=True),
     515            "DATEPART": _build_formatted_time(exp.TimeToStr),
     516            "DATETIMEFROMPARTS": _build_datetimefromparts,
     517            "EOMONTH": _build_eomonth,
     518            "FORMAT": _build_format,
     519            "GETDATE": exp.CurrentTimestamp.from_arg_list,
     520            "HASHBYTES": _build_hashbytes,
     521            "ISNULL": exp.Coalesce.from_arg_list,
     522            "JSON_QUERY": _build_json_query,
     523            "JSON_VALUE": parser.build_extract_json_with_path(exp.JSONExtractScalar),
     524            "LEN": _build_with_arg_as_text(exp.Length),
     525            "LEFT": _build_with_arg_as_text(exp.Left),
     526            "RIGHT": _build_with_arg_as_text(exp.Right),
     527            "REPLICATE": exp.Repeat.from_arg_list,
     528            "SQUARE": lambda args: exp.Pow(this=seq_get(args, 0), expression=exp.Literal.number(2)),
     529            "SYSDATETIME": exp.CurrentTimestamp.from_arg_list,
     530            "SUSER_NAME": exp.CurrentUser.from_arg_list,
     531            "SUSER_SNAME": exp.CurrentUser.from_arg_list,
     532            "SYSTEM_USER": exp.CurrentUser.from_arg_list,
     533            "TIMEFROMPARTS": _build_timefromparts,
     534        }
     535
     536        JOIN_HINTS = {"LOOP", "HASH", "MERGE", "REMOTE"}
     537
     538        RETURNS_TABLE_TOKENS = parser.Parser.ID_VAR_TOKENS - {
     539            TokenType.TABLE,
     540            *parser.Parser.TYPE_TOKENS,
     541        }
     542
     543        STATEMENT_PARSERS = {
     544            **parser.Parser.STATEMENT_PARSERS,
     545            TokenType.DECLARE: lambda self: self._parse_declare(),
     546        }
     547
     548        RANGE_PARSERS = {
     549            **parser.Parser.RANGE_PARSERS,
     550            TokenType.DCOLON: lambda self, this: self.expression(
     551                exp.ScopeResolution,
     552                this=this,
     553                expression=self._parse_function() or self._parse_var(any_token=True),
     554            ),
     555        }
     556
     557        # The DCOLON (::) operator serves as a scope resolution (exp.ScopeResolution) operator in T-SQL
     558        COLUMN_OPERATORS = {
     559            **parser.Parser.COLUMN_OPERATORS,
     560            TokenType.DCOLON: lambda self, this, to: self.expression(exp.Cast, this=this, to=to)
     561            if isinstance(to, exp.DataType) and to.this != exp.DataType.Type.USERDEFINED
     562            else self.expression(exp.ScopeResolution, this=this, expression=to),
     563        }
     564
     565        def _parse_dcolon(self) -> t.Optional[exp.Expression]:
     566            # We want to use _parse_types() if the first token after :: is a known type,
     567            # otherwise we could parse something like x::varchar(max) into a function
     568            if self._match_set(self.TYPE_TOKENS, advance=False):
     569                return self._parse_types()
     570
     571            return self._parse_function() or self._parse_types()
     572
     573        def _parse_options(self) -> t.Optional[t.List[exp.Expression]]:
     574            if not self._match(TokenType.OPTION):
     575                return None
     576
     577            def _parse_option() -> t.Optional[exp.Expression]:
     578                option = self._parse_var_from_options(OPTIONS)
     579                if not option:
     580                    return None
     581
     582                self._match(TokenType.EQ)
     583                return self.expression(
     584                    exp.QueryOption, this=option, expression=self._parse_primary_or_var()
     585                )
     586
     587            return self._parse_wrapped_csv(_parse_option)
     588
     589        def _parse_projections(self) -> t.List[exp.Expression]:
     590            """
     591            T-SQL supports the syntax alias = expression in the SELECT's projection list,
     592            so we transform all parsed Selects to convert their EQ projections into Aliases.
     593
     594            See: https://learn.microsoft.com/en-us/sql/t-sql/queries/select-clause-transact-sql?view=sql-server-ver16#syntax
     595            """
     596            return [
     597                (
     598                    exp.alias_(projection.expression, projection.this.this, copy=False)
     599                    if isinstance(projection, exp.EQ) and isinstance(projection.this, exp.Column)
     600                    else projection
     601                )
     602                for projection in super()._parse_projections()
     603            ]
     604
     605        def _parse_commit_or_rollback(self) -> exp.Commit | exp.Rollback:
     606            """Applies to SQL Server and Azure SQL Database
     607            COMMIT [ { TRAN | TRANSACTION }
     608                [ transaction_name | @tran_name_variable ] ]
     609                [ WITH ( DELAYED_DURABILITY = { OFF | ON } ) ]
     610
     611            ROLLBACK { TRAN | TRANSACTION }
     612                [ transaction_name | @tran_name_variable
     613                | savepoint_name | @savepoint_variable ]
     614            """
     615            rollback = self._prev.token_type == TokenType.ROLLBACK
     616
     617            self._match_texts(("TRAN", "TRANSACTION"))
     618            this = self._parse_id_var()
     619
     620            if rollback:
     621                return self.expression(exp.Rollback, this=this)
     622
     623            durability = None
     624            if self._match_pair(TokenType.WITH, TokenType.L_PAREN):
     625                self._match_text_seq("DELAYED_DURABILITY")
     626                self._match(TokenType.EQ)
     627
     628                if self._match_text_seq("OFF"):
     629                    durability = False
     630                else:
     631                    self._match(TokenType.ON)
     632                    durability = True
     633
     634                self._match_r_paren()
     635
     636            return self.expression(exp.Commit, this=this, durability=durability)
     637
     638        def _parse_transaction(self) -> exp.Transaction | exp.Command:
     639            """Applies to SQL Server and Azure SQL Database
     640            BEGIN { TRAN | TRANSACTION }
     641            [ { transaction_name | @tran_name_variable }
     642            [ WITH MARK [ 'description' ] ]
     643            ]
     644            """
     645            if self._match_texts(("TRAN", "TRANSACTION")):
     646                transaction = self.expression(exp.Transaction, this=self._parse_id_var())
     647                if self._match_text_seq("WITH", "MARK"):
     648                    transaction.set("mark", self._parse_string())
     649
     650                return transaction
     651
     652            return self._parse_as_command(self._prev)
     653
     654        def _parse_returns(self) -> exp.ReturnsProperty:
     655            table = self._parse_id_var(any_token=False, tokens=self.RETURNS_TABLE_TOKENS)
     656            returns = super()._parse_returns()
     657            returns.set("table", table)
     658            return returns
     659
     660        def _parse_convert(
     661            self, strict: bool, safe: t.Optional[bool] = None
     662        ) -> t.Optional[exp.Expression]:
     663            this = self._parse_types()
     664            self._match(TokenType.COMMA)
     665            args = [this, *self._parse_csv(self._parse_assignment)]
     666            convert = exp.Convert.from_arg_list(args)
     667            convert.set("safe", safe)
     668            convert.set("strict", strict)
     669            return convert
     670
     671        def _parse_user_defined_function(
     672            self, kind: t.Optional[TokenType] = None
     673        ) -> t.Optional[exp.Expression]:
     674            this = super()._parse_user_defined_function(kind=kind)
     675
     676            if (
     677                kind == TokenType.FUNCTION
     678                or isinstance(this, exp.UserDefinedFunction)
     679                or self._match(TokenType.ALIAS, advance=False)
     680            ):
     681                return this
     682
     683            expressions = self._parse_csv(self._parse_function_parameter)
     684            return self.expression(exp.UserDefinedFunction, this=this, expressions=expressions)
     685
     686        def _parse_id_var(
     687            self,
     688            any_token: bool = True,
     689            tokens: t.Optional[t.Collection[TokenType]] = None,
     690        ) -> t.Optional[exp.Expression]:
     691            is_temporary = self._match(TokenType.HASH)
     692            is_global = is_temporary and self._match(TokenType.HASH)
     693
     694            this = super()._parse_id_var(any_token=any_token, tokens=tokens)
     695            if this:
     696                if is_global:
     697                    this.set("global", True)
     698                elif is_temporary:
     699                    this.set("temporary", True)
     700
     701            return this
     702
     703        def _parse_create(self) -> exp.Create | exp.Command:
     704            create = super()._parse_create()
     705
     706            if isinstance(create, exp.Create):
     707                table = create.this.this if isinstance(create.this, exp.Schema) else create.this
     708                if isinstance(table, exp.Table) and table.this.args.get("temporary"):
     709                    if not create.args.get("properties"):
     710                        create.set("properties", exp.Properties(expressions=[]))
     711
     712                    create.args["properties"].append("expressions", exp.TemporaryProperty())
     713
     714            return create
     715
     716        def _parse_if(self) -> t.Optional[exp.Expression]:
     717            index = self._index
     718
     719            if self._match_text_seq("OBJECT_ID"):
     720                self._parse_wrapped_csv(self._parse_string)
     721                if self._match_text_seq("IS", "NOT", "NULL") and self._match(TokenType.DROP):
     722                    return self._parse_drop(exists=True)
     723                self._retreat(index)
     724
     725            return super()._parse_if()
     726
     727        def _parse_unique(self) -> exp.UniqueColumnConstraint:
     728            if self._match_texts(("CLUSTERED", "NONCLUSTERED")):
     729                this = self.CONSTRAINT_PARSERS[self._prev.text.upper()](self)
     730            else:
     731                this = self._parse_schema(self._parse_id_var(any_token=False))
     732
     733            return self.expression(exp.UniqueColumnConstraint, this=this)
     734
     735        def _parse_partition(self) -> t.Optional[exp.Partition]:
     736            if not self._match_text_seq("WITH", "(", "PARTITIONS"):
     737                return None
     738
     739            def parse_range():
     740                low = self._parse_bitwise()
     741                high = self._parse_bitwise() if self._match_text_seq("TO") else None
     742
     743                return (
     744                    self.expression(exp.PartitionRange, this=low, expression=high) if high else low
     745                )
     746
     747            partition = self.expression(
     748                exp.Partition, expressions=self._parse_wrapped_csv(parse_range)
     749            )
     750
     751            self._match_r_paren()
     752
     753            return partition
     754
     755        def _parse_declare(self) -> exp.Declare | exp.Command:
     756            index = self._index
     757            expressions = self._try_parse(partial(self._parse_csv, self._parse_declareitem))
     758
     759            if not expressions or self._curr:
     760                self._retreat(index)
     761                return self._parse_as_command(self._prev)
     762
     763            return self.expression(exp.Declare, expressions=expressions)
     764
     765        def _parse_declareitem(self) -> t.Optional[exp.DeclareItem]:
     766            var = self._parse_id_var()
     767            if not var:
     768                return None
     769
     770            value = None
     771            self._match(TokenType.ALIAS)
     772            if self._match(TokenType.TABLE):
     773                data_type = self._parse_schema()
     774            else:
     775                data_type = self._parse_types()
     776                if self._match(TokenType.EQ):
     777                    value = self._parse_bitwise()
     778
     779            return self.expression(exp.DeclareItem, this=var, kind=data_type, default=value)
     780
     781    class Generator(generator.Generator):
     782        LIMIT_IS_TOP = True
     783        QUERY_HINTS = False
     784        RETURNING_END = False
     785        NVL2_SUPPORTED = False
     786        ALTER_TABLE_INCLUDE_COLUMN_KEYWORD = False
     787        LIMIT_FETCH = "FETCH"
     788        COMPUTED_COLUMN_WITH_TYPE = False
     789        CTE_RECURSIVE_KEYWORD_REQUIRED = False
     790        ENSURE_BOOLS = True
     791        NULL_ORDERING_SUPPORTED = None
     792        SUPPORTS_SINGLE_ARG_CONCAT = False
     793        TABLESAMPLE_SEED_KEYWORD = "REPEATABLE"
     794        SUPPORTS_SELECT_INTO = True
     795        JSON_PATH_BRACKETED_KEY_SUPPORTED = False
     796        SUPPORTS_TO_NUMBER = False
     797        SET_OP_MODIFIERS = False
     798        COPY_PARAMS_EQ_REQUIRED = True
     799        PARSE_JSON_NAME = None
     800
     801        EXPRESSIONS_WITHOUT_NESTED_CTES = {
     802            exp.Create,
     803            exp.Delete,
     804            exp.Insert,
     805            exp.Intersect,
     806            exp.Except,
     807            exp.Merge,
     808            exp.Select,
     809            exp.Subquery,
     810            exp.Union,
     811            exp.Update,
     812        }
     813
     814        SUPPORTED_JSON_PATH_PARTS = {
     815            exp.JSONPathKey,
     816            exp.JSONPathRoot,
     817            exp.JSONPathSubscript,
     818        }
     819
     820        TYPE_MAPPING = {
     821            **generator.Generator.TYPE_MAPPING,
     822            exp.DataType.Type.BOOLEAN: "BIT",
     823            exp.DataType.Type.DECIMAL: "NUMERIC",
     824            exp.DataType.Type.DATETIME: "DATETIME2",
     825            exp.DataType.Type.DOUBLE: "FLOAT",
     826            exp.DataType.Type.INT: "INTEGER",
     827            exp.DataType.Type.ROWVERSION: "ROWVERSION",
     828            exp.DataType.Type.TEXT: "VARCHAR(MAX)",
     829            exp.DataType.Type.TIMESTAMP: "DATETIME2",
     830            exp.DataType.Type.TIMESTAMPTZ: "DATETIMEOFFSET",
     831            exp.DataType.Type.UTINYINT: "TINYINT",
     832            exp.DataType.Type.VARIANT: "SQL_VARIANT",
     833        }
     834
     835        TYPE_MAPPING.pop(exp.DataType.Type.NCHAR)
     836        TYPE_MAPPING.pop(exp.DataType.Type.NVARCHAR)
     837
     838        TRANSFORMS = {
     839            **generator.Generator.TRANSFORMS,
     840            exp.AnyValue: any_value_to_max_sql,
     841            exp.ArrayToString: rename_func("STRING_AGG"),
     842            exp.AutoIncrementColumnConstraint: lambda *_: "IDENTITY",
     843            exp.DateAdd: date_delta_sql("DATEADD"),
     844            exp.DateDiff: date_delta_sql("DATEDIFF"),
     845            exp.CTE: transforms.preprocess([qualify_derived_table_outputs]),
     846            exp.CurrentDate: rename_func("GETDATE"),
     847            exp.CurrentTimestamp: rename_func("GETDATE"),
     848            exp.DateStrToDate: datestrtodate_sql,
     849            exp.Extract: rename_func("DATEPART"),
     850            exp.GeneratedAsIdentityColumnConstraint: generatedasidentitycolumnconstraint_sql,
     851            exp.GroupConcat: _string_agg_sql,
     852            exp.If: rename_func("IIF"),
     853            exp.JSONExtract: _json_extract_sql,
     854            exp.JSONExtractScalar: _json_extract_sql,
     855            exp.LastDay: lambda self, e: self.func("EOMONTH", e.this),
     856            exp.Max: max_or_greatest,
     857            exp.MD5: lambda self, e: self.func("HASHBYTES", exp.Literal.string("MD5"), e.this),
     858            exp.Min: min_or_least,
     859            exp.NumberToStr: _format_sql,
     860            exp.Repeat: rename_func("REPLICATE"),
     861            exp.Select: transforms.preprocess(
     862                [
     863                    transforms.eliminate_distinct_on,
     864                    transforms.eliminate_semi_and_anti_joins,
     865                    transforms.eliminate_qualify,
     866                ]
     867            ),
     868            exp.Stddev: rename_func("STDEV"),
     869            exp.StrPosition: lambda self, e: self.func(
     870                "CHARINDEX", e.args.get("substr"), e.this, e.args.get("position")
     871            ),
     872            exp.Subquery: transforms.preprocess([qualify_derived_table_outputs]),
     873            exp.SHA: lambda self, e: self.func("HASHBYTES", exp.Literal.string("SHA1"), e.this),
     874            exp.SHA2: lambda self, e: self.func(
     875                "HASHBYTES", exp.Literal.string(f"SHA2_{e.args.get('length', 256)}"), e.this
     876            ),
     877            exp.TemporaryProperty: lambda self, e: "",
     878            exp.TimeStrToTime: lambda self, e: self.sql(
     879                exp.cast(e.this, exp.DataType.Type.DATETIME)
     880            ),
     881            exp.TimeToStr: _format_sql,
     882            exp.Trim: trim_sql,
     883            exp.TsOrDsAdd: date_delta_sql("DATEADD", cast=True),
     884            exp.TsOrDsDiff: date_delta_sql("DATEDIFF"),
     885        }
     886
     887        TRANSFORMS.pop(exp.ReturnsProperty)
     888
     889        PROPERTIES_LOCATION = {
     890            **generator.Generator.PROPERTIES_LOCATION,
     891            exp.VolatileProperty: exp.Properties.Location.UNSUPPORTED,
     892        }
     893
     894        def scope_resolution(self, rhs: str, scope_name: str) -> str:
     895            return f"{scope_name}::{rhs}"
     896
     897        def select_sql(self, expression: exp.Select) -> str:
     898            if expression.args.get("offset"):
     899                if not expression.args.get("order"):
     900                    # ORDER BY is required in order to use OFFSET in a query, so we use
     901                    # a noop order by, since we don't really care about the order.
     902                    # See: https://www.microsoftpressstore.com/articles/article.aspx?p=2314819
     903                    expression.order_by(exp.select(exp.null()).subquery(), copy=False)
     904
     905                limit = expression.args.get("limit")
     906                if isinstance(limit, exp.Limit):
     907                    # TOP and OFFSET can't be combined, we need use FETCH instead of TOP
     908                    # we replace here because otherwise TOP would be generated in select_sql
     909                    limit.replace(exp.Fetch(direction="FIRST", count=limit.expression))
     910
     911            return super().select_sql(expression)
     912
     913        def convert_sql(self, expression: exp.Convert) -> str:
     914            name = "TRY_CONVERT" if expression.args.get("safe") else "CONVERT"
     915            return self.func(
     916                name, expression.this, expression.expression, expression.args.get("style")
     917            )
     918
     919        def queryoption_sql(self, expression: exp.QueryOption) -> str:
     920            option = self.sql(expression, "this")
     921            value = self.sql(expression, "expression")
     922            if value:
     923                optional_equal_sign = "= " if option in OPTIONS_THAT_REQUIRE_EQUAL else ""
     924                return f"{option} {optional_equal_sign}{value}"
     925            return option
     926
     927        def lateral_op(self, expression: exp.Lateral) -> str:
     928            cross_apply = expression.args.get("cross_apply")
     929            if cross_apply is True:
     930                return "CROSS APPLY"
     931            if cross_apply is False:
     932                return "OUTER APPLY"
     933
     934            # TODO: perhaps we can check if the parent is a Join and transpile it appropriately
     935            self.unsupported("LATERAL clause is not supported.")
     936            return "LATERAL"
     937
     938        def timefromparts_sql(self, expression: exp.TimeFromParts) -> str:
     939            nano = expression.args.get("nano")
     940            if nano is not None:
     941                nano.pop()
     942                self.unsupported("Specifying nanoseconds is not supported in TIMEFROMPARTS.")
     943
     944            if expression.args.get("fractions") is None:
     945                expression.set("fractions", exp.Literal.number(0))
     946            if expression.args.get("precision") is None:
     947                expression.set("precision", exp.Literal.number(0))
     948
     949            return rename_func("TIMEFROMPARTS")(self, expression)
     950
     951        def timestampfromparts_sql(self, expression: exp.TimestampFromParts) -> str:
     952            zone = expression.args.get("zone")
     953            if zone is not None:
     954                zone.pop()
     955                self.unsupported("Time zone is not supported in DATETIMEFROMPARTS.")
     956
     957            nano = expression.args.get("nano")
     958            if nano is not None:
     959                nano.pop()
     960                self.unsupported("Specifying nanoseconds is not supported in DATETIMEFROMPARTS.")
     961
     962            if expression.args.get("milli") is None:
     963                expression.set("milli", exp.Literal.number(0))
     964
     965            return rename_func("DATETIMEFROMPARTS")(self, expression)
     966
     967        def setitem_sql(self, expression: exp.SetItem) -> str:
     968            this = expression.this
     969            if isinstance(this, exp.EQ) and not isinstance(this.left, exp.Parameter):
     970                # T-SQL does not use '=' in SET command, except when the LHS is a variable.
     971                return f"{self.sql(this.left)} {self.sql(this.right)}"
     972
     973            return super().setitem_sql(expression)
     974
     975        def boolean_sql(self, expression: exp.Boolean) -> str:
     976            if type(expression.parent) in BIT_TYPES:
     977                return "1" if expression.this else "0"
     978
     979            return "(1 = 1)" if expression.this else "(1 = 0)"
     980
     981        def is_sql(self, expression: exp.Is) -> str:
     982            if isinstance(expression.expression, exp.Boolean):
     983                return self.binary(expression, "=")
     984            return self.binary(expression, "IS")
     985
     986        def createable_sql(self, expression: exp.Create, locations: t.DefaultDict) -> str:
     987            sql = self.sql(expression, "this")
     988            properties = expression.args.get("properties")
     989
     990            if sql[:1] != "#" and any(
     991                isinstance(prop, exp.TemporaryProperty)
     992                for prop in (properties.expressions if properties else [])
     993            ):
     994                sql = f"[#{sql[1:]}" if sql.startswith("[") else f"#{sql}"
     995
     996            return sql
     997
     998        def create_sql(self, expression: exp.Create) -> str:
     999            kind = expression.kind
    1000            exists = expression.args.pop("exists", None)
    1001
    1002            like_property = expression.find(exp.LikeProperty)
    1003            if like_property:
    1004                ctas_expression = like_property.this
    1005            else:
    1006                ctas_expression = expression.expression
    1007
    1008            if kind == "VIEW":
    1009                expression.this.set("catalog", None)
    1010                with_ = expression.args.get("with")
    1011                if ctas_expression and with_:
    1012                    # We've already preprocessed the Create expression to bubble up any nested CTEs,
    1013                    # but CREATE VIEW actually requires the WITH clause to come after it so we need
    1014                    # to amend the AST by moving the CTEs to the CREATE VIEW statement's query.
    1015                    ctas_expression.set("with", with_.pop())
    1016
    1017            sql = super().create_sql(expression)
    1018
    1019            table = expression.find(exp.Table)
    1020
    1021            # Convert CTAS statement to SELECT .. INTO ..
    1022            if kind == "TABLE" and ctas_expression:
    1023                if isinstance(ctas_expression, exp.UNWRAPPED_QUERIES):
    1024                    ctas_expression = ctas_expression.subquery()
    1025
    1026                select_into = exp.select("*").from_(exp.alias_(ctas_expression, "temp", table=True))
    1027                select_into.set("into", exp.Into(this=table))
    1028
    1029                if like_property:
    1030                    select_into.limit(0, copy=False)
    1031
    1032                sql = self.sql(select_into)
    1033
    1034            if exists:
    1035                identifier = self.sql(exp.Literal.string(exp.table_name(table) if table else ""))
    1036                sql = self.sql(exp.Literal.string(sql))
    1037                if kind == "SCHEMA":
    1038                    sql = f"""IF NOT EXISTS (SELECT * FROM information_schema.schemata WHERE schema_name = {identifier}) EXEC({sql})"""
    1039                elif kind == "TABLE":
    1040                    assert table
    1041                    where = exp.and_(
    1042                        exp.column("table_name").eq(table.name),
    1043                        exp.column("table_schema").eq(table.db) if table.db else None,
    1044                        exp.column("table_catalog").eq(table.catalog) if table.catalog else None,
    1045                    )
    1046                    sql = f"""IF NOT EXISTS (SELECT * FROM information_schema.tables WHERE {where}) EXEC({sql})"""
    1047                elif kind == "INDEX":
    1048                    index = self.sql(exp.Literal.string(expression.this.text("this")))
    1049                    sql = f"""IF NOT EXISTS (SELECT * FROM sys.indexes WHERE object_id = object_id({identifier}) AND name = {index}) EXEC({sql})"""
    1050            elif expression.args.get("replace"):
    1051                sql = sql.replace("CREATE OR REPLACE ", "CREATE OR ALTER ", 1)
    1052
    1053            return self.prepend_ctes(expression, sql)
    1054
    1055        def offset_sql(self, expression: exp.Offset) -> str:
    1056            return f"{super().offset_sql(expression)} ROWS"
    1057
    1058        def version_sql(self, expression: exp.Version) -> str:
    1059            name = "SYSTEM_TIME" if expression.name == "TIMESTAMP" else expression.name
    1060            this = f"FOR {name}"
    1061            expr = expression.expression
    1062            kind = expression.text("kind")
    1063            if kind in ("FROM", "BETWEEN"):
    1064                args = expr.expressions
    1065                sep = "TO" if kind == "FROM" else "AND"
    1066                expr_sql = f"{self.sql(seq_get(args, 0))} {sep} {self.sql(seq_get(args, 1))}"
    1067            else:
    1068                expr_sql = self.sql(expr)
    1069
    1070            expr_sql = f" {expr_sql}" if expr_sql else ""
    1071            return f"{this} {kind}{expr_sql}"
    1072
    1073        def returnsproperty_sql(self, expression: exp.ReturnsProperty) -> str:
    1074            table = expression.args.get("table")
    1075            table = f"{table} " if table else ""
    1076            return f"RETURNS {table}{self.sql(expression, 'this')}"
    1077
    1078        def returning_sql(self, expression: exp.Returning) -> str:
    1079            into = self.sql(expression, "into")
    1080            into = self.seg(f"INTO {into}") if into else ""
    1081            return f"{self.seg('OUTPUT')} {self.expressions(expression, flat=True)}{into}"
    1082
    1083        def transaction_sql(self, expression: exp.Transaction) -> str:
    1084            this = self.sql(expression, "this")
    1085            this = f" {this}" if this else ""
    1086            mark = self.sql(expression, "mark")
    1087            mark = f" WITH MARK {mark}" if mark else ""
    1088            return f"BEGIN TRANSACTION{this}{mark}"
    1089
    1090        def commit_sql(self, expression: exp.Commit) -> str:
    1091            this = self.sql(expression, "this")
    1092            this = f" {this}" if this else ""
    1093            durability = expression.args.get("durability")
    1094            durability = (
    1095                f" WITH (DELAYED_DURABILITY = {'ON' if durability else 'OFF'})"
    1096                if durability is not None
    1097                else ""
    1098            )
    1099            return f"COMMIT TRANSACTION{this}{durability}"
    1100
    1101        def rollback_sql(self, expression: exp.Rollback) -> str:
    1102            this = self.sql(expression, "this")
    1103            this = f" {this}" if this else ""
    1104            return f"ROLLBACK TRANSACTION{this}"
    1105
    1106        def identifier_sql(self, expression: exp.Identifier) -> str:
    1107            identifier = super().identifier_sql(expression)
    1108
    1109            if expression.args.get("global"):
    1110                identifier = f"##{identifier}"
    1111            elif expression.args.get("temporary"):
    1112                identifier = f"#{identifier}"
    1113
    1114            return identifier
    1115
    1116        def constraint_sql(self, expression: exp.Constraint) -> str:
    1117            this = self.sql(expression, "this")
    1118            expressions = self.expressions(expression, flat=True, sep=" ")
    1119            return f"CONSTRAINT {this} {expressions}"
    1120
    1121        def length_sql(self, expression: exp.Length) -> str:
    1122            return self._uncast_text(expression, "LEN")
    1123
    1124        def right_sql(self, expression: exp.Right) -> str:
    1125            return self._uncast_text(expression, "RIGHT")
    1126
    1127        def left_sql(self, expression: exp.Left) -> str:
    1128            return self._uncast_text(expression, "LEFT")
    1129
    1130        def _uncast_text(self, expression: exp.Expression, name: str) -> str:
    1131            this = expression.this
    1132            if isinstance(this, exp.Cast) and this.is_type(exp.DataType.Type.TEXT):
    1133                this_sql = self.sql(this, "this")
    1134            else:
    1135                this_sql = self.sql(this)
    1136            expression_sql = self.sql(expression, "expression")
    1137            return self.func(name, this_sql, expression_sql if expression_sql else None)
    1138
    1139        def partition_sql(self, expression: exp.Partition) -> str:
    1140            return f"WITH (PARTITIONS({self.expressions(expression, flat=True)}))"
    1141
    1142        def altertable_sql(self, expression: exp.AlterTable) -> str:
    1143            action = seq_get(expression.args.get("actions") or [], 0)
    1144            if isinstance(action, exp.RenameTable):
    1145                return f"EXEC sp_rename '{self.sql(expression.this)}', '{action.this.name}'"
    1146            return super().altertable_sql(expression)
    1147
    1148        def drop_sql(self, expression: exp.Drop) -> str:
    1149            if expression.args["kind"] == "VIEW":
    1150                expression.this.set("catalog", None)
    1151            return super().drop_sql(expression)
    1152
    1153        def declare_sql(self, expression: exp.Declare) -> str:
    1154            return f"DECLARE {self.expressions(expression, flat=True)}"
    1155
    1156        def declareitem_sql(self, expression: exp.DeclareItem) -> str:
    1157            variable = self.sql(expression, "this")
    1158            default = self.sql(expression, "default")
    1159            default = f" = {default}" if default else ""
    1160
    1161            kind = self.sql(expression, "kind")
    1162            if isinstance(expression.args.get("kind"), exp.Schema):
    1163                kind = f"TABLE {kind}"
    1164
    1165            return f"{variable} AS {kind}{default}"
    1166
    1167        def options_modifier(self, expression: exp.Expression) -> str:
    1168            options = self.expressions(expression, key="options")
    1169            return f" OPTION{self.wrap(options)}" if options else ""

FULL_FORMAT_TIME_MAPPING = {'weekday': '%A', 'dw': '%A', 'w': '%A', 'month': '%B', 'mm': '%B', 'm': '%B'}
DATE_DELTA_INTERVAL = {'year': 'year', 'yyyy': 'year', 'yy': 'year', 'quarter': 'quarter', 'qq': 'quarter', 'q': 'quarter', 'month': 'month', 'mm': 'month', 'm': 'month', 'week': 'week', 'ww': 'week', 'wk': 'week', 'day': 'day', 'dd': 'day', 'd': 'day'}
DATE_FMT_RE = re.compile('([dD]{1,2})|([mM]{1,2})|([yY]{1,4})|([hH]{1,2})|([sS]{1,2})')
TRANSPILE_SAFE_NUMBER_FMT = {'C', 'N'}
DEFAULT_START_DATE = datetime.date(1900, 1, 1)
BIT_TYPES = {<class 'sqlglot.expressions.EQ'>, <class 'sqlglot.expressions.NEQ'>, <class 'sqlglot.expressions.Select'>, <class 'sqlglot.expressions.In'>, <class 'sqlglot.expressions.Alias'>, <class 'sqlglot.expressions.Is'>}
OPTIONS: Dict[str, Sequence[Union[Sequence[str], str]]] = {'DISABLE_OPTIMIZED_PLAN_FORCING': (), 'FAST': (), 'IGNORE_NONCLUSTERED_COLUMNSTORE_INDEX': (), 'LABEL': (), 'MAXDOP': (), 'MAXRECURSION': (), 'MAX_GRANT_PERCENT': (), 'MIN_GRANT_PERCENT': (), 'NO_PERFORMANCE_SPOOL': (), 'QUERYTRACEON': (), 'RECOMPILE': (), 'CONCAT': ('UNION',), 'DISABLE': ('EXTERNALPUSHDOWN', 'SCALEOUTEXECUTION'), 'EXPAND': ('VIEWS',), 'FORCE': ('EXTERNALPUSHDOWN', 'ORDER', 'SCALEOUTEXECUTION'), 'HASH': ('GROUP', 'JOIN', 'UNION'), 'KEEP': ('PLAN',), 'KEEPFIXED': ('PLAN',), 'LOOP': ('JOIN',), 'MERGE': ('JOIN', 'UNION'), 'OPTIMIZE': (('FOR', 'UNKNOWN'),), 'ORDER': ('GROUP',), 'PARAMETERIZATION': ('FORCED', 'SIMPLE'), 'ROBUST': ('PLAN',), 'USE': ('PLAN',)}
OPTIONS_THAT_REQUIRE_EQUAL = ('MAX_GRANT_PERCENT', 'MIN_GRANT_PERCENT', 'LABEL')
DATEPART_ONLY_FORMATS = {'DW', 'HOUR', 'QUARTER'}
def qualify_derived_table_outputs( expression: sqlglot.expressions.Expression) -> sqlglot.expressions.Expression: View Source

    250def qualify_derived_table_outputs(expression: exp.Expression) -> exp.Expression:
    251    """Ensures all (unnamed) output columns are aliased for CTEs and Subqueries."""
    252    alias = expression.args.get("alias")
    253
    254    if (
    255        isinstance(expression, (exp.CTE, exp.Subquery))
    256        and isinstance(alias, exp.TableAlias)
    257        and not alias.columns
    258    ):
    259        from sqlglot.optimizer.qualify_columns import qualify_outputs
    260
    261        # We keep track of the unaliased column projection indexes instead of the expressions
    262        # themselves, because the latter are going to be replaced by new nodes when the aliases
    263        # are added and hence we won't be able to reach these newly added Alias parents
    264        query = expression.this
    265        unaliased_column_indexes = (
    266            i for i, c in enumerate(query.selects) if isinstance(c, exp.Column) and not c.alias
    267        )
    268
    269        qualify_outputs(query)
    270
    271        # Preserve the quoting information of columns for newly added Alias nodes
    272        query_selects = query.selects
    273        for select_index in unaliased_column_indexes:
    274            alias = query_selects[select_index]
    275            column = alias.this
    276            if isinstance(column.this, exp.Identifier):
    277                alias.args["alias"].set("quoted", column.this.quoted)
    278
    279    return expression

Ensures all (unnamed) output columns are aliased for CTEs and Subqueries.

class TSQL(sqlglot.dialects.dialect.Dialect): View Source

     343class TSQL(Dialect):
     344    NORMALIZATION_STRATEGY = NormalizationStrategy.CASE_INSENSITIVE
     345    TIME_FORMAT = "'yyyy-mm-dd hh:mm:ss'"
     346    SUPPORTS_SEMI_ANTI_JOIN = False
     347    LOG_BASE_FIRST = False
     348    TYPED_DIVISION = True
     349    CONCAT_COALESCE = True
     350
     351    TIME_MAPPING = {
     352        "year": "%Y",
     353        "dayofyear": "%j",
     354        "day": "%d",
     355        "dy": "%d",
     356        "y": "%Y",
     357        "week": "%W",
     358        "ww": "%W",
     359        "wk": "%W",
     360        "hour": "%h",
     361        "hh": "%I",
     362        "minute": "%M",
     363        "mi": "%M",
     364        "n": "%M",
     365        "second": "%S",
     366        "ss": "%S",
     367        "s": "%-S",
     368        "millisecond": "%f",
     369        "ms": "%f",
     370        "weekday": "%W",
     371        "dw": "%W",
     372        "month": "%m",
     373        "mm": "%M",
     374        "m": "%-M",
     375        "Y": "%Y",
     376        "YYYY": "%Y",
     377        "YY": "%y",
     378        "MMMM": "%B",
     379        "MMM": "%b",
     380        "MM": "%m",
     381        "M": "%-m",
     382        "dddd": "%A",
     383        "dd": "%d",
     384        "d": "%-d",
     385        "HH": "%H",
     386        "H": "%-H",
     387        "h": "%-I",
     388        "S": "%f",
     389        "yyyy": "%Y",
     390        "yy": "%y",
     391    }
     392
     393    CONVERT_FORMAT_MAPPING = {
     394        "0": "%b %d %Y %-I:%M%p",
     395        "1": "%m/%d/%y",
     396        "2": "%y.%m.%d",
     397        "3": "%d/%m/%y",
     398        "4": "%d.%m.%y",
     399        "5": "%d-%m-%y",
     400        "6": "%d %b %y",
     401        "7": "%b %d, %y",
     402        "8": "%H:%M:%S",
     403        "9": "%b %d %Y %-I:%M:%S:%f%p",
     404        "10": "mm-dd-yy",
     405        "11": "yy/mm/dd",
     406        "12": "yymmdd",
     407        "13": "%d %b %Y %H:%M:ss:%f",
     408        "14": "%H:%M:%S:%f",
     409        "20": "%Y-%m-%d %H:%M:%S",
     410        "21": "%Y-%m-%d %H:%M:%S.%f",
     411        "22": "%m/%d/%y %-I:%M:%S %p",
     412        "23": "%Y-%m-%d",
     413        "24": "%H:%M:%S",
     414        "25": "%Y-%m-%d %H:%M:%S.%f",
     415        "100": "%b %d %Y %-I:%M%p",
     416        "101": "%m/%d/%Y",
     417        "102": "%Y.%m.%d",
     418        "103": "%d/%m/%Y",
     419        "104": "%d.%m.%Y",
     420        "105": "%d-%m-%Y",
     421        "106": "%d %b %Y",
     422        "107": "%b %d, %Y",
     423        "108": "%H:%M:%S",
     424        "109": "%b %d %Y %-I:%M:%S:%f%p",
     425        "110": "%m-%d-%Y",
     426        "111": "%Y/%m/%d",
     427        "112": "%Y%m%d",
     428        "113": "%d %b %Y %H:%M:%S:%f",
     429        "114": "%H:%M:%S:%f",
     430        "120": "%Y-%m-%d %H:%M:%S",
     431        "121": "%Y-%m-%d %H:%M:%S.%f",
     432    }
     433
     434    FORMAT_TIME_MAPPING = {
     435        "y": "%B %Y",
     436        "d": "%m/%d/%Y",
     437        "H": "%-H",
     438        "h": "%-I",
     439        "s": "%Y-%m-%d %H:%M:%S",
     440        "D": "%A,%B,%Y",
     441        "f": "%A,%B,%Y %-I:%M %p",
     442        "F": "%A,%B,%Y %-I:%M:%S %p",
     443        "g": "%m/%d/%Y %-I:%M %p",
     444        "G": "%m/%d/%Y %-I:%M:%S %p",
     445        "M": "%B %-d",
     446        "m": "%B %-d",
     447        "O": "%Y-%m-%dT%H:%M:%S",
     448        "u": "%Y-%M-%D %H:%M:%S%z",
     449        "U": "%A, %B %D, %Y %H:%M:%S%z",
     450        "T": "%-I:%M:%S %p",
     451        "t": "%-I:%M",
     452        "Y": "%a %Y",
     453    }
     454
     455    class Tokenizer(tokens.Tokenizer):
     456        IDENTIFIERS = [("[", "]"), '"']
     457        QUOTES = ["'", '"']
     458        HEX_STRINGS = [("0x", ""), ("0X", "")]
     459        VAR_SINGLE_TOKENS = {"@", "$", "#"}
     460
     461        KEYWORDS = {
     462            **tokens.Tokenizer.KEYWORDS,
     463            "CLUSTERED INDEX": TokenType.INDEX,
     464            "DATETIME2": TokenType.DATETIME,
     465            "DATETIMEOFFSET": TokenType.TIMESTAMPTZ,
     466            "DECLARE": TokenType.DECLARE,
     467            "EXEC": TokenType.COMMAND,
     468            "FOR SYSTEM_TIME": TokenType.TIMESTAMP_SNAPSHOT,
     469            "IMAGE": TokenType.IMAGE,
     470            "MONEY": TokenType.MONEY,
     471            "NONCLUSTERED INDEX": TokenType.INDEX,
     472            "NTEXT": TokenType.TEXT,
     473            "OPTION": TokenType.OPTION,
     474            "OUTPUT": TokenType.RETURNING,
     475            "PRINT": TokenType.COMMAND,
     476            "PROC": TokenType.PROCEDURE,
     477            "REAL": TokenType.FLOAT,
     478            "ROWVERSION": TokenType.ROWVERSION,
     479            "SMALLDATETIME": TokenType.DATETIME,
     480            "SMALLMONEY": TokenType.SMALLMONEY,
     481            "SQL_VARIANT": TokenType.VARIANT,
     482            "SYSTEM_USER": TokenType.CURRENT_USER,
     483            "TOP": TokenType.TOP,
     484            "TIMESTAMP": TokenType.ROWVERSION,
     485            "TINYINT": TokenType.UTINYINT,
     486            "UNIQUEIDENTIFIER": TokenType.UNIQUEIDENTIFIER,
     487            "UPDATE STATISTICS": TokenType.COMMAND,
     488            "XML": TokenType.XML,
     489        }
     490        KEYWORDS.pop("/*+")
     491
     492        COMMANDS = {*tokens.Tokenizer.COMMANDS, TokenType.END}
     493
     494    class Parser(parser.Parser):
     495        SET_REQUIRES_ASSIGNMENT_DELIMITER = False
     496        LOG_DEFAULTS_TO_LN = True
     497        ALTER_TABLE_ADD_REQUIRED_FOR_EACH_COLUMN = False
     498        STRING_ALIASES = True
     499        NO_PAREN_IF_COMMANDS = False
     500
     501        QUERY_MODIFIER_PARSERS = {
     502            **parser.Parser.QUERY_MODIFIER_PARSERS,
     503            TokenType.OPTION: lambda self: ("options", self._parse_options()),
     504        }
     505
     506        FUNCTIONS = {
     507            **parser.Parser.FUNCTIONS,
     508            "CHARINDEX": lambda args: exp.StrPosition(
     509                this=seq_get(args, 1),
     510                substr=seq_get(args, 0),
     511                position=seq_get(args, 2),
     512            ),
     513            "DATEADD": build_date_delta(exp.DateAdd, unit_mapping=DATE_DELTA_INTERVAL),
     514            "DATEDIFF": _build_date_delta(exp.DateDiff, unit_mapping=DATE_DELTA_INTERVAL),
     515            "DATENAME": _build_formatted_time(exp.TimeToStr, full_format_mapping=True),
     516            "DATEPART": _build_formatted_time(exp.TimeToStr),
     517            "DATETIMEFROMPARTS": _build_datetimefromparts,
     518            "EOMONTH": _build_eomonth,
     519            "FORMAT": _build_format,
     520            "GETDATE": exp.CurrentTimestamp.from_arg_list,
     521            "HASHBYTES": _build_hashbytes,
     522            "ISNULL": exp.Coalesce.from_arg_list,
     523            "JSON_QUERY": _build_json_query,
     524            "JSON_VALUE": parser.build_extract_json_with_path(exp.JSONExtractScalar),
     525            "LEN": _build_with_arg_as_text(exp.Length),
     526            "LEFT": _build_with_arg_as_text(exp.Left),
     527            "RIGHT": _build_with_arg_as_text(exp.Right),
     528            "REPLICATE": exp.Repeat.from_arg_list,
     529            "SQUARE": lambda args: exp.Pow(this=seq_get(args, 0), expression=exp.Literal.number(2)),
     530            "SYSDATETIME": exp.CurrentTimestamp.from_arg_list,
     531            "SUSER_NAME": exp.CurrentUser.from_arg_list,
     532            "SUSER_SNAME": exp.CurrentUser.from_arg_list,
     533            "SYSTEM_USER": exp.CurrentUser.from_arg_list,
     534            "TIMEFROMPARTS": _build_timefromparts,
     535        }
     536
     537        JOIN_HINTS = {"LOOP", "HASH", "MERGE", "REMOTE"}
     538
     539        RETURNS_TABLE_TOKENS = parser.Parser.ID_VAR_TOKENS - {
     540            TokenType.TABLE,
     541            *parser.Parser.TYPE_TOKENS,
     542        }
     543
     544        STATEMENT_PARSERS = {
     545            **parser.Parser.STATEMENT_PARSERS,
     546            TokenType.DECLARE: lambda self: self._parse_declare(),
     547        }
     548
     549        RANGE_PARSERS = {
     550            **parser.Parser.RANGE_PARSERS,
     551            TokenType.DCOLON: lambda self, this: self.expression(
     552                exp.ScopeResolution,
     553                this=this,
     554                expression=self._parse_function() or self._parse_var(any_token=True),
     555            ),
     556        }
     557
     558        # The DCOLON (::) operator serves as a scope resolution (exp.ScopeResolution) operator in T-SQL
     559        COLUMN_OPERATORS = {
     560            **parser.Parser.COLUMN_OPERATORS,
     561            TokenType.DCOLON: lambda self, this, to: self.expression(exp.Cast, this=this, to=to)
     562            if isinstance(to, exp.DataType) and to.this != exp.DataType.Type.USERDEFINED
     563            else self.expression(exp.ScopeResolution, this=this, expression=to),
     564        }
     565
     566        def _parse_dcolon(self) -> t.Optional[exp.Expression]:
     567            # We want to use _parse_types() if the first token after :: is a known type,
     568            # otherwise we could parse something like x::varchar(max) into a function
     569            if self._match_set(self.TYPE_TOKENS, advance=False):
     570                return self._parse_types()
     571
     572            return self._parse_function() or self._parse_types()
     573
     574        def _parse_options(self) -> t.Optional[t.List[exp.Expression]]:
     575            if not self._match(TokenType.OPTION):
     576                return None
     577
     578            def _parse_option() -> t.Optional[exp.Expression]:
     579                option = self._parse_var_from_options(OPTIONS)
     580                if not option:
     581                    return None
     582
     583                self._match(TokenType.EQ)
     584                return self.expression(
     585                    exp.QueryOption, this=option, expression=self._parse_primary_or_var()
     586                )
     587
     588            return self._parse_wrapped_csv(_parse_option)
     589
     590        def _parse_projections(self) -> t.List[exp.Expression]:
     591            """
     592            T-SQL supports the syntax alias = expression in the SELECT's projection list,
     593            so we transform all parsed Selects to convert their EQ projections into Aliases.
     594
     595            See: https://learn.microsoft.com/en-us/sql/t-sql/queries/select-clause-transact-sql?view=sql-server-ver16#syntax
     596            """
     597            return [
     598                (
     599                    exp.alias_(projection.expression, projection.this.this, copy=False)
     600                    if isinstance(projection, exp.EQ) and isinstance(projection.this, exp.Column)
     601                    else projection
     602                )
     603                for projection in super()._parse_projections()
     604            ]
     605
     606        def _parse_commit_or_rollback(self) -> exp.Commit | exp.Rollback:
     607            """Applies to SQL Server and Azure SQL Database
     608            COMMIT [ { TRAN | TRANSACTION }
     609                [ transaction_name | @tran_name_variable ] ]
     610                [ WITH ( DELAYED_DURABILITY = { OFF | ON } ) ]
     611
     612            ROLLBACK { TRAN | TRANSACTION }
     613                [ transaction_name | @tran_name_variable
     614                | savepoint_name | @savepoint_variable ]
     615            """
     616            rollback = self._prev.token_type == TokenType.ROLLBACK
     617
     618            self._match_texts(("TRAN", "TRANSACTION"))
     619            this = self._parse_id_var()
     620
     621            if rollback:
     622                return self.expression(exp.Rollback, this=this)
     623
     624            durability = None
     625            if self._match_pair(TokenType.WITH, TokenType.L_PAREN):
     626                self._match_text_seq("DELAYED_DURABILITY")
     627                self._match(TokenType.EQ)
     628
     629                if self._match_text_seq("OFF"):
     630                    durability = False
     631                else:
     632                    self._match(TokenType.ON)
     633                    durability = True
     634
     635                self._match_r_paren()
     636
     637            return self.expression(exp.Commit, this=this, durability=durability)
     638
     639        def _parse_transaction(self) -> exp.Transaction | exp.Command:
     640            """Applies to SQL Server and Azure SQL Database
     641            BEGIN { TRAN | TRANSACTION }
     642            [ { transaction_name | @tran_name_variable }
     643            [ WITH MARK [ 'description' ] ]
     644            ]
     645            """
     646            if self._match_texts(("TRAN", "TRANSACTION")):
     647                transaction = self.expression(exp.Transaction, this=self._parse_id_var())
     648                if self._match_text_seq("WITH", "MARK"):
     649                    transaction.set("mark", self._parse_string())
     650
     651                return transaction
     652
     653            return self._parse_as_command(self._prev)
     654
     655        def _parse_returns(self) -> exp.ReturnsProperty:
     656            table = self._parse_id_var(any_token=False, tokens=self.RETURNS_TABLE_TOKENS)
     657            returns = super()._parse_returns()
     658            returns.set("table", table)
     659            return returns
     660
     661        def _parse_convert(
     662            self, strict: bool, safe: t.Optional[bool] = None
     663        ) -> t.Optional[exp.Expression]:
     664            this = self._parse_types()
     665            self._match(TokenType.COMMA)
     666            args = [this, *self._parse_csv(self._parse_assignment)]
     667            convert = exp.Convert.from_arg_list(args)
     668            convert.set("safe", safe)
     669            convert.set("strict", strict)
     670            return convert
     671
     672        def _parse_user_defined_function(
     673            self, kind: t.Optional[TokenType] = None
     674        ) -> t.Optional[exp.Expression]:
     675            this = super()._parse_user_defined_function(kind=kind)
     676
     677            if (
     678                kind == TokenType.FUNCTION
     679                or isinstance(this, exp.UserDefinedFunction)
     680                or self._match(TokenType.ALIAS, advance=False)
     681            ):
     682                return this
     683
     684            expressions = self._parse_csv(self._parse_function_parameter)
     685            return self.expression(exp.UserDefinedFunction, this=this, expressions=expressions)
     686
     687        def _parse_id_var(
     688            self,
     689            any_token: bool = True,
     690            tokens: t.Optional[t.Collection[TokenType]] = None,
     691        ) -> t.Optional[exp.Expression]:
     692            is_temporary = self._match(TokenType.HASH)
     693            is_global = is_temporary and self._match(TokenType.HASH)
     694
     695            this = super()._parse_id_var(any_token=any_token, tokens=tokens)
     696            if this:
     697                if is_global:
     698                    this.set("global", True)
     699                elif is_temporary:
     700                    this.set("temporary", True)
     701
     702            return this
     703
     704        def _parse_create(self) -> exp.Create | exp.Command:
     705            create = super()._parse_create()
     706
     707            if isinstance(create, exp.Create):
     708                table = create.this.this if isinstance(create.this, exp.Schema) else create.this
     709                if isinstance(table, exp.Table) and table.this.args.get("temporary"):
     710                    if not create.args.get("properties"):
     711                        create.set("properties", exp.Properties(expressions=[]))
     712
     713                    create.args["properties"].append("expressions", exp.TemporaryProperty())
     714
     715            return create
     716
     717        def _parse_if(self) -> t.Optional[exp.Expression]:
     718            index = self._index
     719
     720            if self._match_text_seq("OBJECT_ID"):
     721                self._parse_wrapped_csv(self._parse_string)
     722                if self._match_text_seq("IS", "NOT", "NULL") and self._match(TokenType.DROP):
     723                    return self._parse_drop(exists=True)
     724                self._retreat(index)
     725
     726            return super()._parse_if()
     727
     728        def _parse_unique(self) -> exp.UniqueColumnConstraint:
     729            if self._match_texts(("CLUSTERED", "NONCLUSTERED")):
     730                this = self.CONSTRAINT_PARSERS[self._prev.text.upper()](self)
     731            else:
     732                this = self._parse_schema(self._parse_id_var(any_token=False))
     733
     734            return self.expression(exp.UniqueColumnConstraint, this=this)
     735
     736        def _parse_partition(self) -> t.Optional[exp.Partition]:
     737            if not self._match_text_seq("WITH", "(", "PARTITIONS"):
     738                return None
     739
     740            def parse_range():
     741                low = self._parse_bitwise()
     742                high = self._parse_bitwise() if self._match_text_seq("TO") else None
     743
     744                return (
     745                    self.expression(exp.PartitionRange, this=low, expression=high) if high else low
     746                )
     747
     748            partition = self.expression(
     749                exp.Partition, expressions=self._parse_wrapped_csv(parse_range)
     750            )
     751
     752            self._match_r_paren()
     753
     754            return partition
     755
     756        def _parse_declare(self) -> exp.Declare | exp.Command:
     757            index = self._index
     758            expressions = self._try_parse(partial(self._parse_csv, self._parse_declareitem))
     759
     760            if not expressions or self._curr:
     761                self._retreat(index)
     762                return self._parse_as_command(self._prev)
     763
     764            return self.expression(exp.Declare, expressions=expressions)
     765
     766        def _parse_declareitem(self) -> t.Optional[exp.DeclareItem]:
     767            var = self._parse_id_var()
     768            if not var:
     769                return None
     770
     771            value = None
     772            self._match(TokenType.ALIAS)
     773            if self._match(TokenType.TABLE):
     774                data_type = self._parse_schema()
     775            else:
     776                data_type = self._parse_types()
     777                if self._match(TokenType.EQ):
     778                    value = self._parse_bitwise()
     779
     780            return self.expression(exp.DeclareItem, this=var, kind=data_type, default=value)
     781
     782    class Generator(generator.Generator):
     783        LIMIT_IS_TOP = True
     784        QUERY_HINTS = False
     785        RETURNING_END = False
     786        NVL2_SUPPORTED = False
     787        ALTER_TABLE_INCLUDE_COLUMN_KEYWORD = False
     788        LIMIT_FETCH = "FETCH"
     789        COMPUTED_COLUMN_WITH_TYPE = False
     790        CTE_RECURSIVE_KEYWORD_REQUIRED = False
     791        ENSURE_BOOLS = True
     792        NULL_ORDERING_SUPPORTED = None
     793        SUPPORTS_SINGLE_ARG_CONCAT = False
     794        TABLESAMPLE_SEED_KEYWORD = "REPEATABLE"
     795        SUPPORTS_SELECT_INTO = True
     796        JSON_PATH_BRACKETED_KEY_SUPPORTED = False
     797        SUPPORTS_TO_NUMBER = False
     798        SET_OP_MODIFIERS = False
     799        COPY_PARAMS_EQ_REQUIRED = True
     800        PARSE_JSON_NAME = None
     801
     802        EXPRESSIONS_WITHOUT_NESTED_CTES = {
     803            exp.Create,
     804            exp.Delete,
     805            exp.Insert,
     806            exp.Intersect,
     807            exp.Except,
     808            exp.Merge,
     809            exp.Select,
     810            exp.Subquery,
     811            exp.Union,
     812            exp.Update,
     813        }
     814
     815        SUPPORTED_JSON_PATH_PARTS = {
     816            exp.JSONPathKey,
     817            exp.JSONPathRoot,
     818            exp.JSONPathSubscript,
     819        }
     820
     821        TYPE_MAPPING = {
     822            **generator.Generator.TYPE_MAPPING,
     823            exp.DataType.Type.BOOLEAN: "BIT",
     824            exp.DataType.Type.DECIMAL: "NUMERIC",
     825            exp.DataType.Type.DATETIME: "DATETIME2",
     826            exp.DataType.Type.DOUBLE: "FLOAT",
     827            exp.DataType.Type.INT: "INTEGER",
     828            exp.DataType.Type.ROWVERSION: "ROWVERSION",
     829            exp.DataType.Type.TEXT: "VARCHAR(MAX)",
     830            exp.DataType.Type.TIMESTAMP: "DATETIME2",
     831            exp.DataType.Type.TIMESTAMPTZ: "DATETIMEOFFSET",
     832            exp.DataType.Type.UTINYINT: "TINYINT",
     833            exp.DataType.Type.VARIANT: "SQL_VARIANT",
     834        }
     835
     836        TYPE_MAPPING.pop(exp.DataType.Type.NCHAR)
     837        TYPE_MAPPING.pop(exp.DataType.Type.NVARCHAR)
     838
     839        TRANSFORMS = {
     840            **generator.Generator.TRANSFORMS,
     841            exp.AnyValue: any_value_to_max_sql,
     842            exp.ArrayToString: rename_func("STRING_AGG"),
     843            exp.AutoIncrementColumnConstraint: lambda *_: "IDENTITY",
     844            exp.DateAdd: date_delta_sql("DATEADD"),
     845            exp.DateDiff: date_delta_sql("DATEDIFF"),
     846            exp.CTE: transforms.preprocess([qualify_derived_table_outputs]),
     847            exp.CurrentDate: rename_func("GETDATE"),
     848            exp.CurrentTimestamp: rename_func("GETDATE"),
     849            exp.DateStrToDate: datestrtodate_sql,
     850            exp.Extract: rename_func("DATEPART"),
     851            exp.GeneratedAsIdentityColumnConstraint: generatedasidentitycolumnconstraint_sql,
     852            exp.GroupConcat: _string_agg_sql,
     853            exp.If: rename_func("IIF"),
     854            exp.JSONExtract: _json_extract_sql,
     855            exp.JSONExtractScalar: _json_extract_sql,
     856            exp.LastDay: lambda self, e: self.func("EOMONTH", e.this),
     857            exp.Max: max_or_greatest,
     858            exp.MD5: lambda self, e: self.func("HASHBYTES", exp.Literal.string("MD5"), e.this),
     859            exp.Min: min_or_least,
     860            exp.NumberToStr: _format_sql,
     861            exp.Repeat: rename_func("REPLICATE"),
     862            exp.Select: transforms.preprocess(
     863                [
     864                    transforms.eliminate_distinct_on,
     865                    transforms.eliminate_semi_and_anti_joins,
     866                    transforms.eliminate_qualify,
     867                ]
     868            ),
     869            exp.Stddev: rename_func("STDEV"),
     870            exp.StrPosition: lambda self, e: self.func(
     871                "CHARINDEX", e.args.get("substr"), e.this, e.args.get("position")
     872            ),
     873            exp.Subquery: transforms.preprocess([qualify_derived_table_outputs]),
     874            exp.SHA: lambda self, e: self.func("HASHBYTES", exp.Literal.string("SHA1"), e.this),
     875            exp.SHA2: lambda self, e: self.func(
     876                "HASHBYTES", exp.Literal.string(f"SHA2_{e.args.get('length', 256)}"), e.this
     877            ),
     878            exp.TemporaryProperty: lambda self, e: "",
     879            exp.TimeStrToTime: lambda self, e: self.sql(
     880                exp.cast(e.this, exp.DataType.Type.DATETIME)
     881            ),
     882            exp.TimeToStr: _format_sql,
     883            exp.Trim: trim_sql,
     884            exp.TsOrDsAdd: date_delta_sql("DATEADD", cast=True),
     885            exp.TsOrDsDiff: date_delta_sql("DATEDIFF"),
     886        }
     887
     888        TRANSFORMS.pop(exp.ReturnsProperty)
     889
     890        PROPERTIES_LOCATION = {
     891            **generator.Generator.PROPERTIES_LOCATION,
     892            exp.VolatileProperty: exp.Properties.Location.UNSUPPORTED,
     893        }
     894
     895        def scope_resolution(self, rhs: str, scope_name: str) -> str:
     896            return f"{scope_name}::{rhs}"
     897
     898        def select_sql(self, expression: exp.Select) -> str:
     899            if expression.args.get("offset"):
     900                if not expression.args.get("order"):
     901                    # ORDER BY is required in order to use OFFSET in a query, so we use
     902                    # a noop order by, since we don't really care about the order.
     903                    # See: https://www.microsoftpressstore.com/articles/article.aspx?p=2314819
     904                    expression.order_by(exp.select(exp.null()).subquery(), copy=False)
     905
     906                limit = expression.args.get("limit")
     907                if isinstance(limit, exp.Limit):
     908                    # TOP and OFFSET can't be combined, we need use FETCH instead of TOP
     909                    # we replace here because otherwise TOP would be generated in select_sql
     910                    limit.replace(exp.Fetch(direction="FIRST", count=limit.expression))
     911
     912            return super().select_sql(expression)
     913
     914        def convert_sql(self, expression: exp.Convert) -> str:
     915            name = "TRY_CONVERT" if expression.args.get("safe") else "CONVERT"
     916            return self.func(
     917                name, expression.this, expression.expression, expression.args.get("style")
     918            )
     919
     920        def queryoption_sql(self, expression: exp.QueryOption) -> str:
     921            option = self.sql(expression, "this")
     922            value = self.sql(expression, "expression")
     923            if value:
     924                optional_equal_sign = "= " if option in OPTIONS_THAT_REQUIRE_EQUAL else ""
     925                return f"{option} {optional_equal_sign}{value}"
     926            return option
     927
     928        def lateral_op(self, expression: exp.Lateral) -> str:
     929            cross_apply = expression.args.get("cross_apply")
     930            if cross_apply is True:
     931                return "CROSS APPLY"
     932            if cross_apply is False:
     933                return "OUTER APPLY"
     934
     935            # TODO: perhaps we can check if the parent is a Join and transpile it appropriately
     936            self.unsupported("LATERAL clause is not supported.")
     937            return "LATERAL"
     938
     939        def timefromparts_sql(self, expression: exp.TimeFromParts) -> str:
     940            nano = expression.args.get("nano")
     941            if nano is not None:
     942                nano.pop()
     943                self.unsupported("Specifying nanoseconds is not supported in TIMEFROMPARTS.")
     944
     945            if expression.args.get("fractions") is None:
     946                expression.set("fractions", exp.Literal.number(0))
     947            if expression.args.get("precision") is None:
     948                expression.set("precision", exp.Literal.number(0))
     949
     950            return rename_func("TIMEFROMPARTS")(self, expression)
     951
     952        def timestampfromparts_sql(self, expression: exp.TimestampFromParts) -> str:
     953            zone = expression.args.get("zone")
     954            if zone is not None:
     955                zone.pop()
     956                self.unsupported("Time zone is not supported in DATETIMEFROMPARTS.")
     957
     958            nano = expression.args.get("nano")
     959            if nano is not None:
     960                nano.pop()
     961                self.unsupported("Specifying nanoseconds is not supported in DATETIMEFROMPARTS.")
     962
     963            if expression.args.get("milli") is None:
     964                expression.set("milli", exp.Literal.number(0))
     965
     966            return rename_func("DATETIMEFROMPARTS")(self, expression)
     967
     968        def setitem_sql(self, expression: exp.SetItem) -> str:
     969            this = expression.this
     970            if isinstance(this, exp.EQ) and not isinstance(this.left, exp.Parameter):
     971                # T-SQL does not use '=' in SET command, except when the LHS is a variable.
     972                return f"{self.sql(this.left)} {self.sql(this.right)}"
     973
     974            return super().setitem_sql(expression)
     975
     976        def boolean_sql(self, expression: exp.Boolean) -> str:
     977            if type(expression.parent) in BIT_TYPES:
     978                return "1" if expression.this else "0"
     979
     980            return "(1 = 1)" if expression.this else "(1 = 0)"
     981
     982        def is_sql(self, expression: exp.Is) -> str:
     983            if isinstance(expression.expression, exp.Boolean):
     984                return self.binary(expression, "=")
     985            return self.binary(expression, "IS")
     986
     987        def createable_sql(self, expression: exp.Create, locations: t.DefaultDict) -> str:
     988            sql = self.sql(expression, "this")
     989            properties = expression.args.get("properties")
     990
     991            if sql[:1] != "#" and any(
     992                isinstance(prop, exp.TemporaryProperty)
     993                for prop in (properties.expressions if properties else [])
     994            ):
     995                sql = f"[#{sql[1:]}" if sql.startswith("[") else f"#{sql}"
     996
     997            return sql
     998
     999        def create_sql(self, expression: exp.Create) -> str:
    1000            kind = expression.kind
    1001            exists = expression.args.pop("exists", None)
    1002
    1003            like_property = expression.find(exp.LikeProperty)
    1004            if like_property:
    1005                ctas_expression = like_property.this
    1006            else:
    1007                ctas_expression = expression.expression
    1008
    1009            if kind == "VIEW":
    1010                expression.this.set("catalog", None)
    1011                with_ = expression.args.get("with")
    1012                if ctas_expression and with_:
    1013                    # We've already preprocessed the Create expression to bubble up any nested CTEs,
    1014                    # but CREATE VIEW actually requires the WITH clause to come after it so we need
    1015                    # to amend the AST by moving the CTEs to the CREATE VIEW statement's query.
    1016                    ctas_expression.set("with", with_.pop())
    1017
    1018            sql = super().create_sql(expression)
    1019
    1020            table = expression.find(exp.Table)
    1021
    1022            # Convert CTAS statement to SELECT .. INTO ..
    1023            if kind == "TABLE" and ctas_expression:
    1024                if isinstance(ctas_expression, exp.UNWRAPPED_QUERIES):
    1025                    ctas_expression = ctas_expression.subquery()
    1026
    1027                select_into = exp.select("*").from_(exp.alias_(ctas_expression, "temp", table=True))
    1028                select_into.set("into", exp.Into(this=table))
    1029
    1030                if like_property:
    1031                    select_into.limit(0, copy=False)
    1032
    1033                sql = self.sql(select_into)
    1034
    1035            if exists:
    1036                identifier = self.sql(exp.Literal.string(exp.table_name(table) if table else ""))
    1037                sql = self.sql(exp.Literal.string(sql))
    1038                if kind == "SCHEMA":
    1039                    sql = f"""IF NOT EXISTS (SELECT * FROM information_schema.schemata WHERE schema_name = {identifier}) EXEC({sql})"""
    1040                elif kind == "TABLE":
    1041                    assert table
    1042                    where = exp.and_(
    1043                        exp.column("table_name").eq(table.name),
    1044                        exp.column("table_schema").eq(table.db) if table.db else None,
    1045                        exp.column("table_catalog").eq(table.catalog) if table.catalog else None,
    1046                    )
    1047                    sql = f"""IF NOT EXISTS (SELECT * FROM information_schema.tables WHERE {where}) EXEC({sql})"""
    1048                elif kind == "INDEX":
    1049                    index = self.sql(exp.Literal.string(expression.this.text("this")))
    1050                    sql = f"""IF NOT EXISTS (SELECT * FROM sys.indexes WHERE object_id = object_id({identifier}) AND name = {index}) EXEC({sql})"""
    1051            elif expression.args.get("replace"):
    1052                sql = sql.replace("CREATE OR REPLACE ", "CREATE OR ALTER ", 1)
    1053
    1054            return self.prepend_ctes(expression, sql)
    1055
    1056        def offset_sql(self, expression: exp.Offset) -> str:
    1057            return f"{super().offset_sql(expression)} ROWS"
    1058
    1059        def version_sql(self, expression: exp.Version) -> str:
    1060            name = "SYSTEM_TIME" if expression.name == "TIMESTAMP" else expression.name
    1061            this = f"FOR {name}"
    1062            expr = expression.expression
    1063            kind = expression.text("kind")
    1064            if kind in ("FROM", "BETWEEN"):
    1065                args = expr.expressions
    1066                sep = "TO" if kind == "FROM" else "AND"
    1067                expr_sql = f"{self.sql(seq_get(args, 0))} {sep} {self.sql(seq_get(args, 1))}"
    1068            else:
    1069                expr_sql = self.sql(expr)
    1070
    1071            expr_sql = f" {expr_sql}" if expr_sql else ""
    1072            return f"{this} {kind}{expr_sql}"
    1073
    1074        def returnsproperty_sql(self, expression: exp.ReturnsProperty) -> str:
    1075            table = expression.args.get("table")
    1076            table = f"{table} " if table else ""
    1077            return f"RETURNS {table}{self.sql(expression, 'this')}"
    1078
    1079        def returning_sql(self, expression: exp.Returning) -> str:
    1080            into = self.sql(expression, "into")
    1081            into = self.seg(f"INTO {into}") if into else ""
    1082            return f"{self.seg('OUTPUT')} {self.expressions(expression, flat=True)}{into}"
    1083
    1084        def transaction_sql(self, expression: exp.Transaction) -> str:
    1085            this = self.sql(expression, "this")
    1086            this = f" {this}" if this else ""
    1087            mark = self.sql(expression, "mark")
    1088            mark = f" WITH MARK {mark}" if mark else ""
    1089            return f"BEGIN TRANSACTION{this}{mark}"
    1090
    1091        def commit_sql(self, expression: exp.Commit) -> str:
    1092            this = self.sql(expression, "this")
    1093            this = f" {this}" if this else ""
    1094            durability = expression.args.get("durability")
    1095            durability = (
    1096                f" WITH (DELAYED_DURABILITY = {'ON' if durability else 'OFF'})"
    1097                if durability is not None
    1098                else ""
    1099            )
    1100            return f"COMMIT TRANSACTION{this}{durability}"
    1101
    1102        def rollback_sql(self, expression: exp.Rollback) -> str:
    1103            this = self.sql(expression, "this")
    1104            this = f" {this}" if this else ""
    1105            return f"ROLLBACK TRANSACTION{this}"
    1106
    1107        def identifier_sql(self, expression: exp.Identifier) -> str:
    1108            identifier = super().identifier_sql(expression)
    1109
    1110            if expression.args.get("global"):
    1111                identifier = f"##{identifier}"
    1112            elif expression.args.get("temporary"):
    1113                identifier = f"#{identifier}"
    1114
    1115            return identifier
    1116
    1117        def constraint_sql(self, expression: exp.Constraint) -> str:
    1118            this = self.sql(expression, "this")
    1119            expressions = self.expressions(expression, flat=True, sep=" ")
    1120            return f"CONSTRAINT {this} {expressions}"
    1121
    1122        def length_sql(self, expression: exp.Length) -> str:
    1123            return self._uncast_text(expression, "LEN")
    1124
    1125        def right_sql(self, expression: exp.Right) -> str:
    1126            return self._uncast_text(expression, "RIGHT")
    1127
    1128        def left_sql(self, expression: exp.Left) -> str:
    1129            return self._uncast_text(expression, "LEFT")
    1130
    1131        def _uncast_text(self, expression: exp.Expression, name: str) -> str:
    1132            this = expression.this
    1133            if isinstance(this, exp.Cast) and this.is_type(exp.DataType.Type.TEXT):
    1134                this_sql = self.sql(this, "this")
    1135            else:
    1136                this_sql = self.sql(this)
    1137            expression_sql = self.sql(expression, "expression")
    1138            return self.func(name, this_sql, expression_sql if expression_sql else None)
    1139
    1140        def partition_sql(self, expression: exp.Partition) -> str:
    1141            return f"WITH (PARTITIONS({self.expressions(expression, flat=True)}))"
    1142
    1143        def altertable_sql(self, expression: exp.AlterTable) -> str:
    1144            action = seq_get(expression.args.get("actions") or [], 0)
    1145            if isinstance(action, exp.RenameTable):
    1146                return f"EXEC sp_rename '{self.sql(expression.this)}', '{action.this.name}'"
    1147            return super().altertable_sql(expression)
    1148
    1149        def drop_sql(self, expression: exp.Drop) -> str:
    1150            if expression.args["kind"] == "VIEW":
    1151                expression.this.set("catalog", None)
    1152            return super().drop_sql(expression)
    1153
    1154        def declare_sql(self, expression: exp.Declare) -> str:
    1155            return f"DECLARE {self.expressions(expression, flat=True)}"
    1156
    1157        def declareitem_sql(self, expression: exp.DeclareItem) -> str:
    1158            variable = self.sql(expression, "this")
    1159            default = self.sql(expression, "default")
    1160            default = f" = {default}" if default else ""
    1161
    1162            kind = self.sql(expression, "kind")
    1163            if isinstance(expression.args.get("kind"), exp.Schema):
    1164                kind = f"TABLE {kind}"
    1165
    1166            return f"{variable} AS {kind}{default}"
    1167
    1168        def options_modifier(self, expression: exp.Expression) -> str:
    1169            options = self.expressions(expression, key="options")
    1170            return f" OPTION{self.wrap(options)}" if options else ""

NORMALIZATION_STRATEGY = <NormalizationStrategy.CASE_INSENSITIVE: 'CASE_INSENSITIVE'>

Specifies the strategy according to which identifiers should be normalized.

TIME_FORMAT = "'yyyy-mm-dd hh:mm:ss'"
SUPPORTS_SEMI_ANTI_JOIN = False

Whether `SEMI` or `ANTI` joins are supported.

LOG_BASE_FIRST: Optional[bool] = False

Whether the base comes first in the `LOG` function. Possible values: `True`, `False`, `None` (two arguments are not supported by `LOG`)

TYPED_DIVISION = True

Whether the behavior of `a / b` depends on the types of `a` and `b`. False means `a / b` is always float division. True means `a / b` is integer division if both `a` and `b` are integers.

CONCAT_COALESCE = True

A `NULL` arg in `CONCAT` yields `NULL` by default, but in some dialects it yields an empty string.

TIME_MAPPING: Dict[str, str] = {'year': '%Y', 'dayofyear': '%j', 'day': '%d', 'dy': '%d', 'y': '%Y', 'week': '%W', 'ww': '%W', 'wk': '%W', 'hour': '%h', 'hh': '%I', 'minute': '%M', 'mi': '%M', 'n': '%M', 'second': '%S', 'ss': '%S', 's': '%-S', 'millisecond': '%f', 'ms': '%f', 'weekday': '%W', 'dw': '%W', 'month': '%m', 'mm': '%M', 'm': '%-M', 'Y': '%Y', 'YYYY': '%Y', 'YY': '%y', 'MMMM': '%B', 'MMM': '%b', 'MM': '%m', 'M': '%-m', 'dddd': '%A', 'dd': '%d', 'd': '%-d', 'HH': '%H', 'H': '%-H', 'h': '%-I', 'S': '%f', 'yyyy': '%Y', 'yy': '%y'}

Associates this dialect's time formats with their equivalent Python `strftime` formats.

CONVERT_FORMAT_MAPPING = {'0': '%b %d %Y %-I:%M%p', '1': '%m/%d/%y', '2': '%y.%m.%d', '3': '%d/%m/%y', '4': '%d.%m.%y', '5': '%d-%m-%y', '6': '%d %b %y', '7': '%b %d, %y', '8': '%H:%M:%S', '9': '%b %d %Y %-I:%M:%S:%f%p', '10': 'mm-dd-yy', '11': 'yy/mm/dd', '12': 'yymmdd', '13': '%d %b %Y %H:%M:ss:%f', '14': '%H:%M:%S:%f', '20': '%Y-%m-%d %H:%M:%S', '21': '%Y-%m-%d %H:%M:%S.%f', '22': '%m/%d/%y %-I:%M:%S %p', '23': '%Y-%m-%d', '24': '%H:%M:%S', '25': '%Y-%m-%d %H:%M:%S.%f', '100': '%b %d %Y %-I:%M%p', '101': '%m/%d/%Y', '102': '%Y.%m.%d', '103': '%d/%m/%Y', '104': '%d.%m.%Y', '105': '%d-%m-%Y', '106': '%d %b %Y', '107': '%b %d, %Y', '108': '%H:%M:%S', '109': '%b %d %Y %-I:%M:%S:%f%p', '110': '%m-%d-%Y', '111': '%Y/%m/%d', '112': '%Y%m%d', '113': '%d %b %Y %H:%M:%S:%f', '114': '%H:%M:%S:%f', '120': '%Y-%m-%d %H:%M:%S', '121': '%Y-%m-%d %H:%M:%S.%f'}
FORMAT_TIME_MAPPING = {'y': '%B %Y', 'd': '%m/%d/%Y', 'H': '%-H', 'h': '%-I', 's': '%Y-%m-%d %H:%M:%S', 'D': '%A,%B,%Y', 'f': '%A,%B,%Y %-I:%M %p', 'F': '%A,%B,%Y %-I:%M:%S %p', 'g': '%m/%d/%Y %-I:%M %p', 'G': '%m/%d/%Y %-I:%M:%S %p', 'M': '%B %-d', 'm': '%B %-d', 'O': '%Y-%m-%dT%H:%M:%S', 'u': '%Y-%M-%D %H:%M:%S%z', 'U': '%A, %B %D, %Y %H:%M:%S%z', 'T': '%-I:%M:%S %p', 't': '%-I:%M', 'Y': '%a %Y'}
SUPPORTS_COLUMN_JOIN_MARKS = False

Whether the old-style outer join (+) syntax is supported.

tokenizer_class = <class 'TSQL.Tokenizer'>
jsonpath_tokenizer_class = <class 'sqlglot.tokens.JSONPathTokenizer'>
parser_class = <class 'TSQL.Parser'>
generator_class = <class 'TSQL.Generator'>
TIME_TRIE: Dict = {'y': {'e': {'a': {'r': {0: True}}}, 0: True, 'y': {'y': {'y': {0: True}}, 0: True}}, 'd': {'a': {'y': {'o': {'f': {'y': {'e': {'a': {'r': {0: True}}}}}}, 0: True}}, 'y': {0: True}, 'w': {0: True}, 'd': {'d': {'d': {0: True}}, 0: True}, 0: True}, 'w': {'e': {'e': {'k': {0: True, 'd': {'a': {'y': {0: True}}}}}}, 'w': {0: True}, 'k': {0: True}}, 'h': {'o': {'u': {'r': {0: True}}}, 'h': {0: True}, 0: True}, 'm': {'i': {'n': {'u': {'t': {'e': {0: True}}}}, 0: True, 'l': {'l': {'i': {'s': {'e': {'c': {'o': {'n': {'d': {0: True}}}}}}}}}}, 's': {0: True}, 'o': {'n': {'t': {'h': {0: True}}}}, 'm': {0: True}, 0: True}, 'n': {0: True}, 's': {'e': {'c': {'o': {'n': {'d': {0: True}}}}}, 's': {0: True}, 0: True}, 'Y': {0: True, 'Y': {'Y': {'Y': {0: True}}, 0: True}}, 'M': {'M': {'M': {'M': {0: True}, 0: True}, 0: True}, 0: True}, 'H': {'H': {0: True}, 0: True}, 'S': {0: True}}
FORMAT_TRIE: Dict = {'y': {'e': {'a': {'r': {0: True}}}, 0: True, 'y': {'y': {'y': {0: True}}, 0: True}}, 'd': {'a': {'y': {'o': {'f': {'y': {'e': {'a': {'r': {0: True}}}}}}, 0: True}}, 'y': {0: True}, 'w': {0: True}, 'd': {'d': {'d': {0: True}}, 0: True}, 0: True}, 'w': {'e': {'e': {'k': {0: True, 'd': {'a': {'y': {0: True}}}}}}, 'w': {0: True}, 'k': {0: True}}, 'h': {'o': {'u': {'r': {0: True}}}, 'h': {0: True}, 0: True}, 'm': {'i': {'n': {'u': {'t': {'e': {0: True}}}}, 0: True, 'l': {'l': {'i': {'s': {'e': {'c': {'o': {'n': {'d': {0: True}}}}}}}}}}, 's': {0: True}, 'o': {'n': {'t': {'h': {0: True}}}}, 'm': {0: True}, 0: True}, 'n': {0: True}, 's': {'e': {'c': {'o': {'n': {'d': {0: True}}}}}, 's': {0: True}, 0: True}, 'Y': {0: True, 'Y': {'Y': {'Y': {0: True}}, 0: True}}, 'M': {'M': {'M': {'M': {0: True}, 0: True}, 0: True}, 0: True}, 'H': {'H': {0: True}, 0: True}, 'S': {0: True}}
INVERSE_TIME_MAPPING: Dict[str, str] = {'%Y': 'yyyy', '%j': 'dayofyear', '%d': 'dd', '%W': 'dw', '%h': 'hour', '%I': 'hh', '%M': 'mm', '%S': 'ss', '%-S': 's', '%f': 'S', '%m': 'MM', '%-M': 'm', '%y': 'yy', '%B': 'MMMM', '%b': 'MMM', '%-m': 'M', '%A': 'dddd', '%-d': 'd', '%H': 'HH', '%-H': 'H', '%-I': 'h'}
INVERSE_TIME_TRIE: Dict = {'%': {'Y': {0: True}, 'j': {0: True}, 'd': {0: True}, 'W': {0: True}, 'h': {0: True}, 'I': {0: True}, 'M': {0: True}, 'S': {0: True}, '-': {'S': {0: True}, 'M': {0: True}, 'm': {0: True}, 'd': {0: True}, 'H': {0: True}, 'I': {0: True}}, 'f': {0: True}, 'm': {0: True}, 'y': {0: True}, 'B': {0: True}, 'b': {0: True}, 'A': {0: True}, 'H': {0: True}}}
INVERSE_FORMAT_MAPPING: Dict[str, str] = {}
INVERSE_FORMAT_TRIE: Dict = {}
ESCAPED_SEQUENCES: Dict[str, str] = {}
QUOTE_START = "'"
QUOTE_END = "'"
IDENTIFIER_START = '['
IDENTIFIER_END = ']'
BIT_START: Optional[str] = None
BIT_END: Optional[str] = None
HEX_START: Optional[str] = '0x'
HEX_END: Optional[str] = ''
BYTE_START: Optional[str] = None
BYTE_END: Optional[str] = None
UNICODE_START: Optional[str] = None
UNICODE_END: Optional[str] = None

##### Inherited Members

sqlglot.dialects.dialect.Dialect
  Dialect
  INDEX_OFFSET
  WEEK_OFFSET
  UNNEST_COLUMN_ONLY
  ALIAS_POST_TABLESAMPLE
  TABLESAMPLE_SIZE_IS_PERCENT
  IDENTIFIERS_CAN_START_WITH_DIGIT
  DPIPE_IS_STRING_CONCAT
  STRICT_STRING_CONCAT
  SUPPORTS_USER_DEFINED_TYPES
  COPY_PARAMS_ARE_CSV
  NORMALIZE_FUNCTIONS
  NULL_ORDERING
  SAFE_DIVISION
  HEX_LOWERCASE
  DATE_FORMAT
  DATEINT_FORMAT
  FORMAT_MAPPING
  UNESCAPED_SEQUENCES
  PSEUDOCOLUMNS
  PREFER_CTE_ALIAS_COLUMN
  FORCE_EARLY_ALIAS_REF_EXPANSION
  EXPAND_ALIAS_REFS_EARLY_ONLY_IN_GROUP_BY
  SUPPORTS_ORDER_BY_ALL
  DATE_PART_MAPPING
  TYPE_TO_EXPRESSIONS
  ANNOTATORS
  get_or_raise
  format_time
  settings
  normalize_identifier
  case_sensitive
  can_identify
  quote_identifier
  to_json_path
  parse
  parse_into
  generate
  transpile
  tokenize
  tokenizer
  jsonpath_tokenizer
  parser
  generator

class TSQL.Tokenizer(sqlglot.tokens.Tokenizer): View Source

    455    class Tokenizer(tokens.Tokenizer):
    456        IDENTIFIERS = [("[", "]"), '"']
    457        QUOTES = ["'", '"']
    458        HEX_STRINGS = [("0x", ""), ("0X", "")]
    459        VAR_SINGLE_TOKENS = {"@", "$", "#"}
    460
    461        KEYWORDS = {
    462            **tokens.Tokenizer.KEYWORDS,
    463            "CLUSTERED INDEX": TokenType.INDEX,
    464            "DATETIME2": TokenType.DATETIME,
    465            "DATETIMEOFFSET": TokenType.TIMESTAMPTZ,
    466            "DECLARE": TokenType.DECLARE,
    467            "EXEC": TokenType.COMMAND,
    468            "FOR SYSTEM_TIME": TokenType.TIMESTAMP_SNAPSHOT,
    469            "IMAGE": TokenType.IMAGE,
    470            "MONEY": TokenType.MONEY,
    471            "NONCLUSTERED INDEX": TokenType.INDEX,
    472            "NTEXT": TokenType.TEXT,
    473            "OPTION": TokenType.OPTION,
    474            "OUTPUT": TokenType.RETURNING,
    475            "PRINT": TokenType.COMMAND,
    476            "PROC": TokenType.PROCEDURE,
    477            "REAL": TokenType.FLOAT,
    478            "ROWVERSION": TokenType.ROWVERSION,
    479            "SMALLDATETIME": TokenType.DATETIME,
    480            "SMALLMONEY": TokenType.SMALLMONEY,
    481            "SQL_VARIANT": TokenType.VARIANT,
    482            "SYSTEM_USER": TokenType.CURRENT_USER,
    483            "TOP": TokenType.TOP,
    484            "TIMESTAMP": TokenType.ROWVERSION,
    485            "TINYINT": TokenType.UTINYINT,
    486            "UNIQUEIDENTIFIER": TokenType.UNIQUEIDENTIFIER,
    487            "UPDATE STATISTICS": TokenType.COMMAND,
    488            "XML": TokenType.XML,
    489        }
    490        KEYWORDS.pop("/*+")
    491
    492        COMMANDS = {*tokens.Tokenizer.COMMANDS, TokenType.END}

IDENTIFIERS = [('[', ']'), '"']
QUOTES = ["'", '"']
HEX_STRINGS = [('0x', ''), ('0X', '')]
VAR_SINGLE_TOKENS = {'@', '$', '#'}
KEYWORDS = {'{%': <TokenType.BLOCK_START: 'BLOCK_START'>, '{%+': <TokenType.BLOCK_START: 'BLOCK_START'>, '{%-': <TokenType.BLOCK_START: 'BLOCK_START'>, '%}': <TokenType.BLOCK_END: 'BLOCK_END'>, '+%}': <TokenType.BLOCK_END: 'BLOCK_END'>, '-%}': <TokenType.BLOCK_END: 'BLOCK_END'>, '{{+': <TokenType.BLOCK_START: 'BLOCK_START'>, '{{-': <TokenType.BLOCK_START: 'BLOCK_START'>, '+}}': <TokenType.BLOCK_END: 'BLOCK_END'>, '-}}': <TokenType.BLOCK_END: 'BLOCK_END'>, '==': <TokenType.EQ: 'EQ'>, '::': <TokenType.DCOLON: 'DCOLON'>, '||': <TokenType.DPIPE: 'DPIPE'>, '>=': <TokenType.GTE: 'GTE'>, '<=': <TokenType.LTE: 'LTE'>, '<>': <TokenType.NEQ: 'NEQ'>, '!=': <TokenType.NEQ: 'NEQ'>, ':=': <TokenType.COLON_EQ: 'COLON_EQ'>, '<=>': <TokenType.NULLSAFE_EQ: 'NULLSAFE_EQ'>, '->': <TokenType.ARROW: 'ARROW'>, '->>': <TokenType.DARROW: 'DARROW'>, '=>': <TokenType.FARROW: 'FARROW'>, '#>': <TokenType.HASH_ARROW: 'HASH_ARROW'>, '#>>': <TokenType.DHASH_ARROW: 'DHASH_ARROW'>, '<->': <TokenType.LR_ARROW: 'LR_ARROW'>, '&&': <TokenType.DAMP: 'DAMP'>, '??': <TokenType.DQMARK: 'DQMARK'>, 'ALL': <TokenType.ALL: 'ALL'>, 'ALWAYS': <TokenType.ALWAYS: 'ALWAYS'>, 'AND': <TokenType.AND: 'AND'>, 'ANTI': <TokenType.ANTI: 'ANTI'>, 'ANY': <TokenType.ANY: 'ANY'>, 'ASC': <TokenType.ASC: 'ASC'>, 'AS': <TokenType.ALIAS: 'ALIAS'>, 'ASOF': <TokenType.ASOF: 'ASOF'>, 'AUTOINCREMENT': <TokenType.AUTO_INCREMENT: 'AUTO_INCREMENT'>, 'AUTO_INCREMENT': <TokenType.AUTO_INCREMENT: 'AUTO_INCREMENT'>, 'BEGIN': <TokenType.BEGIN: 'BEGIN'>, 'BETWEEN': <TokenType.BETWEEN: 'BETWEEN'>, 'CACHE': <TokenType.CACHE: 'CACHE'>, 'UNCACHE': <TokenType.UNCACHE: 'UNCACHE'>, 'CASE': <TokenType.CASE: 'CASE'>, 'CHARACTER SET': <TokenType.CHARACTER_SET: 'CHARACTER_SET'>, 'CLUSTER BY': <TokenType.CLUSTER_BY: 'CLUSTER_BY'>, 'COLLATE': <TokenType.COLLATE: 'COLLATE'>, 'COLUMN': <TokenType.COLUMN: 'COLUMN'>, 'COMMIT': <TokenType.COMMIT: 'COMMIT'>, 'CONNECT BY': <TokenType.CONNECT_BY: 'CONNECT_BY'>, 'CONSTRAINT': <TokenType.CONSTRAINT: 'CONSTRAINT'>, 'COPY': <TokenType.COPY: 'COPY'>, 'CREATE': <TokenType.CREATE: 'CREATE'>, 'CROSS': <TokenType.CROSS: 'CROSS'>, 'CUBE': <TokenType.CUBE: 'CUBE'>, 'CURRENT_DATE': <TokenType.CURRENT_DATE: 'CURRENT_DATE'>, 'CURRENT_TIME': <TokenType.CURRENT_TIME: 'CURRENT_TIME'>, 'CURRENT_TIMESTAMP': <TokenType.CURRENT_TIMESTAMP: 'CURRENT_TIMESTAMP'>, 'CURRENT_USER': <TokenType.CURRENT_USER: 'CURRENT_USER'>, 'DATABASE': <TokenType.DATABASE: 'DATABASE'>, 'DEFAULT': <TokenType.DEFAULT: 'DEFAULT'>, 'DELETE': <TokenType.DELETE: 'DELETE'>, 'DESC': <TokenType.DESC: 'DESC'>, 'DESCRIBE': <TokenType.DESCRIBE: 'DESCRIBE'>, 'DISTINCT': <TokenType.DISTINCT: 'DISTINCT'>, 'DISTRIBUTE BY': <TokenType.DISTRIBUTE_BY: 'DISTRIBUTE_BY'>, 'DIV': <TokenType.DIV: 'DIV'>, 'DROP': <TokenType.DROP: 'DROP'>, 'ELSE': <TokenType.ELSE: 'ELSE'>, 'END': <TokenType.END: 'END'>, 'ENUM': <TokenType.ENUM: 'ENUM'>, 'ESCAPE': <TokenType.ESCAPE: 'ESCAPE'>, 'EXCEPT': <TokenType.EXCEPT: 'EXCEPT'>, 'EXECUTE': <TokenType.EXECUTE: 'EXECUTE'>, 'EXISTS': <TokenType.EXISTS: 'EXISTS'>, 'FALSE': <TokenType.FALSE: 'FALSE'>, 'FETCH': <TokenType.FETCH: 'FETCH'>, 'FILTER': <TokenType.FILTER: 'FILTER'>, 'FIRST': <TokenType.FIRST: 'FIRST'>, 'FULL': <TokenType.FULL: 'FULL'>, 'FUNCTION': <TokenType.FUNCTION: 'FUNCTION'>, 'FOR': <TokenType.FOR: 'FOR'>, 'FOREIGN KEY': <TokenType.FOREIGN_KEY: 'FOREIGN_KEY'>, 'FORMAT': <TokenType.FORMAT: 'FORMAT'>, 'FROM': <TokenType.FROM: 'FROM'>, 'GEOGRAPHY': <TokenType.GEOGRAPHY: 'GEOGRAPHY'>, 'GEOMETRY': <TokenType.GEOMETRY: 'GEOMETRY'>, 'GLOB': <TokenType.GLOB: 'GLOB'>, 'GROUP BY': <TokenType.GROUP_BY: 'GROUP_BY'>, 'GROUPING SETS': <TokenType.GROUPING_SETS: 'GROUPING_SETS'>, 'HAVING': <TokenType.HAVING: 'HAVING'>, 'ILIKE': <TokenType.ILIKE: 'ILIKE'>, 'IN': <TokenType.IN: 'IN'>, 'INDEX': <TokenType.INDEX: 'INDEX'>, 'INET': <TokenType.INET: 'INET'>, 'INNER': <TokenType.INNER: 'INNER'>, 'INSERT': <TokenType.INSERT: 'INSERT'>, 'INTERVAL': <TokenType.INTERVAL: 'INTERVAL'>, 'INTERSECT': <TokenType.INTERSECT: 'INTERSECT'>, 'INTO': <TokenType.INTO: 'INTO'>, 'IS': <TokenType.IS: 'IS'>, 'ISNULL': <TokenType.ISNULL: 'ISNULL'>, 'JOIN': <TokenType.JOIN: 'JOIN'>, 'KEEP': <TokenType.KEEP: 'KEEP'>, 'KILL': <TokenType.KILL: 'KILL'>, 'LATERAL': <TokenType.LATERAL: 'LATERAL'>, 'LEFT': <TokenType.LEFT: 'LEFT'>, 'LIKE': <TokenType.LIKE: 'LIKE'>, 'LIMIT': <TokenType.LIMIT: 'LIMIT'>, 'LOAD': <TokenType.LOAD: 'LOAD'>, 'LOCK': <TokenType.LOCK: 'LOCK'>, 'MERGE': <TokenType.MERGE: 'MERGE'>, 'NATURAL': <TokenType.NATURAL: 'NATURAL'>, 'NEXT': <TokenType.NEXT: 'NEXT'>, 'NOT': <TokenType.NOT: 'NOT'>, 'NOTNULL': <TokenType.NOTNULL: 'NOTNULL'>, 'NULL': <TokenType.NULL: 'NULL'>, 'OBJECT': <TokenType.OBJECT: 'OBJECT'>, 'OFFSET': <TokenType.OFFSET: 'OFFSET'>, 'ON': <TokenType.ON: 'ON'>, 'OR': <TokenType.OR: 'OR'>, 'XOR': <TokenType.XOR: 'XOR'>, 'ORDER BY': <TokenType.ORDER_BY: 'ORDER_BY'>, 'ORDINALITY': <TokenType.ORDINALITY: 'ORDINALITY'>, 'OUTER': <TokenType.OUTER: 'OUTER'>, 'OVER': <TokenType.OVER: 'OVER'>, 'OVERLAPS': <TokenType.OVERLAPS: 'OVERLAPS'>, 'OVERWRITE': <TokenType.OVERWRITE: 'OVERWRITE'>, 'PARTITION': <TokenType.PARTITION: 'PARTITION'>, 'PARTITION BY': <TokenType.PARTITION_BY: 'PARTITION_BY'>, 'PARTITIONED BY': <TokenType.PARTITION_BY: 'PARTITION_BY'>, 'PARTITIONED_BY': <TokenType.PARTITION_BY: 'PARTITION_BY'>, 'PERCENT': <TokenType.PERCENT: 'PERCENT'>, 'PIVOT': <TokenType.PIVOT: 'PIVOT'>, 'PRAGMA': <TokenType.PRAGMA: 'PRAGMA'>, 'PRIMARY KEY': <TokenType.PRIMARY_KEY: 'PRIMARY_KEY'>, 'PROCEDURE': <TokenType.PROCEDURE: 'PROCEDURE'>, 'QUALIFY': <TokenType.QUALIFY: 'QUALIFY'>, 'RANGE': <TokenType.RANGE: 'RANGE'>, 'RECURSIVE': <TokenType.RECURSIVE: 'RECURSIVE'>, 'REGEXP': <TokenType.RLIKE: 'RLIKE'>, 'REPLACE': <TokenType.REPLACE: 'REPLACE'>, 'RETURNING': <TokenType.RETURNING: 'RETURNING'>, 'REFERENCES': <TokenType.REFERENCES: 'REFERENCES'>, 'RIGHT': <TokenType.RIGHT: 'RIGHT'>, 'RLIKE': <TokenType.RLIKE: 'RLIKE'>, 'ROLLBACK': <TokenType.ROLLBACK: 'ROLLBACK'>, 'ROLLUP': <TokenType.ROLLUP: 'ROLLUP'>, 'ROW': <TokenType.ROW: 'ROW'>, 'ROWS': <TokenType.ROWS: 'ROWS'>, 'SCHEMA': <TokenType.SCHEMA: 'SCHEMA'>, 'SELECT': <TokenType.SELECT: 'SELECT'>, 'SEMI': <TokenType.SEMI: 'SEMI'>, 'SET': <TokenType.SET: 'SET'>, 'SETTINGS': <TokenType.SETTINGS: 'SETTINGS'>, 'SHOW': <TokenType.SHOW: 'SHOW'>, 'SIMILAR TO': <TokenType.SIMILAR_TO: 'SIMILAR_TO'>, 'SOME': <TokenType.SOME: 'SOME'>, 'SORT BY': <TokenType.SORT_BY: 'SORT_BY'>, 'START WITH': <TokenType.START_WITH: 'START_WITH'>, 'STRAIGHT_JOIN': <TokenType.STRAIGHT_JOIN: 'STRAIGHT_JOIN'>, 'TABLE': <TokenType.TABLE: 'TABLE'>, 'TABLESAMPLE': <TokenType.TABLE_SAMPLE: 'TABLE_SAMPLE'>, 'TEMP': <TokenType.TEMPORARY: 'TEMPORARY'>, 'TEMPORARY': <TokenType.TEMPORARY: 'TEMPORARY'>, 'THEN': <TokenType.THEN: 'THEN'>, 'TRUE': <TokenType.TRUE: 'TRUE'>, 'TRUNCATE': <TokenType.TRUNCATE: 'TRUNCATE'>, 'UNION': <TokenType.UNION: 'UNION'>, 'UNKNOWN': <TokenType.UNKNOWN: 'UNKNOWN'>, 'UNNEST': <TokenType.UNNEST: 'UNNEST'>, 'UNPIVOT': <TokenType.UNPIVOT: 'UNPIVOT'>, 'UPDATE': <TokenType.UPDATE: 'UPDATE'>, 'USE': <TokenType.USE: 'USE'>, 'USING': <TokenType.USING: 'USING'>, 'UUID': <TokenType.UUID: 'UUID'>, 'VALUES': <TokenType.VALUES: 'VALUES'>, 'VIEW': <TokenType.VIEW: 'VIEW'>, 'VOLATILE': <TokenType.VOLATILE: 'VOLATILE'>, 'WHEN': <TokenType.WHEN: 'WHEN'>, 'WHERE': <TokenType.WHERE: 'WHERE'>, 'WINDOW': <TokenType.WINDOW: 'WINDOW'>, 'WITH': <TokenType.WITH: 'WITH'>, 'APPLY': <TokenType.APPLY: 'APPLY'>, 'ARRAY': <TokenType.ARRAY: 'ARRAY'>, 'BIT': <TokenType.BIT: 'BIT'>, 'BOOL': <TokenType.BOOLEAN: 'BOOLEAN'>, 'BOOLEAN': <TokenType.BOOLEAN: 'BOOLEAN'>, 'BYTE': <TokenType.TINYINT: 'TINYINT'>, 'MEDIUMINT': <TokenType.MEDIUMINT: 'MEDIUMINT'>, 'INT1': <TokenType.TINYINT: 'TINYINT'>, 'TINYINT': <TokenType.UTINYINT: 'UTINYINT'>, 'INT16': <TokenType.SMALLINT: 'SMALLINT'>, 'SHORT': <TokenType.SMALLINT: 'SMALLINT'>, 'SMALLINT': <TokenType.SMALLINT: 'SMALLINT'>, 'INT128': <TokenType.INT128: 'INT128'>, 'HUGEINT': <TokenType.INT128: 'INT128'>, 'INT2': <TokenType.SMALLINT: 'SMALLINT'>, 'INTEGER': <TokenType.INT: 'INT'>, 'INT': <TokenType.INT: 'INT'>, 'INT4': <TokenType.INT: 'INT'>, 'INT32': <TokenType.INT: 'INT'>, 'INT64': <TokenType.BIGINT: 'BIGINT'>, 'LONG': <TokenType.BIGINT: 'BIGINT'>, 'BIGINT': <TokenType.BIGINT: 'BIGINT'>, 'INT8': <TokenType.TINYINT: 'TINYINT'>, 'UINT': <TokenType.UINT: 'UINT'>, 'DEC': <TokenType.DECIMAL: 'DECIMAL'>, 'DECIMAL': <TokenType.DECIMAL: 'DECIMAL'>, 'BIGDECIMAL': <TokenType.BIGDECIMAL: 'BIGDECIMAL'>, 'BIGNUMERIC': <TokenType.BIGDECIMAL: 'BIGDECIMAL'>, 'LIST': <TokenType.LIST: 'LIST'>, 'MAP': <TokenType.MAP: 'MAP'>, 'NULLABLE': <TokenType.NULLABLE: 'NULLABLE'>, 'NUMBER': <TokenType.DECIMAL: 'DECIMAL'>, 'NUMERIC': <TokenType.DECIMAL: 'DECIMAL'>, 'FIXED': <TokenType.DECIMAL: 'DECIMAL'>, 'REAL': <TokenType.FLOAT: 'FLOAT'>, 'FLOAT': <TokenType.FLOAT: 'FLOAT'>, 'FLOAT4': <TokenType.FLOAT: 'FLOAT'>, 'FLOAT8': <TokenType.DOUBLE: 'DOUBLE'>, 'DOUBLE': <TokenType.DOUBLE: 'DOUBLE'>, 'DOUBLE PRECISION': <TokenType.DOUBLE: 'DOUBLE'>, 'JSON': <TokenType.JSON: 'JSON'>, 'JSONB': <TokenType.JSONB: 'JSONB'>, 'CHAR': <TokenType.CHAR: 'CHAR'>, 'CHARACTER': <TokenType.CHAR: 'CHAR'>, 'NCHAR': <TokenType.NCHAR: 'NCHAR'>, 'VARCHAR': <TokenType.VARCHAR: 'VARCHAR'>, 'VARCHAR2': <TokenType.VARCHAR: 'VARCHAR'>, 'NVARCHAR': <TokenType.NVARCHAR: 'NVARCHAR'>, 'NVARCHAR2': <TokenType.NVARCHAR: 'NVARCHAR'>, 'BPCHAR': <TokenType.BPCHAR: 'BPCHAR'>, 'STR': <TokenType.TEXT: 'TEXT'>, 'STRING': <TokenType.TEXT: 'TEXT'>, 'TEXT': <TokenType.TEXT: 'TEXT'>, 'LONGTEXT': <TokenType.LONGTEXT: 'LONGTEXT'>, 'MEDIUMTEXT': <TokenType.MEDIUMTEXT: 'MEDIUMTEXT'>, 'TINYTEXT': <TokenType.TINYTEXT: 'TINYTEXT'>, 'CLOB': <TokenType.TEXT: 'TEXT'>, 'LONGVARCHAR': <TokenType.TEXT: 'TEXT'>, 'BINARY': <TokenType.BINARY: 'BINARY'>, 'BLOB': <TokenType.VARBINARY: 'VARBINARY'>, 'LONGBLOB': <TokenType.LONGBLOB: 'LONGBLOB'>, 'MEDIUMBLOB': <TokenType.MEDIUMBLOB: 'MEDIUMBLOB'>, 'TINYBLOB': <TokenType.TINYBLOB: 'TINYBLOB'>, 'BYTEA': <TokenType.VARBINARY: 'VARBINARY'>, 'VARBINARY': <TokenType.VARBINARY: 'VARBINARY'>, 'TIME': <TokenType.TIME: 'TIME'>, 'TIMETZ': <TokenType.TIMETZ: 'TIMETZ'>, 'TIMESTAMP': <TokenType.ROWVERSION: 'ROWVERSION'>, 'TIMESTAMPTZ': <TokenType.TIMESTAMPTZ: 'TIMESTAMPTZ'>, 'TIMESTAMPLTZ': <TokenType.TIMESTAMPLTZ: 'TIMESTAMPLTZ'>, 'TIMESTAMP_LTZ': <TokenType.TIMESTAMPLTZ: 'TIMESTAMPLTZ'>, 'TIMESTAMPNTZ': <TokenType.TIMESTAMPNTZ: 'TIMESTAMPNTZ'>, 'TIMESTAMP_NTZ': <TokenType.TIMESTAMPNTZ: 'TIMESTAMPNTZ'>, 'DATE': <TokenType.DATE: 'DATE'>, 'DATETIME': <TokenType.DATETIME: 'DATETIME'>, 'INT4RANGE': <TokenType.INT4RANGE: 'INT4RANGE'>, 'INT4MULTIRANGE': <TokenType.INT4MULTIRANGE: 'INT4MULTIRANGE'>, 'INT8RANGE': <TokenType.INT8RANGE: 'INT8RANGE'>, 'INT8MULTIRANGE': <TokenType.INT8MULTIRANGE: 'INT8MULTIRANGE'>, 'NUMRANGE': <TokenType.NUMRANGE: 'NUMRANGE'>, 'NUMMULTIRANGE': <TokenType.NUMMULTIRANGE: 'NUMMULTIRANGE'>, 'TSRANGE': <TokenType.TSRANGE: 'TSRANGE'>, 'TSMULTIRANGE': <TokenType.TSMULTIRANGE: 'TSMULTIRANGE'>, 'TSTZRANGE': <TokenType.TSTZRANGE: 'TSTZRANGE'>, 'TSTZMULTIRANGE': <TokenType.TSTZMULTIRANGE: 'TSTZMULTIRANGE'>, 'DATERANGE': <TokenType.DATERANGE: 'DATERANGE'>, 'DATEMULTIRANGE': <TokenType.DATEMULTIRANGE: 'DATEMULTIRANGE'>, 'UNIQUE': <TokenType.UNIQUE: 'UNIQUE'>, 'VECTOR': <TokenType.VECTOR: 'VECTOR'>, 'STRUCT': <TokenType.STRUCT: 'STRUCT'>, 'SEQUENCE': <TokenType.SEQUENCE: 'SEQUENCE'>, 'VARIANT': <TokenType.VARIANT: 'VARIANT'>, 'ALTER': <TokenType.ALTER: 'ALTER'>, 'ANALYZE': <TokenType.COMMAND: 'COMMAND'>, 'CALL': <TokenType.COMMAND: 'COMMAND'>, 'COMMENT': <TokenType.COMMENT: 'COMMENT'>, 'EXPLAIN': <TokenType.COMMAND: 'COMMAND'>, 'GRANT': <TokenType.COMMAND: 'COMMAND'>, 'OPTIMIZE': <TokenType.COMMAND: 'COMMAND'>, 'PREPARE': <TokenType.COMMAND: 'COMMAND'>, 'VACUUM': <TokenType.COMMAND: 'COMMAND'>, 'USER-DEFINED': <TokenType.USERDEFINED: 'USERDEFINED'>, 'FOR VERSION': <TokenType.VERSION_SNAPSHOT: 'VERSION_SNAPSHOT'>, 'FOR TIMESTAMP': <TokenType.TIMESTAMP_SNAPSHOT: 'TIMESTAMP_SNAPSHOT'>, 'CLUSTERED INDEX': <TokenType.INDEX: 'INDEX'>, 'DATETIME2': <TokenType.DATETIME: 'DATETIME'>, 'DATETIMEOFFSET': <TokenType.TIMESTAMPTZ: 'TIMESTAMPTZ'>, 'DECLARE': <TokenType.DECLARE: 'DECLARE'>, 'EXEC': <TokenType.COMMAND: 'COMMAND'>, 'FOR SYSTEM_TIME': <TokenType.TIMESTAMP_SNAPSHOT: 'TIMESTAMP_SNAPSHOT'>, 'IMAGE': <TokenType.IMAGE: 'IMAGE'>, 'MONEY': <TokenType.MONEY: 'MONEY'>, 'NONCLUSTERED INDEX': <TokenType.INDEX: 'INDEX'>, 'NTEXT': <TokenType.TEXT: 'TEXT'>, 'OPTION': <TokenType.OPTION: 'OPTION'>, 'OUTPUT': <TokenType.RETURNING: 'RETURNING'>, 'PRINT': <TokenType.COMMAND: 'COMMAND'>, 'PROC': <TokenType.PROCEDURE: 'PROCEDURE'>, 'ROWVERSION': <TokenType.ROWVERSION: 'ROWVERSION'>, 'SMALLDATETIME': <TokenType.DATETIME: 'DATETIME'>, 'SMALLMONEY': <TokenType.SMALLMONEY: 'SMALLMONEY'>, 'SQL_VARIANT': <TokenType.VARIANT: 'VARIANT'>, 'SYSTEM_USER': <TokenType.CURRENT_USER: 'CURRENT_USER'>, 'TOP': <TokenType.TOP: 'TOP'>, 'UNIQUEIDENTIFIER': <TokenType.UNIQUEIDENTIFIER: 'UNIQUEIDENTIFIER'>, 'UPDATE STATISTICS': <TokenType.COMMAND: 'COMMAND'>, 'XML': <TokenType.XML: 'XML'>}
COMMANDS = {<TokenType.SHOW: 'SHOW'>, <TokenType.COMMAND: 'COMMAND'>, <TokenType.EXECUTE: 'EXECUTE'>, <TokenType.FETCH: 'FETCH'>, <TokenType.END: 'END'>}

##### Inherited Members

sqlglot.tokens.Tokenizer
  Tokenizer
  SINGLE_TOKENS
  BIT_STRINGS
  BYTE_STRINGS
  RAW_STRINGS
  HEREDOC_STRINGS
  UNICODE_STRINGS
  IDENTIFIER_ESCAPES
  STRING_ESCAPES
  HEREDOC_TAG_IS_IDENTIFIER
  HEREDOC_STRING_ALTERNATIVE
  STRING_ESCAPES_ALLOWED_IN_RAW_STRINGS
  WHITE_SPACE
  COMMAND_PREFIX_TOKENS
  NUMERIC_LITERALS
  COMMENTS
  dialect
  reset
  tokenize
  tokenize_rs
  size
  sql
  tokens

class TSQL.Parser(sqlglot.parser.Parser): View Source

    494    class Parser(parser.Parser):
    495        SET_REQUIRES_ASSIGNMENT_DELIMITER = False
    496        LOG_DEFAULTS_TO_LN = True
    497        ALTER_TABLE_ADD_REQUIRED_FOR_EACH_COLUMN = False
    498        STRING_ALIASES = True
    499        NO_PAREN_IF_COMMANDS = False
    500
    501        QUERY_MODIFIER_PARSERS = {
    502            **parser.Parser.QUERY_MODIFIER_PARSERS,
    503            TokenType.OPTION: lambda self: ("options", self._parse_options()),
    504        }
    505
    506        FUNCTIONS = {
    507            **parser.Parser.FUNCTIONS,
    508            "CHARINDEX": lambda args: exp.StrPosition(
    509                this=seq_get(args, 1),
    510                substr=seq_get(args, 0),
    511                position=seq_get(args, 2),
    512            ),
    513            "DATEADD": build_date_delta(exp.DateAdd, unit_mapping=DATE_DELTA_INTERVAL),
    514            "DATEDIFF": _build_date_delta(exp.DateDiff, unit_mapping=DATE_DELTA_INTERVAL),
    515            "DATENAME": _build_formatted_time(exp.TimeToStr, full_format_mapping=True),
    516            "DATEPART": _build_formatted_time(exp.TimeToStr),
    517            "DATETIMEFROMPARTS": _build_datetimefromparts,
    518            "EOMONTH": _build_eomonth,
    519            "FORMAT": _build_format,
    520            "GETDATE": exp.CurrentTimestamp.from_arg_list,
    521            "HASHBYTES": _build_hashbytes,
    522            "ISNULL": exp.Coalesce.from_arg_list,
    523            "JSON_QUERY": _build_json_query,
    524            "JSON_VALUE": parser.build_extract_json_with_path(exp.JSONExtractScalar),
    525            "LEN": _build_with_arg_as_text(exp.Length),
    526            "LEFT": _build_with_arg_as_text(exp.Left),
    527            "RIGHT": _build_with_arg_as_text(exp.Right),
    528            "REPLICATE": exp.Repeat.from_arg_list,
    529            "SQUARE": lambda args: exp.Pow(this=seq_get(args, 0), expression=exp.Literal.number(2)),
    530            "SYSDATETIME": exp.CurrentTimestamp.from_arg_list,
    531            "SUSER_NAME": exp.CurrentUser.from_arg_list,
    532            "SUSER_SNAME": exp.CurrentUser.from_arg_list,
    533            "SYSTEM_USER": exp.CurrentUser.from_arg_list,
    534            "TIMEFROMPARTS": _build_timefromparts,
    535        }
    536
    537        JOIN_HINTS = {"LOOP", "HASH", "MERGE", "REMOTE"}
    538
    539        RETURNS_TABLE_TOKENS = parser.Parser.ID_VAR_TOKENS - {
    540            TokenType.TABLE,
    541            *parser.Parser.TYPE_TOKENS,
    542        }
    543
    544        STATEMENT_PARSERS = {
    545            **parser.Parser.STATEMENT_PARSERS,
    546            TokenType.DECLARE: lambda self: self._parse_declare(),
    547        }
    548
    549        RANGE_PARSERS = {
    550            **parser.Parser.RANGE_PARSERS,
    551            TokenType.DCOLON: lambda self, this: self.expression(
    552                exp.ScopeResolution,
    553                this=this,
    554                expression=self._parse_function() or self._parse_var(any_token=True),
    555            ),
    556        }
    557
    558        # The DCOLON (::) operator serves as a scope resolution (exp.ScopeResolution) operator in T-SQL
    559        COLUMN_OPERATORS = {
    560            **parser.Parser.COLUMN_OPERATORS,
    561            TokenType.DCOLON: lambda self, this, to: self.expression(exp.Cast, this=this, to=to)
    562            if isinstance(to, exp.DataType) and to.this != exp.DataType.Type.USERDEFINED
    563            else self.expression(exp.ScopeResolution, this=this, expression=to),
    564        }
    565
    566        def _parse_dcolon(self) -> t.Optional[exp.Expression]:
    567            # We want to use _parse_types() if the first token after :: is a known type,
    568            # otherwise we could parse something like x::varchar(max) into a function
    569            if self._match_set(self.TYPE_TOKENS, advance=False):
    570                return self._parse_types()
    571
    572            return self._parse_function() or self._parse_types()
    573
    574        def _parse_options(self) -> t.Optional[t.List[exp.Expression]]:
    575            if not self._match(TokenType.OPTION):
    576                return None
    577
    578            def _parse_option() -> t.Optional[exp.Expression]:
    579                option = self._parse_var_from_options(OPTIONS)
    580                if not option:
    581                    return None
    582
    583                self._match(TokenType.EQ)
    584                return self.expression(
    585                    exp.QueryOption, this=option, expression=self._parse_primary_or_var()
    586                )
    587
    588            return self._parse_wrapped_csv(_parse_option)
    589
    590        def _parse_projections(self) -> t.List[exp.Expression]:
    591            """
    592            T-SQL supports the syntax alias = expression in the SELECT's projection list,
    593            so we transform all parsed Selects to convert their EQ projections into Aliases.
    594
    595            See: https://learn.microsoft.com/en-us/sql/t-sql/queries/select-clause-transact-sql?view=sql-server-ver16#syntax
    596            """
    597            return [
    598                (
    599                    exp.alias_(projection.expression, projection.this.this, copy=False)
    600                    if isinstance(projection, exp.EQ) and isinstance(projection.this, exp.Column)
    601                    else projection
    602                )
    603                for projection in super()._parse_projections()
    604            ]
    605
    606        def _parse_commit_or_rollback(self) -> exp.Commit | exp.Rollback:
    607            """Applies to SQL Server and Azure SQL Database
    608            COMMIT [ { TRAN | TRANSACTION }
    609                [ transaction_name | @tran_name_variable ] ]
    610                [ WITH ( DELAYED_DURABILITY = { OFF | ON } ) ]
    611
    612            ROLLBACK { TRAN | TRANSACTION }
    613                [ transaction_name | @tran_name_variable
    614                | savepoint_name | @savepoint_variable ]
    615            """
    616            rollback = self._prev.token_type == TokenType.ROLLBACK
    617
    618            self._match_texts(("TRAN", "TRANSACTION"))
    619            this = self._parse_id_var()
    620
    621            if rollback:
    622                return self.expression(exp.Rollback, this=this)
    623
    624            durability = None
    625            if self._match_pair(TokenType.WITH, TokenType.L_PAREN):
    626                self._match_text_seq("DELAYED_DURABILITY")
    627                self._match(TokenType.EQ)
    628
    629                if self._match_text_seq("OFF"):
    630                    durability = False
    631                else:
    632                    self._match(TokenType.ON)
    633                    durability = True
    634
    635                self._match_r_paren()
    636
    637            return self.expression(exp.Commit, this=this, durability=durability)
    638
    639        def _parse_transaction(self) -> exp.Transaction | exp.Command:
    640            """Applies to SQL Server and Azure SQL Database
    641            BEGIN { TRAN | TRANSACTION }
    642            [ { transaction_name | @tran_name_variable }
    643            [ WITH MARK [ 'description' ] ]
    644            ]
    645            """
    646            if self._match_texts(("TRAN", "TRANSACTION")):
    647                transaction = self.expression(exp.Transaction, this=self._parse_id_var())
    648                if self._match_text_seq("WITH", "MARK"):
    649                    transaction.set("mark", self._parse_string())
    650
    651                return transaction
    652
    653            return self._parse_as_command(self._prev)
    654
    655        def _parse_returns(self) -> exp.ReturnsProperty:
    656            table = self._parse_id_var(any_token=False, tokens=self.RETURNS_TABLE_TOKENS)
    657            returns = super()._parse_returns()
    658            returns.set("table", table)
    659            return returns
    660
    661        def _parse_convert(
    662            self, strict: bool, safe: t.Optional[bool] = None
    663        ) -> t.Optional[exp.Expression]:
    664            this = self._parse_types()
    665            self._match(TokenType.COMMA)
    666            args = [this, *self._parse_csv(self._parse_assignment)]
    667            convert = exp.Convert.from_arg_list(args)
    668            convert.set("safe", safe)
    669            convert.set("strict", strict)
    670            return convert
    671
    672        def _parse_user_defined_function(
    673            self, kind: t.Optional[TokenType] = None
    674        ) -> t.Optional[exp.Expression]:
    675            this = super()._parse_user_defined_function(kind=kind)
    676
    677            if (
    678                kind == TokenType.FUNCTION
    679                or isinstance(this, exp.UserDefinedFunction)
    680                or self._match(TokenType.ALIAS, advance=False)
    681            ):
    682                return this
    683
    684            expressions = self._parse_csv(self._parse_function_parameter)
    685            return self.expression(exp.UserDefinedFunction, this=this, expressions=expressions)
    686
    687        def _parse_id_var(
    688            self,
    689            any_token: bool = True,
    690            tokens: t.Optional[t.Collection[TokenType]] = None,
    691        ) -> t.Optional[exp.Expression]:
    692            is_temporary = self._match(TokenType.HASH)
    693            is_global = is_temporary and self._match(TokenType.HASH)
    694
    695            this = super()._parse_id_var(any_token=any_token, tokens=tokens)
    696            if this:
    697                if is_global:
    698                    this.set("global", True)
    699                elif is_temporary:
    700                    this.set("temporary", True)
    701
    702            return this
    703
    704        def _parse_create(self) -> exp.Create | exp.Command:
    705            create = super()._parse_create()
    706
    707            if isinstance(create, exp.Create):
    708                table = create.this.this if isinstance(create.this, exp.Schema) else create.this
    709                if isinstance(table, exp.Table) and table.this.args.get("temporary"):
    710                    if not create.args.get("properties"):
    711                        create.set("properties", exp.Properties(expressions=[]))
    712
    713                    create.args["properties"].append("expressions", exp.TemporaryProperty())
    714
    715            return create
    716
    717        def _parse_if(self) -> t.Optional[exp.Expression]:
    718            index = self._index
    719
    720            if self._match_text_seq("OBJECT_ID"):
    721                self._parse_wrapped_csv(self._parse_string)
    722                if self._match_text_seq("IS", "NOT", "NULL") and self._match(TokenType.DROP):
    723                    return self._parse_drop(exists=True)
    724                self._retreat(index)
    725
    726            return super()._parse_if()
    727
    728        def _parse_unique(self) -> exp.UniqueColumnConstraint:
    729            if self._match_texts(("CLUSTERED", "NONCLUSTERED")):
    730                this = self.CONSTRAINT_PARSERS[self._prev.text.upper()](self)
    731            else:
    732                this = self._parse_schema(self._parse_id_var(any_token=False))
    733
    734            return self.expression(exp.UniqueColumnConstraint, this=this)
    735
    736        def _parse_partition(self) -> t.Optional[exp.Partition]:
    737            if not self._match_text_seq("WITH", "(", "PARTITIONS"):
    738                return None
    739
    740            def parse_range():
    741                low = self._parse_bitwise()
    742                high = self._parse_bitwise() if self._match_text_seq("TO") else None
    743
    744                return (
    745                    self.expression(exp.PartitionRange, this=low, expression=high) if high else low
    746                )
    747
    748            partition = self.expression(
    749                exp.Partition, expressions=self._parse_wrapped_csv(parse_range)
    750            )
    751
    752            self._match_r_paren()
    753
    754            return partition
    755
    756        def _parse_declare(self) -> exp.Declare | exp.Command:
    757            index = self._index
    758            expressions = self._try_parse(partial(self._parse_csv, self._parse_declareitem))
    759
    760            if not expressions or self._curr:
    761                self._retreat(index)
    762                return self._parse_as_command(self._prev)
    763
    764            return self.expression(exp.Declare, expressions=expressions)
    765
    766        def _parse_declareitem(self) -> t.Optional[exp.DeclareItem]:
    767            var = self._parse_id_var()
    768            if not var:
    769                return None
    770
    771            value = None
    772            self._match(TokenType.ALIAS)
    773            if self._match(TokenType.TABLE):
    774                data_type = self._parse_schema()
    775            else:
    776                data_type = self._parse_types()
    777                if self._match(TokenType.EQ):
    778                    value = self._parse_bitwise()
    779
    780            return self.expression(exp.DeclareItem, this=var, kind=data_type, default=value)

Parser consumes a list of tokens produced by the Tokenizer and produces a parsed syntax tree.

###### Arguments:

- error_level: The desired error level. Default: ErrorLevel.IMMEDIATE
- error_message_context: The amount of context to capture from a query string when displaying the error message (in number of characters). Default: 100
- max_errors: Maximum number of error messages to include in a raised ParseError. This is only relevant if error_level is ErrorLevel.RAISE. Default: 3

SET_REQUIRES_ASSIGNMENT_DELIMITER = False
LOG_DEFAULTS_TO_LN = True
ALTER_TABLE_ADD_REQUIRED_FOR_EACH_COLUMN = False
STRING_ALIASES = True
NO_PAREN_IF_COMMANDS = False
QUERY_MODIFIER_PARSERS = {<TokenType.MATCH_RECOGNIZE: 'MATCH_RECOGNIZE'>: <function Parser.<lambda>>, <TokenType.PREWHERE: 'PREWHERE'>: <function Parser.<lambda>>, <TokenType.WHERE: 'WHERE'>: <function Parser.<lambda>>, <TokenType.GROUP_BY: 'GROUP_BY'>: <function Parser.<lambda>>, <TokenType.HAVING: 'HAVING'>: <function Parser.<lambda>>, <TokenType.QUALIFY: 'QUALIFY'>: <function Parser.<lambda>>, <TokenType.WINDOW: 'WINDOW'>: <function Parser.<lambda>>, <TokenType.ORDER_BY: 'ORDER_BY'>: <function Parser.<lambda>>, <TokenType.LIMIT: 'LIMIT'>: <function Parser.<lambda>>, <TokenType.FETCH: 'FETCH'>: <function Parser.<lambda>>, <TokenType.OFFSET: 'OFFSET'>: <function Parser.<lambda>>, <TokenType.FOR: 'FOR'>: <function Parser.<lambda>>, <TokenType.LOCK: 'LOCK'>: <function Parser.<lambda>>, <TokenType.TABLE_SAMPLE: 'TABLE_SAMPLE'>: <function Parser.<lambda>>, <TokenType.USING: 'USING'>: <function Parser.<lambda>>, <TokenType.CLUSTER_BY: 'CLUSTER_BY'>: <function Parser.<lambda>>, <TokenType.DISTRIBUTE_BY: 'DISTRIBUTE_BY'>: <function Parser.<lambda>>, <TokenType.SORT_BY: 'SORT_BY'>: <function Parser.<lambda>>, <TokenType.CONNECT_BY: 'CONNECT_BY'>: <function Parser.<lambda>>, <TokenType.START_WITH: 'START_WITH'>: <function Parser.<lambda>>, <TokenType.OPTION: 'OPTION'>: <function TSQL.Parser.<lambda>>}
FUNCTIONS = {'ABS': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Abs'>>, 'ADD_MONTHS': <bound method Func.from_arg_list of <class 'sqlglot.expressions.AddMonths'>>, 'ANONYMOUS_AGG_FUNC': <bound method Func.from_arg_list of <class 'sqlglot.expressions.AnonymousAggFunc'>>, 'ANY_VALUE': <bound method Func.from_arg_list of <class 'sqlglot.expressions.AnyValue'>>, 'APPROX_DISTINCT': <bound method Func.from_arg_list of <class 'sqlglot.expressions.ApproxDistinct'>>, 'APPROX_COUNT_DISTINCT': <bound method Func.from_arg_list of <class 'sqlglot.expressions.ApproxDistinct'>>, 'APPROX_QUANTILE': <bound method Func.from_arg_list of <class 'sqlglot.expressions.ApproxQuantile'>>, 'APPROX_TOP_K': <bound method Func.from_arg_list of <class 'sqlglot.expressions.ApproxTopK'>>, 'ARG_MAX': <bound method Func.from_arg_list of <class 'sqlglot.expressions.ArgMax'>>, 'ARGMAX': <bound method Func.from_arg_list of <class 'sqlglot.expressions.ArgMax'>>, 'MAX_BY': <bound method Func.from_arg_list of <class 'sqlglot.expressions.ArgMax'>>, 'ARG_MIN': <bound method Func.from_arg_list of <class 'sqlglot.expressions.ArgMin'>>, 'ARGMIN': <bound method Func.from_arg_list of <class 'sqlglot.expressions.ArgMin'>>, 'MIN_BY': <bound method Func.from_arg_list of <class 'sqlglot.expressions.ArgMin'>>, 'ARRAY': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Array'>>, 'ARRAY_AGG': <bound method Func.from_arg_list of <class 'sqlglot.expressions.ArrayAgg'>>, 'ARRAY_ALL': <bound method Func.from_arg_list of <class 'sqlglot.expressions.ArrayAll'>>, 'ARRAY_ANY': <bound method Func.from_arg_list of <class 'sqlglot.expressions.ArrayAny'>>, 'ARRAY_CONCAT': <bound method Func.from_arg_list of <class 'sqlglot.expressions.ArrayConcat'>>, 'ARRAY_CAT': <bound method Func.from_arg_list of <class 'sqlglot.expressions.ArrayConcat'>>, 'ARRAY_CONSTRUCT_COMPACT': <bound method Func.from_arg_list of <class 'sqlglot.expressions.ArrayConstructCompact'>>, 'ARRAY_CONTAINS': <bound method Func.from_arg_list of <class 'sqlglot.expressions.ArrayContains'>>, 'ARRAY_HAS': <bound method Func.from_arg_list of <class 'sqlglot.expressions.ArrayContains'>>, 'ARRAY_CONTAINS_ALL': <bound method Func.from_arg_list of <class 'sqlglot.expressions.ArrayContainsAll'>>, 'ARRAY_HAS_ALL': <bound method Func.from_arg_list of <class 'sqlglot.expressions.ArrayContainsAll'>>, 'FILTER': <bound method Func.from_arg_list of <class 'sqlglot.expressions.ArrayFilter'>>, 'ARRAY_FILTER': <bound method Func.from_arg_list of <class 'sqlglot.expressions.ArrayFilter'>>, 'ARRAY_OVERLAPS': <bound method Func.from_arg_list of <class 'sqlglot.expressions.ArrayOverlaps'>>, 'ARRAY_SIZE': <bound method Func.from_arg_list of <class 'sqlglot.expressions.ArraySize'>>, 'ARRAY_LENGTH': <bound method Func.from_arg_list of <class 'sqlglot.expressions.ArraySize'>>, 'ARRAY_SORT': <bound method Func.from_arg_list of <class 'sqlglot.expressions.ArraySort'>>, 'ARRAY_SUM': <bound method Func.from_arg_list of <class 'sqlglot.expressions.ArraySum'>>, 'ARRAY_TO_STRING': <bound method Func.from_arg_list of <class 'sqlglot.expressions.ArrayToString'>>, 'ARRAY_JOIN': <bound method Func.from_arg_list of <class 'sqlglot.expressions.ArrayToString'>>, 'ARRAY_UNION_AGG': <bound method Func.from_arg_list of <class 'sqlglot.expressions.ArrayUnionAgg'>>, 'ARRAY_UNIQUE_AGG': <bound method Func.from_arg_list of <class 'sqlglot.expressions.ArrayUniqueAgg'>>, 'AVG': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Avg'>>, 'CASE': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Case'>>, 'CAST': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Cast'>>, 'CAST_TO_STR_TYPE': <bound method Func.from_arg_list of <class 'sqlglot.expressions.CastToStrType'>>, 'CBRT': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Cbrt'>>, 'CEIL': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Ceil'>>, 'CEILING': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Ceil'>>, 'CHR': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Chr'>>, 'CHAR': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Chr'>>, 'COALESCE': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Coalesce'>>, 'IFNULL': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Coalesce'>>, 'NVL': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Coalesce'>>, 'COLLATE': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Collate'>>, 'COMBINED_AGG_FUNC': <bound method Func.from_arg_list of <class 'sqlglot.expressions.CombinedAggFunc'>>, 'COMBINED_PARAMETERIZED_AGG': <bound method Func.from_arg_list of <class 'sqlglot.expressions.CombinedParameterizedAgg'>>, 'CONCAT': <function Parser.<lambda>>, 'CONCAT_WS': <function Parser.<lambda>>, 'CONNECT_BY_ROOT': <bound method Func.from_arg_list of <class 'sqlglot.expressions.ConnectByRoot'>>, 'CONVERT': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Convert'>>, 'CORR': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Corr'>>, 'COUNT': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Count'>>, 'COUNT_IF': <bound method Func.from_arg_list of <class 'sqlglot.expressions.CountIf'>>, 'COUNTIF': <bound method Func.from_arg_list of <class 'sqlglot.expressions.CountIf'>>, 'COVAR_POP': <bound method Func.from_arg_list of <class 'sqlglot.expressions.CovarPop'>>, 'COVAR_SAMP': <bound method Func.from_arg_list of <class 'sqlglot.expressions.CovarSamp'>>, 'CURRENT_DATE': <bound method Func.from_arg_list of <class 'sqlglot.expressions.CurrentDate'>>, 'CURRENT_DATETIME': <bound method Func.from_arg_list of <class 'sqlglot.expressions.CurrentDatetime'>>, 'CURRENT_TIME': <bound method Func.from_arg_list of <class 'sqlglot.expressions.CurrentTime'>>, 'CURRENT_TIMESTAMP': <bound method Func.from_arg_list of <class 'sqlglot.expressions.CurrentTimestamp'>>, 'CURRENT_USER': <bound method Func.from_arg_list of <class 'sqlglot.expressions.CurrentUser'>>, 'DATE': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Date'>>, 'DATE_ADD': <bound method Func.from_arg_list of <class 'sqlglot.expressions.DateAdd'>>, 'DATEDIFF': <function _build_date_delta.<locals>._builder>, 'DATE_DIFF': <bound method Func.from_arg_list of <class 'sqlglot.expressions.DateDiff'>>, 'DATE_FROM_PARTS': <bound method Func.from_arg_list of <class 'sqlglot.expressions.DateFromParts'>>, 'DATEFROMPARTS': <bound method Func.from_arg_list of <class 'sqlglot.expressions.DateFromParts'>>, 'DATE_STR_TO_DATE': <bound method Func.from_arg_list of <class 'sqlglot.expressions.DateStrToDate'>>, 'DATE_SUB': <bound method Func.from_arg_list of <class 'sqlglot.expressions.DateSub'>>, 'DATE_TO_DATE_STR': <function Parser.<lambda>>, 'DATE_TO_DI': <bound method Func.from_arg_list of <class 'sqlglot.expressions.DateToDi'>>, 'DATE_TRUNC': <bound method Func.from_arg_list of <class 'sqlglot.expressions.DateTrunc'>>, 'DATETIME': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Datetime'>>, 'DATETIME_ADD': <bound method Func.from_arg_list of <class 'sqlglot.expressions.DatetimeAdd'>>, 'DATETIME_DIFF': <bound method Func.from_arg_list of <class 'sqlglot.expressions.DatetimeDiff'>>, 'DATETIME_SUB': <bound method Func.from_arg_list of <class 'sqlglot.expressions.DatetimeSub'>>, 'DATETIME_TRUNC': <bound method Func.from_arg_list of <class 'sqlglot.expressions.DatetimeTrunc'>>, 'DAY': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Day'>>, 'DAY_OF_MONTH': <bound method Func.from_arg_list of <class 'sqlglot.expressions.DayOfMonth'>>, 'DAYOFMONTH': <bound method Func.from_arg_list of <class 'sqlglot.expressions.DayOfMonth'>>, 'DAY_OF_WEEK': <bound method Func.from_arg_list of <class 'sqlglot.expressions.DayOfWeek'>>, 'DAYOFWEEK': <bound method Func.from_arg_list of <class 'sqlglot.expressions.DayOfWeek'>>, 'DAY_OF_YEAR': <bound method Func.from_arg_list of <class 'sqlglot.expressions.DayOfYear'>>, 'DAYOFYEAR': <bound method Func.from_arg_list of <class 'sqlglot.expressions.DayOfYear'>>, 'DECODE': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Decode'>>, 'DI_TO_DATE': <bound method Func.from_arg_list of <class 'sqlglot.expressions.DiToDate'>>, 'ENCODE': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Encode'>>, 'EXP': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Exp'>>, 'EXPLODE': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Explode'>>, 'EXPLODE_OUTER': <bound method Func.from_arg_list of <class 'sqlglot.expressions.ExplodeOuter'>>, 'EXTRACT': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Extract'>>, 'FIRST': <bound method Func.from_arg_list of <class 'sqlglot.expressions.First'>>, 'FIRST_VALUE': <bound method Func.from_arg_list of <class 'sqlglot.expressions.FirstValue'>>, 'FLATTEN': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Flatten'>>, 'FLOOR': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Floor'>>, 'FROM_BASE': <bound method Func.from_arg_list of <class 'sqlglot.expressions.FromBase'>>, 'FROM_BASE64': <bound method Func.from_arg_list of <class 'sqlglot.expressions.FromBase64'>>, 'GAP_FILL': <bound method Func.from_arg_list of <class 'sqlglot.expressions.GapFill'>>, 'GENERATE_DATE_ARRAY': <bound method Func.from_arg_list of <class 'sqlglot.expressions.GenerateDateArray'>>, 'GENERATE_SERIES': <bound method Func.from_arg_list of <class 'sqlglot.expressions.GenerateSeries'>>, 'GREATEST': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Greatest'>>, 'GROUP_CONCAT': <bound method Func.from_arg_list of <class 'sqlglot.expressions.GroupConcat'>>, 'HEX': <function build_hex>, 'HLL': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Hll'>>, 'IF': <bound method Func.from_arg_list of <class 'sqlglot.expressions.If'>>, 'IIF': <bound method Func.from_arg_list of <class 'sqlglot.expressions.If'>>, 'INITCAP': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Initcap'>>, 'IS_INF': <bound method Func.from_arg_list of <class 'sqlglot.expressions.IsInf'>>, 'ISINF': <bound method Func.from_arg_list of <class 'sqlglot.expressions.IsInf'>>, 'IS_NAN': <bound method Func.from_arg_list of <class 'sqlglot.expressions.IsNan'>>, 'ISNAN': <bound method Func.from_arg_list of <class 'sqlglot.expressions.IsNan'>>, 'J_S_O_N_ARRAY': <bound method Func.from_arg_list of <class 'sqlglot.expressions.JSONArray'>>, 'J_S_O_N_ARRAY_AGG': <bound method Func.from_arg_list of <class 'sqlglot.expressions.JSONArrayAgg'>>, 'JSON_ARRAY_CONTAINS': <bound method Func.from_arg_list of <class 'sqlglot.expressions.JSONArrayContains'>>, 'JSONB_CONTAINS': <bound method Func.from_arg_list of <class 'sqlglot.expressions.JSONBContains'>>, 'JSONB_EXTRACT': <bound method Func.from_arg_list of <class 'sqlglot.expressions.JSONBExtract'>>, 'JSONB_EXTRACT_SCALAR': <bound method Func.from_arg_list of <class 'sqlglot.expressions.JSONBExtractScalar'>>, 'JSON_EXTRACT': <function build_extract_json_with_path.<locals>._builder>, 'JSON_EXTRACT_SCALAR': <function build_extract_json_with_path.<locals>._builder>, 'JSON_FORMAT': <bound method Func.from_arg_list of <class 'sqlglot.expressions.JSONFormat'>>, 'J_S_O_N_OBJECT': <bound method Func.from_arg_list of <class 'sqlglot.expressions.JSONObject'>>, 'J_S_O_N_OBJECT_AGG': <bound method Func.from_arg_list of <class 'sqlglot.expressions.JSONObjectAgg'>>, 'J_S_O_N_TABLE': <bound method Func.from_arg_list of <class 'sqlglot.expressions.JSONTable'>>, 'LAG': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Lag'>>, 'LAST': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Last'>>, 'LAST_DAY': <bound method Func.from_arg_list of <class 'sqlglot.expressions.LastDay'>>, 'LAST_DAY_OF_MONTH': <bound method Func.from_arg_list of <class 'sqlglot.expressions.LastDay'>>, 'LAST_VALUE': <bound method Func.from_arg_list of <class 'sqlglot.expressions.LastValue'>>, 'LEAD': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Lead'>>, 'LEAST': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Least'>>, 'LEFT': <function _build_with_arg_as_text.<locals>._parse>, 'LENGTH': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Length'>>, 'LEN': <function _build_with_arg_as_text.<locals>._parse>, 'LEVENSHTEIN': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Levenshtein'>>, 'LIST': <bound method Func.from_arg_list of <class 'sqlglot.expressions.List'>>, 'LN': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Ln'>>, 'LOG': <function build_logarithm>, 'LOGICAL_AND': <bound method Func.from_arg_list of <class 'sqlglot.expressions.LogicalAnd'>>, 'BOOL_AND': <bound method Func.from_arg_list of <class 'sqlglot.expressions.LogicalAnd'>>, 'BOOLAND_AGG': <bound method Func.from_arg_list of <class 'sqlglot.expressions.LogicalAnd'>>, 'LOGICAL_OR': <bound method Func.from_arg_list of <class 'sqlglot.expressions.LogicalOr'>>, 'BOOL_OR': <bound method Func.from_arg_list of <class 'sqlglot.expressions.LogicalOr'>>, 'BOOLOR_AGG': <bound method Func.from_arg_list of <class 'sqlglot.expressions.LogicalOr'>>, 'LOWER': <function build_lower>, 'LCASE': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Lower'>>, 'LOWER_HEX': <bound method Func.from_arg_list of <class 'sqlglot.expressions.LowerHex'>>, 'MD5': <bound method Func.from_arg_list of <class 'sqlglot.expressions.MD5'>>, 'MD5_DIGEST': <bound method Func.from_arg_list of <class 'sqlglot.expressions.MD5Digest'>>, 'MAP': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Map'>>, 'MAP_FROM_ENTRIES': <bound method Func.from_arg_list of <class 'sqlglot.expressions.MapFromEntries'>>, 'MATCH_AGAINST': <bound method Func.from_arg_list of <class 'sqlglot.expressions.MatchAgainst'>>, 'MAX': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Max'>>, 'MIN': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Min'>>, 'MONTH': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Month'>>, 'MONTHS_BETWEEN': <bound method Func.from_arg_list of <class 'sqlglot.expressions.MonthsBetween'>>, 'NEXT_VALUE_FOR': <bound method Func.from_arg_list of <class 'sqlglot.expressions.NextValueFor'>>, 'NTH_VALUE': <bound method Func.from_arg_list of <class 'sqlglot.expressions.NthValue'>>, 'NULLIF': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Nullif'>>, 'NUMBER_TO_STR': <bound method Func.from_arg_list of <class 'sqlglot.expressions.NumberToStr'>>, 'NVL2': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Nvl2'>>, 'OBJECT_INSERT': <bound method Func.from_arg_list of <class 'sqlglot.expressions.ObjectInsert'>>, 'OPEN_J_S_O_N': <bound method Func.from_arg_list of <class 'sqlglot.expressions.OpenJSON'>>, 'PAD': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Pad'>>, 'PARAMETERIZED_AGG': <bound method Func.from_arg_list of <class 'sqlglot.expressions.ParameterizedAgg'>>, 'PARSE_JSON': <bound method Func.from_arg_list of <class 'sqlglot.expressions.ParseJSON'>>, 'JSON_PARSE': <bound method Func.from_arg_list of <class 'sqlglot.expressions.ParseJSON'>>, 'PERCENTILE_CONT': <bound method Func.from_arg_list of <class 'sqlglot.expressions.PercentileCont'>>, 'PERCENTILE_DISC': <bound method Func.from_arg_list of <class 'sqlglot.expressions.PercentileDisc'>>, 'POSEXPLODE': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Posexplode'>>, 'POSEXPLODE_OUTER': <bound method Func.from_arg_list of <class 'sqlglot.expressions.PosexplodeOuter'>>, 'POWER': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Pow'>>, 'POW': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Pow'>>, 'PREDICT': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Predict'>>, 'QUANTILE': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Quantile'>>, 'QUARTER': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Quarter'>>, 'RAND': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Rand'>>, 'RANDOM': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Rand'>>, 'RANDN': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Randn'>>, 'RANGE_N': <bound method Func.from_arg_list of <class 'sqlglot.expressions.RangeN'>>, 'READ_CSV': <bound method Func.from_arg_list of <class 'sqlglot.expressions.ReadCSV'>>, 'REDUCE': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Reduce'>>, 'REGEXP_EXTRACT': <bound method Func.from_arg_list of <class 'sqlglot.expressions.RegexpExtract'>>, 'REGEXP_I_LIKE': <bound method Func.from_arg_list of <class 'sqlglot.expressions.RegexpILike'>>, 'REGEXP_LIKE': <bound method Func.from_arg_list of <class 'sqlglot.expressions.RegexpLike'>>, 'REGEXP_REPLACE': <bound method Func.from_arg_list of <class 'sqlglot.expressions.RegexpReplace'>>, 'REGEXP_SPLIT': <bound method Func.from_arg_list of <class 'sqlglot.expressions.RegexpSplit'>>, 'REPEAT': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Repeat'>>, 'RIGHT': <function _build_with_arg_as_text.<locals>._parse>, 'ROUND': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Round'>>, 'ROW_NUMBER': <bound method Func.from_arg_list of <class 'sqlglot.expressions.RowNumber'>>, 'SHA': <bound method Func.from_arg_list of <class 'sqlglot.expressions.SHA'>>, 'SHA1': <bound method Func.from_arg_list of <class 'sqlglot.expressions.SHA'>>, 'SHA2': <bound method Func.from_arg_list of <class 'sqlglot.expressions.SHA2'>>, 'SAFE_DIVIDE': <bound method Func.from_arg_list of <class 'sqlglot.expressions.SafeDivide'>>, 'SIGN': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Sign'>>, 'SIGNUM': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Sign'>>, 'SORT_ARRAY': <bound method Func.from_arg_list of <class 'sqlglot.expressions.SortArray'>>, 'SPLIT': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Split'>>, 'SQRT': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Sqrt'>>, 'STANDARD_HASH': <bound method Func.from_arg_list of <class 'sqlglot.expressions.StandardHash'>>, 'STAR_MAP': <bound method Func.from_arg_list of <class 'sqlglot.expressions.StarMap'>>, 'STARTS_WITH': <bound method Func.from_arg_list of <class 'sqlglot.expressions.StartsWith'>>, 'STARTSWITH': <bound method Func.from_arg_list of <class 'sqlglot.expressions.StartsWith'>>, 'STDDEV': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Stddev'>>, 'STDEV': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Stddev'>>, 'STDDEV_POP': <bound method Func.from_arg_list of <class 'sqlglot.expressions.StddevPop'>>, 'STDDEV_SAMP': <bound method Func.from_arg_list of <class 'sqlglot.expressions.StddevSamp'>>, 'STR_POSITION': <bound method Func.from_arg_list of <class 'sqlglot.expressions.StrPosition'>>, 'STR_TO_DATE': <bound method Func.from_arg_list of <class 'sqlglot.expressions.StrToDate'>>, 'STR_TO_MAP': <bound method Func.from_arg_list of <class 'sqlglot.expressions.StrToMap'>>, 'STR_TO_TIME': <bound method Func.from_arg_list of <class 'sqlglot.expressions.StrToTime'>>, 'STR_TO_UNIX': <bound method Func.from_arg_list of <class 'sqlglot.expressions.StrToUnix'>>, 'STRING_TO_ARRAY': <bound method Func.from_arg_list of <class 'sqlglot.expressions.StringToArray'>>, 'SPLIT_BY_STRING': <bound method Func.from_arg_list of <class 'sqlglot.expressions.StringToArray'>>, 'STRUCT': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Struct'>>, 'STRUCT_EXTRACT': <bound method Func.from_arg_list of <class 'sqlglot.expressions.StructExtract'>>, 'STUFF': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Stuff'>>, 'INSERT': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Stuff'>>, 'SUBSTRING': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Substring'>>, 'SUM': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Sum'>>, 'TIME': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Time'>>, 'TIME_ADD': <bound method Func.from_arg_list of <class 'sqlglot.expressions.TimeAdd'>>, 'TIME_DIFF': <bound method Func.from_arg_list of <class 'sqlglot.expressions.TimeDiff'>>, 'TIME_FROM_PARTS': <bound method Func.from_arg_list of <class 'sqlglot.expressions.TimeFromParts'>>, 'TIMEFROMPARTS': <function _build_timefromparts>, 'TIME_STR_TO_DATE': <bound method Func.from_arg_list of <class 'sqlglot.expressions.TimeStrToDate'>>, 'TIME_STR_TO_TIME': <bound method Func.from_arg_list of <class 'sqlglot.expressions.TimeStrToTime'>>, 'TIME_STR_TO_UNIX': <bound method Func.from_arg_list of <class 'sqlglot.expressions.TimeStrToUnix'>>, 'TIME_SUB': <bound method Func.from_arg_list of <class 'sqlglot.expressions.TimeSub'>>, 'TIME_TO_STR': <bound method Func.from_arg_list of <class 'sqlglot.expressions.TimeToStr'>>, 'TIME_TO_TIME_STR': <function Parser.<lambda>>, 'TIME_TO_UNIX': <bound method Func.from_arg_list of <class 'sqlglot.expressions.TimeToUnix'>>, 'TIME_TRUNC': <bound method Func.from_arg_list of <class 'sqlglot.expressions.TimeTrunc'>>, 'TIMESTAMP': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Timestamp'>>, 'TIMESTAMP_ADD': <bound method Func.from_arg_list of <class 'sqlglot.expressions.TimestampAdd'>>, 'TIMESTAMPDIFF': <bound method Func.from_arg_list of <class 'sqlglot.expressions.TimestampDiff'>>, 'TIMESTAMP_DIFF': <bound method Func.from_arg_list of <class 'sqlglot.expressions.TimestampDiff'>>, 'TIMESTAMP_FROM_PARTS': <bound method Func.from_arg_list of <class 'sqlglot.expressions.TimestampFromParts'>>, 'TIMESTAMPFROMPARTS': <bound method Func.from_arg_list of <class 'sqlglot.expressions.TimestampFromParts'>>, 'TIMESTAMP_SUB': <bound method Func.from_arg_list of <class 'sqlglot.expressions.TimestampSub'>>, 'TIMESTAMP_TRUNC': <bound method Func.from_arg_list of <class 'sqlglot.expressions.TimestampTrunc'>>, 'TO_ARRAY': <bound method Func.from_arg_list of <class 'sqlglot.expressions.ToArray'>>, 'TO_BASE64': <bound method Func.from_arg_list of <class 'sqlglot.expressions.ToBase64'>>, 'TO_CHAR': <bound method Func.from_arg_list of <class 'sqlglot.expressions.ToChar'>>, 'TO_DAYS': <bound method Func.from_arg_list of <class 'sqlglot.expressions.ToDays'>>, 'TO_MAP': <bound method Func.from_arg_list of <class 'sqlglot.expressions.ToMap'>>, 'TO_NUMBER': <bound method Func.from_arg_list of <class 'sqlglot.expressions.ToNumber'>>, 'TRANSFORM': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Transform'>>, 'TRIM': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Trim'>>, 'TRY': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Try'>>, 'TRY_CAST': <bound method Func.from_arg_list of <class 'sqlglot.expressions.TryCast'>>, 'TS_OR_DI_TO_DI': <bound method Func.from_arg_list of <class 'sqlglot.expressions.TsOrDiToDi'>>, 'TS_OR_DS_ADD': <bound method Func.from_arg_list of <class 'sqlglot.expressions.TsOrDsAdd'>>, 'TS_OR_DS_DIFF': <bound method Func.from_arg_list of <class 'sqlglot.expressions.TsOrDsDiff'>>, 'TS_OR_DS_TO_DATE': <bound method Func.from_arg_list of <class 'sqlglot.expressions.TsOrDsToDate'>>, 'TS_OR_DS_TO_DATE_STR': <function Parser.<lambda>>, 'TS_OR_DS_TO_TIME': <bound method Func.from_arg_list of <class 'sqlglot.expressions.TsOrDsToTime'>>, 'TS_OR_DS_TO_TIMESTAMP': <bound method Func.from_arg_list of <class 'sqlglot.expressions.TsOrDsToTimestamp'>>, 'UNHEX': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Unhex'>>, 'UNIX_DATE': <bound method Func.from_arg_list of <class 'sqlglot.expressions.UnixDate'>>, 'UNIX_TO_STR': <bound method Func.from_arg_list of <class 'sqlglot.expressions.UnixToStr'>>, 'UNIX_TO_TIME': <bound method Func.from_arg_list of <class 'sqlglot.expressions.UnixToTime'>>, 'UNIX_TO_TIME_STR': <bound method Func.from_arg_list of <class 'sqlglot.expressions.UnixToTimeStr'>>, 'UNNEST': <function Parser.<lambda>>, 'UPPER': <function build_upper>, 'UCASE': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Upper'>>, 'VAR_MAP': <function build_var_map>, 'VARIANCE': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Variance'>>, 'VARIANCE_SAMP': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Variance'>>, 'VAR_SAMP': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Variance'>>, 'VARIANCE_POP': <bound method Func.from_arg_list of <class 'sqlglot.expressions.VariancePop'>>, 'VAR_POP': <bound method Func.from_arg_list of <class 'sqlglot.expressions.VariancePop'>>, 'WEEK': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Week'>>, 'WEEK_OF_YEAR': <bound method Func.from_arg_list of <class 'sqlglot.expressions.WeekOfYear'>>, 'WEEKOFYEAR': <bound method Func.from_arg_list of <class 'sqlglot.expressions.WeekOfYear'>>, 'WHEN': <bound method Func.from_arg_list of <class 'sqlglot.expressions.When'>>, 'X_M_L_TABLE': <bound method Func.from_arg_list of <class 'sqlglot.expressions.XMLTable'>>, 'XOR': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Xor'>>, 'YEAR': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Year'>>, 'GLOB': <function Parser.<lambda>>, 'JSON_EXTRACT_PATH_TEXT': <function build_extract_json_with_path.<locals>._builder>, 'LIKE': <function build_like>, 'LOG2': <function Parser.<lambda>>, 'LOG10': <function Parser.<lambda>>, 'LPAD': <function Parser.<lambda>>, 'LEFTPAD': <function Parser.<lambda>>, 'MOD': <function build_mod>, 'RPAD': <function Parser.<lambda>>, 'RIGHTPAD': <function Parser.<lambda>>, 'SCOPE_RESOLUTION': <function Parser.<lambda>>, 'TO_HEX': <function build_hex>, 'CHARINDEX': <function TSQL.Parser.<lambda>>, 'DATEADD': <function build_date_delta.<locals>._builder>, 'DATENAME': <function _build_formatted_time.<locals>._builder>, 'DATEPART': <function _build_formatted_time.<locals>._builder>, 'DATETIMEFROMPARTS': <function _build_datetimefromparts>, 'EOMONTH': <function _build_eomonth>, 'FORMAT': <function _build_format>, 'GETDATE': <bound method Func.from_arg_list of <class 'sqlglot.expressions.CurrentTimestamp'>>, 'HASHBYTES': <function _build_hashbytes>, 'ISNULL': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Coalesce'>>, 'JSON_QUERY': <function _build_json_query>, 'JSON_VALUE': <function build_extract_json_with_path.<locals>._builder>, 'REPLICATE': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Repeat'>>, 'SQUARE': <function TSQL.Parser.<lambda>>, 'SYSDATETIME': <bound method Func.from_arg_list of <class 'sqlglot.expressions.CurrentTimestamp'>>, 'SUSER_NAME': <bound method Func.from_arg_list of <class 'sqlglot.expressions.CurrentUser'>>, 'SUSER_SNAME': <bound method Func.from_arg_list of <class 'sqlglot.expressions.CurrentUser'>>, 'SYSTEM_USER': <bound method Func.from_arg_list of <class 'sqlglot.expressions.CurrentUser'>>}
JOIN_HINTS = {'REMOTE', 'MERGE', 'HASH', 'LOOP'}
RETURNS_TABLE_TOKENS = {<TokenType.WINDOW: 'WINDOW'>, <TokenType.COPY: 'COPY'>, <TokenType.DICTIONARY: 'DICTIONARY'>, <TokenType.IS: 'IS'>, <TokenType.VIEW: 'VIEW'>, <TokenType.REFERENCES: 'REFERENCES'>, <TokenType.TOP: 'TOP'>, <TokenType.DIV: 'DIV'>, <TokenType.ROW: 'ROW'>, <TokenType.SEMI: 'SEMI'>, <TokenType.OPERATOR: 'OPERATOR'>, <TokenType.SOME: 'SOME'>, <TokenType.LEFT: 'LEFT'>, <TokenType.CURRENT_DATETIME: 'CURRENT_DATETIME'>, <TokenType.SETTINGS: 'SETTINGS'>, <TokenType.FOREIGN_KEY: 'FOREIGN_KEY'>, <TokenType.FALSE: 'FALSE'>, <TokenType.RIGHT: 'RIGHT'>, <TokenType.ORDINALITY: 'ORDINALITY'>, <TokenType.NEXT: 'NEXT'>, <TokenType.REPLACE: 'REPLACE'>, <TokenType.SHOW: 'SHOW'>, <TokenType.USE: 'USE'>, <TokenType.ESCAPE: 'ESCAPE'>, <TokenType.COMMAND: 'COMMAND'>, <TokenType.CURRENT_USER: 'CURRENT_USER'>, <TokenType.MERGE: 'MERGE'>, <TokenType.ASOF: 'ASOF'>, <TokenType.UNNEST: 'UNNEST'>, <TokenType.UPDATE: 'UPDATE'>, <TokenType.PROCEDURE: 'PROCEDURE'>, <TokenType.ALL: 'ALL'>, <TokenType.APPLY: 'APPLY'>, <TokenType.OVERWRITE: 'OVERWRITE'>, <TokenType.NATURAL: 'NATURAL'>, <TokenType.MODEL: 'MODEL'>, <TokenType.PERCENT: 'PERCENT'>, <TokenType.UNIQUE: 'UNIQUE'>, <TokenType.ROLLUP: 'ROLLUP'>, <TokenType.FIRST: 'FIRST'>, <TokenType.STREAMLIT: 'STREAMLIT'>, <TokenType.RANGE: 'RANGE'>, <TokenType.FORMAT: 'FORMAT'>, <TokenType.KEEP: 'KEEP'>, <TokenType.LOAD: 'LOAD'>, <TokenType.PIVOT: 'PIVOT'>, <TokenType.ROWS: 'ROWS'>, <TokenType.SCHEMA: 'SCHEMA'>, <TokenType.EXISTS: 'EXISTS'>, <TokenType.COLLATE: 'COLLATE'>, <TokenType.VAR: 'VAR'>, <TokenType.CUBE: 'CUBE'>, <TokenType.SET: 'SET'>, <TokenType.IDENTIFIER: 'IDENTIFIER'>, <TokenType.EXECUTE: 'EXECUTE'>, <TokenType.ASC: 'ASC'>, <TokenType.WAREHOUSE: 'WAREHOUSE'>, <TokenType.COMMIT: 'COMMIT'>, <TokenType.DELETE: 'DELETE'>, <TokenType.DEFAULT: 'DEFAULT'>, <TokenType.INDEX: 'INDEX'>, <TokenType.COMMENT: 'COMMENT'>, <TokenType.BEGIN: 'BEGIN'>, <TokenType.UNPIVOT: 'UNPIVOT'>, <TokenType.DESCRIBE: 'DESCRIBE'>, <TokenType.CURRENT_TIMESTAMP: 'CURRENT_TIMESTAMP'>, <TokenType.REFRESH: 'REFRESH'>, <TokenType.END: 'END'>, <TokenType.CACHE: 'CACHE'>, <TokenType.ANY: 'ANY'>, <TokenType.ANTI: 'ANTI'>, <TokenType.FUNCTION: 'FUNCTION'>, <TokenType.DESC: 'DESC'>, <TokenType.FULL: 'FULL'>, <TokenType.ISNULL: 'ISNULL'>, <TokenType.FINAL: 'FINAL'>, <TokenType.TRUNCATE: 'TRUNCATE'>, <TokenType.OVERLAPS: 'OVERLAPS'>, <TokenType.OFFSET: 'OFFSET'>, <TokenType.FILTER: 'FILTER'>, <TokenType.PRAGMA: 'PRAGMA'>, <TokenType.TRUE: 'TRUE'>, <TokenType.TEMPORARY: 'TEMPORARY'>, <TokenType.CASE: 'CASE'>, <TokenType.AUTO_INCREMENT: 'AUTO_INCREMENT'>, <TokenType.KILL: 'KILL'>, <TokenType.VOLATILE: 'VOLATILE'>, <TokenType.PARTITION: 'PARTITION'>, <TokenType.CURRENT_TIME: 'CURRENT_TIME'>, <TokenType.TAG: 'TAG'>, <TokenType.CURRENT_DATE: 'CURRENT_DATE'>, <TokenType.COLUMN: 'COLUMN'>, <TokenType.DATABASE: 'DATABASE'>, <TokenType.CONSTRAINT: 'CONSTRAINT'>, <TokenType.RECURSIVE: 'RECURSIVE'>, <TokenType.STORAGE_INTEGRATION: 'STORAGE_INTEGRATION'>, <TokenType.SEQUENCE: 'SEQUENCE'>}
STATEMENT_PARSERS = {<TokenType.ALTER: 'ALTER'>: <function Parser.<lambda>>, <TokenType.BEGIN: 'BEGIN'>: <function Parser.<lambda>>, <TokenType.CACHE: 'CACHE'>: <function Parser.<lambda>>, <TokenType.COMMENT: 'COMMENT'>: <function Parser.<lambda>>, <TokenType.COMMIT: 'COMMIT'>: <function Parser.<lambda>>, <TokenType.COPY: 'COPY'>: <function Parser.<lambda>>, <TokenType.CREATE: 'CREATE'>: <function Parser.<lambda>>, <TokenType.DELETE: 'DELETE'>: <function Parser.<lambda>>, <TokenType.DESC: 'DESC'>: <function Parser.<lambda>>, <TokenType.DESCRIBE: 'DESCRIBE'>: <function Parser.<lambda>>, <TokenType.DROP: 'DROP'>: <function Parser.<lambda>>, <TokenType.INSERT: 'INSERT'>: <function Parser.<lambda>>, <TokenType.KILL: 'KILL'>: <function Parser.<lambda>>, <TokenType.LOAD: 'LOAD'>: <function Parser.<lambda>>, <TokenType.MERGE: 'MERGE'>: <function Parser.<lambda>>, <TokenType.PIVOT: 'PIVOT'>: <function Parser.<lambda>>, <TokenType.PRAGMA: 'PRAGMA'>: <function Parser.<lambda>>, <TokenType.REFRESH: 'REFRESH'>: <function Parser.<lambda>>, <TokenType.ROLLBACK: 'ROLLBACK'>: <function Parser.<lambda>>, <TokenType.SET: 'SET'>: <function Parser.<lambda>>, <TokenType.TRUNCATE: 'TRUNCATE'>: <function Parser.<lambda>>, <TokenType.UNCACHE: 'UNCACHE'>: <function Parser.<lambda>>, <TokenType.UPDATE: 'UPDATE'>: <function Parser.<lambda>>, <TokenType.USE: 'USE'>: <function Parser.<lambda>>, <TokenType.SEMICOLON: 'SEMICOLON'>: <function Parser.<lambda>>, <TokenType.DECLARE: 'DECLARE'>: <function TSQL.Parser.<lambda>>}
RANGE_PARSERS = {<TokenType.BETWEEN: 'BETWEEN'>: <function Parser.<lambda>>, <TokenType.GLOB: 'GLOB'>: <function binary_range_parser.<locals>._parse_binary_range>, <TokenType.ILIKE: 'ILIKE'>: <function binary_range_parser.<locals>._parse_binary_range>, <TokenType.IN: 'IN'>: <function Parser.<lambda>>, <TokenType.IRLIKE: 'IRLIKE'>: <function binary_range_parser.<locals>._parse_binary_range>, <TokenType.IS: 'IS'>: <function Parser.<lambda>>, <TokenType.LIKE: 'LIKE'>: <function binary_range_parser.<locals>._parse_binary_range>, <TokenType.OVERLAPS: 'OVERLAPS'>: <function binary_range_parser.<locals>._parse_binary_range>, <TokenType.RLIKE: 'RLIKE'>: <function binary_range_parser.<locals>._parse_binary_range>, <TokenType.SIMILAR_TO: 'SIMILAR_TO'>: <function binary_range_parser.<locals>._parse_binary_range>, <TokenType.FOR: 'FOR'>: <function Parser.<lambda>>, <TokenType.DCOLON: 'DCOLON'>: <function TSQL.Parser.<lambda>>}
COLUMN_OPERATORS = {<TokenType.DOT: 'DOT'>: None, <TokenType.DCOLON: 'DCOLON'>: <function TSQL.Parser.<lambda>>, <TokenType.ARROW: 'ARROW'>: <function Parser.<lambda>>, <TokenType.DARROW: 'DARROW'>: <function Parser.<lambda>>, <TokenType.HASH_ARROW: 'HASH_ARROW'>: <function Parser.<lambda>>, <TokenType.DHASH_ARROW: 'DHASH_ARROW'>: <function Parser.<lambda>>, <TokenType.PLACEHOLDER: 'PLACEHOLDER'>: <function Parser.<lambda>>}
ID_VAR_TOKENS = {<TokenType.DATE32: 'DATE32'>, <TokenType.MAP: 'MAP'>, <TokenType.WINDOW: 'WINDOW'>, <TokenType.UMEDIUMINT: 'UMEDIUMINT'>, <TokenType.COPY: 'COPY'>, <TokenType.UBIGINT: 'UBIGINT'>, <TokenType.PSEUDO_TYPE: 'PSEUDO_TYPE'>, <TokenType.UNKNOWN: 'UNKNOWN'>, <TokenType.DICTIONARY: 'DICTIONARY'>, <TokenType.UNIQUEIDENTIFIER: 'UNIQUEIDENTIFIER'>, <TokenType.TDIGEST: 'TDIGEST'>, <TokenType.IS: 'IS'>, <TokenType.XML: 'XML'>, <TokenType.VIEW: 'VIEW'>, <TokenType.TIMESTAMPTZ: 'TIMESTAMPTZ'>, <TokenType.ROWVERSION: 'ROWVERSION'>, <TokenType.REFERENCES: 'REFERENCES'>, <TokenType.TOP: 'TOP'>, <TokenType.UUID: 'UUID'>, <TokenType.DIV: 'DIV'>, <TokenType.IPADDRESS: 'IPADDRESS'>, <TokenType.SERIAL: 'SERIAL'>, <TokenType.VARCHAR: 'VARCHAR'>, <TokenType.INT4RANGE: 'INT4RANGE'>, <TokenType.BIT: 'BIT'>, <TokenType.ROW: 'ROW'>, <TokenType.DATETIME: 'DATETIME'>, <TokenType.BPCHAR: 'BPCHAR'>, <TokenType.SEMI: 'SEMI'>, <TokenType.OPERATOR: 'OPERATOR'>, <TokenType.SOME: 'SOME'>, <TokenType.NUMMULTIRANGE: 'NUMMULTIRANGE'>, <TokenType.LEFT: 'LEFT'>, <TokenType.IMAGE: 'IMAGE'>, <TokenType.CURRENT_DATETIME: 'CURRENT_DATETIME'>, <TokenType.AGGREGATEFUNCTION: 'AGGREGATEFUNCTION'>, <TokenType.TIMESTAMP_NS: 'TIMESTAMP_NS'>, <TokenType.SETTINGS: 'SETTINGS'>, <TokenType.TINYINT: 'TINYINT'>, <TokenType.FOREIGN_KEY: 'FOREIGN_KEY'>, <TokenType.BOOLEAN: 'BOOLEAN'>, <TokenType.BIGDECIMAL: 'BIGDECIMAL'>, <TokenType.DOUBLE: 'DOUBLE'>, <TokenType.MEDIUMBLOB: 'MEDIUMBLOB'>, <TokenType.FALSE: 'FALSE'>, <TokenType.STRAIGHT_JOIN: 'STRAIGHT_JOIN'>, <TokenType.RIGHT: 'RIGHT'>, <TokenType.ORDINALITY: 'ORDINALITY'>, <TokenType.UDECIMAL: 'UDECIMAL'>, <TokenType.NEXT: 'NEXT'>, <TokenType.GEOGRAPHY: 'GEOGRAPHY'>, <TokenType.REPLACE: 'REPLACE'>, <TokenType.USERDEFINED: 'USERDEFINED'>, <TokenType.SHOW: 'SHOW'>, <TokenType.USE: 'USE'>, <TokenType.FIXEDSTRING: 'FIXEDSTRING'>, <TokenType.TIME: 'TIME'>, <TokenType.ESCAPE: 'ESCAPE'>, <TokenType.COMMAND: 'COMMAND'>, <TokenType.TIMESTAMP: 'TIMESTAMP'>, <TokenType.CURRENT_USER: 'CURRENT_USER'>, <TokenType.MERGE: 'MERGE'>, <TokenType.LONGTEXT: 'LONGTEXT'>, <TokenType.INT8RANGE: 'INT8RANGE'>, <TokenType.ARRAY: 'ARRAY'>, <TokenType.NULLABLE: 'NULLABLE'>, <TokenType.ENUM8: 'ENUM8'>, <TokenType.ENUM: 'ENUM'>, <TokenType.ASOF: 'ASOF'>, <TokenType.UTINYINT: 'UTINYINT'>, <TokenType.TIMESTAMP_S: 'TIMESTAMP_S'>, <TokenType.UNNEST: 'UNNEST'>, <TokenType.OBJECT_IDENTIFIER: 'OBJECT_IDENTIFIER'>, <TokenType.UPDATE: 'UPDATE'>, <TokenType.PROCEDURE: 'PROCEDURE'>, <TokenType.ALL: 'ALL'>, <TokenType.APPLY: 'APPLY'>, <TokenType.OVERWRITE: 'OVERWRITE'>, <TokenType.NATURAL: 'NATURAL'>, <TokenType.MODEL: 'MODEL'>, <TokenType.INTERVAL: 'INTERVAL'>, <TokenType.PERCENT: 'PERCENT'>, <TokenType.VECTOR: 'VECTOR'>, <TokenType.OBJECT: 'OBJECT'>, <TokenType.UNIQUE: 'UNIQUE'>, <TokenType.NAME: 'NAME'>, <TokenType.ROLLUP: 'ROLLUP'>, <TokenType.FIRST: 'FIRST'>, <TokenType.TIMESTAMP_MS: 'TIMESTAMP_MS'>, <TokenType.TIMESTAMPLTZ: 'TIMESTAMPLTZ'>, <TokenType.STREAMLIT: 'STREAMLIT'>, <TokenType.LIST: 'LIST'>, <TokenType.RANGE: 'RANGE'>, <TokenType.FORMAT: 'FORMAT'>, <TokenType.YEAR: 'YEAR'>, <TokenType.KEEP: 'KEEP'>, <TokenType.LOAD: 'LOAD'>, <TokenType.PIVOT: 'PIVOT'>, <TokenType.ROWS: 'ROWS'>, <TokenType.SCHEMA: 'SCHEMA'>, <TokenType.EXISTS: 'EXISTS'>, <TokenType.IPV6: 'IPV6'>, <TokenType.UINT: 'UINT'>, <TokenType.INT256: 'INT256'>, <TokenType.TEXT: 'TEXT'>, <TokenType.VARBINARY: 'VARBINARY'>, <TokenType.COLLATE: 'COLLATE'>, <TokenType.BIGINT: 'BIGINT'>, <TokenType.GEOMETRY: 'GEOMETRY'>, <TokenType.VAR: 'VAR'>, <TokenType.CUBE: 'CUBE'>, <TokenType.SET: 'SET'>, <TokenType.SMALLINT: 'SMALLINT'>, <TokenType.IDENTIFIER: 'IDENTIFIER'>, <TokenType.SMALLMONEY: 'SMALLMONEY'>, <TokenType.SMALLSERIAL: 'SMALLSERIAL'>, <TokenType.EXECUTE: 'EXECUTE'>, <TokenType.ASC: 'ASC'>, <TokenType.INET: 'INET'>, <TokenType.NCHAR: 'NCHAR'>, <TokenType.DECIMAL: 'DECIMAL'>, <TokenType.UINT256: 'UINT256'>, <TokenType.WAREHOUSE: 'WAREHOUSE'>, <TokenType.MEDIUMINT: 'MEDIUMINT'>, <TokenType.MEDIUMTEXT: 'MEDIUMTEXT'>, <TokenType.COMMIT: 'COMMIT'>, <TokenType.DELETE: 'DELETE'>, <TokenType.NVARCHAR: 'NVARCHAR'>, <TokenType.INT: 'INT'>, <TokenType.STRUCT: 'STRUCT'>, <TokenType.DEFAULT: 'DEFAULT'>, <TokenType.HLLSKETCH: 'HLLSKETCH'>, <TokenType.INDEX: 'INDEX'>, <TokenType.COMMENT: 'COMMENT'>, <TokenType.BEGIN: 'BEGIN'>, <TokenType.DATETIME64: 'DATETIME64'>, <TokenType.TINYTEXT: 'TINYTEXT'>, <TokenType.FLOAT: 'FLOAT'>, <TokenType.UNPIVOT: 'UNPIVOT'>, <TokenType.TIMETZ: 'TIMETZ'>, <TokenType.TSRANGE: 'TSRANGE'>, <TokenType.MONEY: 'MONEY'>, <TokenType.ENUM16: 'ENUM16'>, <TokenType.HSTORE: 'HSTORE'>, <TokenType.DESCRIBE: 'DESCRIBE'>, <TokenType.TABLE: 'TABLE'>, <TokenType.CURRENT_TIMESTAMP: 'CURRENT_TIMESTAMP'>, <TokenType.REFRESH: 'REFRESH'>, <TokenType.DATE: 'DATE'>, <TokenType.LOWCARDINALITY: 'LOWCARDINALITY'>, <TokenType.END: 'END'>, <TokenType.CACHE: 'CACHE'>, <TokenType.NUMRANGE: 'NUMRANGE'>, <TokenType.ANY: 'ANY'>, <TokenType.ANTI: 'ANTI'>, <TokenType.TINYBLOB: 'TINYBLOB'>, <TokenType.UINT128: 'UINT128'>, <TokenType.JSON: 'JSON'>, <TokenType.BIGSERIAL: 'BIGSERIAL'>, <TokenType.DATEMULTIRANGE: 'DATEMULTIRANGE'>, <TokenType.LONGBLOB: 'LONGBLOB'>, <TokenType.FUNCTION: 'FUNCTION'>, <TokenType.DESC: 'DESC'>, <TokenType.FULL: 'FULL'>, <TokenType.ISNULL: 'ISNULL'>, <TokenType.FINAL: 'FINAL'>, <TokenType.TRUNCATE: 'TRUNCATE'>, <TokenType.OVERLAPS: 'OVERLAPS'>, <TokenType.OFFSET: 'OFFSET'>, <TokenType.TSMULTIRANGE: 'TSMULTIRANGE'>, <TokenType.FILTER: 'FILTER'>, <TokenType.PRAGMA: 'PRAGMA'>, <TokenType.TRUE: 'TRUE'>, <TokenType.INT8MULTIRANGE: 'INT8MULTIRANGE'>, <TokenType.CHAR: 'CHAR'>, <TokenType.SIMPLEAGGREGATEFUNCTION: 'SIMPLEAGGREGATEFUNCTION'>, <TokenType.VARIANT: 'VARIANT'>, <TokenType.SUPER: 'SUPER'>, <TokenType.TEMPORARY: 'TEMPORARY'>, <TokenType.CASE: 'CASE'>, <TokenType.AUTO_INCREMENT: 'AUTO_INCREMENT'>, <TokenType.IPV4: 'IPV4'>, <TokenType.KILL: 'KILL'>, <TokenType.NESTED: 'NESTED'>, <TokenType.BINARY: 'BINARY'>, <TokenType.VOLATILE: 'VOLATILE'>, <TokenType.PARTITION: 'PARTITION'>, <TokenType.DATERANGE: 'DATERANGE'>, <TokenType.TSTZMULTIRANGE: 'TSTZMULTIRANGE'>, <TokenType.CURRENT_TIME: 'CURRENT_TIME'>, <TokenType.TAG: 'TAG'>, <TokenType.NULL: 'NULL'>, <TokenType.JSONB: 'JSONB'>, <TokenType.USMALLINT: 'USMALLINT'>, <TokenType.TIMESTAMPNTZ: 'TIMESTAMPNTZ'>, <TokenType.INT4MULTIRANGE: 'INT4MULTIRANGE'>, <TokenType.TSTZRANGE: 'TSTZRANGE'>, <TokenType.CURRENT_DATE: 'CURRENT_DATE'>, <TokenType.COLUMN: 'COLUMN'>, <TokenType.DATABASE: 'DATABASE'>, <TokenType.CONSTRAINT: 'CONSTRAINT'>, <TokenType.INT128: 'INT128'>, <TokenType.RECURSIVE: 'RECURSIVE'>, <TokenType.STORAGE_INTEGRATION: 'STORAGE_INTEGRATION'>, <TokenType.SEQUENCE: 'SEQUENCE'>, <TokenType.IPPREFIX: 'IPPREFIX'>}
TABLE_ALIAS_TOKENS = {<TokenType.DATE32: 'DATE32'>, <TokenType.MAP: 'MAP'>, <TokenType.UMEDIUMINT: 'UMEDIUMINT'>, <TokenType.COPY: 'COPY'>, <TokenType.UBIGINT: 'UBIGINT'>, <TokenType.PSEUDO_TYPE: 'PSEUDO_TYPE'>, <TokenType.UNKNOWN: 'UNKNOWN'>, <TokenType.DICTIONARY: 'DICTIONARY'>, <TokenType.UNIQUEIDENTIFIER: 'UNIQUEIDENTIFIER'>, <TokenType.TDIGEST: 'TDIGEST'>, <TokenType.IS: 'IS'>, <TokenType.XML: 'XML'>, <TokenType.VIEW: 'VIEW'>, <TokenType.TIMESTAMPTZ: 'TIMESTAMPTZ'>, <TokenType.ROWVERSION: 'ROWVERSION'>, <TokenType.REFERENCES: 'REFERENCES'>, <TokenType.TOP: 'TOP'>, <TokenType.UUID: 'UUID'>, <TokenType.DIV: 'DIV'>, <TokenType.IPADDRESS: 'IPADDRESS'>, <TokenType.SERIAL: 'SERIAL'>, <TokenType.VARCHAR: 'VARCHAR'>, <TokenType.INT4RANGE: 'INT4RANGE'>, <TokenType.BIT: 'BIT'>, <TokenType.ROW: 'ROW'>, <TokenType.DATETIME: 'DATETIME'>, <TokenType.BPCHAR: 'BPCHAR'>, <TokenType.SEMI: 'SEMI'>, <TokenType.OPERATOR: 'OPERATOR'>, <TokenType.SOME: 'SOME'>, <TokenType.NUMMULTIRANGE: 'NUMMULTIRANGE'>, <TokenType.IMAGE: 'IMAGE'>, <TokenType.CURRENT_DATETIME: 'CURRENT_DATETIME'>, <TokenType.AGGREGATEFUNCTION: 'AGGREGATEFUNCTION'>, <TokenType.TIMESTAMP_NS: 'TIMESTAMP_NS'>, <TokenType.SETTINGS: 'SETTINGS'>, <TokenType.TINYINT: 'TINYINT'>, <TokenType.FOREIGN_KEY: 'FOREIGN_KEY'>, <TokenType.BOOLEAN: 'BOOLEAN'>, <TokenType.BIGDECIMAL: 'BIGDECIMAL'>, <TokenType.DOUBLE: 'DOUBLE'>, <TokenType.MEDIUMBLOB: 'MEDIUMBLOB'>, <TokenType.FALSE: 'FALSE'>, <TokenType.STRAIGHT_JOIN: 'STRAIGHT_JOIN'>, <TokenType.ORDINALITY: 'ORDINALITY'>, <TokenType.UDECIMAL: 'UDECIMAL'>, <TokenType.NEXT: 'NEXT'>, <TokenType.GEOGRAPHY: 'GEOGRAPHY'>, <TokenType.REPLACE: 'REPLACE'>, <TokenType.SEQUENCE: 'SEQUENCE'>, <TokenType.USERDEFINED: 'USERDEFINED'>, <TokenType.SHOW: 'SHOW'>, <TokenType.USE: 'USE'>, <TokenType.FIXEDSTRING: 'FIXEDSTRING'>, <TokenType.TIME: 'TIME'>, <TokenType.ESCAPE: 'ESCAPE'>, <TokenType.COMMAND: 'COMMAND'>, <TokenType.TIMESTAMP: 'TIMESTAMP'>, <TokenType.CURRENT_USER: 'CURRENT_USER'>, <TokenType.MERGE: 'MERGE'>, <TokenType.LONGTEXT: 'LONGTEXT'>, <TokenType.INT8RANGE: 'INT8RANGE'>, <TokenType.ARRAY: 'ARRAY'>, <TokenType.NULLABLE: 'NULLABLE'>, <TokenType.ENUM8: 'ENUM8'>, <TokenType.UTINYINT: 'UTINYINT'>, <TokenType.TIMESTAMP_S: 'TIMESTAMP_S'>, <TokenType.UNNEST: 'UNNEST'>, <TokenType.OBJECT_IDENTIFIER: 'OBJECT_IDENTIFIER'>, <TokenType.UPDATE: 'UPDATE'>, <TokenType.PROCEDURE: 'PROCEDURE'>, <TokenType.ALL: 'ALL'>, <TokenType.OVERWRITE: 'OVERWRITE'>, <TokenType.MODEL: 'MODEL'>, <TokenType.INTERVAL: 'INTERVAL'>, <TokenType.PERCENT: 'PERCENT'>, <TokenType.VECTOR: 'VECTOR'>, <TokenType.OBJECT: 'OBJECT'>, <TokenType.UNIQUE: 'UNIQUE'>, <TokenType.NAME: 'NAME'>, <TokenType.ROLLUP: 'ROLLUP'>, <TokenType.FIRST: 'FIRST'>, <TokenType.TIMESTAMP_MS: 'TIMESTAMP_MS'>, <TokenType.TIMESTAMPLTZ: 'TIMESTAMPLTZ'>, <TokenType.STREAMLIT: 'STREAMLIT'>, <TokenType.LIST: 'LIST'>, <TokenType.RANGE: 'RANGE'>, <TokenType.FORMAT: 'FORMAT'>, <TokenType.YEAR: 'YEAR'>, <TokenType.KEEP: 'KEEP'>, <TokenType.LOAD: 'LOAD'>, <TokenType.PIVOT: 'PIVOT'>, <TokenType.ROWS: 'ROWS'>, <TokenType.SCHEMA: 'SCHEMA'>, <TokenType.EXISTS: 'EXISTS'>, <TokenType.IPV6: 'IPV6'>, <TokenType.UINT: 'UINT'>, <TokenType.INT256: 'INT256'>, <TokenType.TEXT: 'TEXT'>, <TokenType.VARBINARY: 'VARBINARY'>, <TokenType.COLLATE: 'COLLATE'>, <TokenType.BIGINT: 'BIGINT'>, <TokenType.GEOMETRY: 'GEOMETRY'>, <TokenType.VAR: 'VAR'>, <TokenType.CUBE: 'CUBE'>, <TokenType.SET: 'SET'>, <TokenType.SMALLINT: 'SMALLINT'>, <TokenType.IDENTIFIER: 'IDENTIFIER'>, <TokenType.SMALLMONEY: 'SMALLMONEY'>, <TokenType.SMALLSERIAL: 'SMALLSERIAL'>, <TokenType.EXECUTE: 'EXECUTE'>, <TokenType.ASC: 'ASC'>, <TokenType.INET: 'INET'>, <TokenType.NCHAR: 'NCHAR'>, <TokenType.DECIMAL: 'DECIMAL'>, <TokenType.UINT256: 'UINT256'>, <TokenType.WAREHOUSE: 'WAREHOUSE'>, <TokenType.MEDIUMINT: 'MEDIUMINT'>, <TokenType.MEDIUMTEXT: 'MEDIUMTEXT'>, <TokenType.COMMIT: 'COMMIT'>, <TokenType.DELETE: 'DELETE'>, <TokenType.NVARCHAR: 'NVARCHAR'>, <TokenType.INT: 'INT'>, <TokenType.STRUCT: 'STRUCT'>, <TokenType.DEFAULT: 'DEFAULT'>, <TokenType.HLLSKETCH: 'HLLSKETCH'>, <TokenType.INDEX: 'INDEX'>, <TokenType.COMMENT: 'COMMENT'>, <TokenType.BEGIN: 'BEGIN'>, <TokenType.DATETIME64: 'DATETIME64'>, <TokenType.TINYTEXT: 'TINYTEXT'>, <TokenType.FLOAT: 'FLOAT'>, <TokenType.UNPIVOT: 'UNPIVOT'>, <TokenType.TIMETZ: 'TIMETZ'>, <TokenType.TSRANGE: 'TSRANGE'>, <TokenType.MONEY: 'MONEY'>, <TokenType.ENUM16: 'ENUM16'>, <TokenType.HSTORE: 'HSTORE'>, <TokenType.DESCRIBE: 'DESCRIBE'>, <TokenType.TABLE: 'TABLE'>, <TokenType.CURRENT_TIMESTAMP: 'CURRENT_TIMESTAMP'>, <TokenType.REFRESH: 'REFRESH'>, <TokenType.DATE: 'DATE'>, <TokenType.LOWCARDINALITY: 'LOWCARDINALITY'>, <TokenType.END: 'END'>, <TokenType.CACHE: 'CACHE'>, <TokenType.NUMRANGE: 'NUMRANGE'>, <TokenType.ANY: 'ANY'>, <TokenType.TINYBLOB: 'TINYBLOB'>, <TokenType.UINT128: 'UINT128'>, <TokenType.JSON: 'JSON'>, <TokenType.BIGSERIAL: 'BIGSERIAL'>, <TokenType.ANTI: 'ANTI'>, <TokenType.DATEMULTIRANGE: 'DATEMULTIRANGE'>, <TokenType.LONGBLOB: 'LONGBLOB'>, <TokenType.FUNCTION: 'FUNCTION'>, <TokenType.DESC: 'DESC'>, <TokenType.ISNULL: 'ISNULL'>, <TokenType.FINAL: 'FINAL'>, <TokenType.TRUNCATE: 'TRUNCATE'>, <TokenType.OVERLAPS: 'OVERLAPS'>, <TokenType.TSMULTIRANGE: 'TSMULTIRANGE'>, <TokenType.FILTER: 'FILTER'>, <TokenType.PRAGMA: 'PRAGMA'>, <TokenType.TRUE: 'TRUE'>, <TokenType.INT8MULTIRANGE: 'INT8MULTIRANGE'>, <TokenType.CHAR: 'CHAR'>, <TokenType.SIMPLEAGGREGATEFUNCTION: 'SIMPLEAGGREGATEFUNCTION'>, <TokenType.VARIANT: 'VARIANT'>, <TokenType.SUPER: 'SUPER'>, <TokenType.TEMPORARY: 'TEMPORARY'>, <TokenType.CASE: 'CASE'>, <TokenType.AUTO_INCREMENT: 'AUTO_INCREMENT'>, <TokenType.IPV4: 'IPV4'>, <TokenType.KILL: 'KILL'>, <TokenType.NESTED: 'NESTED'>, <TokenType.BINARY: 'BINARY'>, <TokenType.VOLATILE: 'VOLATILE'>, <TokenType.PARTITION: 'PARTITION'>, <TokenType.DATERANGE: 'DATERANGE'>, <TokenType.TSTZMULTIRANGE: 'TSTZMULTIRANGE'>, <TokenType.CURRENT_TIME: 'CURRENT_TIME'>, <TokenType.TAG: 'TAG'>, <TokenType.NULL: 'NULL'>, <TokenType.JSONB: 'JSONB'>, <TokenType.USMALLINT: 'USMALLINT'>, <TokenType.TIMESTAMPNTZ: 'TIMESTAMPNTZ'>, <TokenType.INT4MULTIRANGE: 'INT4MULTIRANGE'>, <TokenType.TSTZRANGE: 'TSTZRANGE'>, <TokenType.CURRENT_DATE: 'CURRENT_DATE'>, <TokenType.COLUMN: 'COLUMN'>, <TokenType.DATABASE: 'DATABASE'>, <TokenType.CONSTRAINT: 'CONSTRAINT'>, <TokenType.INT128: 'INT128'>, <TokenType.RECURSIVE: 'RECURSIVE'>, <TokenType.STORAGE_INTEGRATION: 'STORAGE_INTEGRATION'>, <TokenType.ENUM: 'ENUM'>, <TokenType.IPPREFIX: 'IPPREFIX'>}
SHOW_TRIE: Dict = {}
SET_TRIE: Dict = {'GLOBAL': {0: True}, 'LOCAL': {0: True}, 'SESSION': {0: True}, 'TRANSACTION': {0: True}}

##### Inherited Members

sqlglot.parser.Parser
  Parser
  NO_PAREN_FUNCTIONS
  STRUCT_TYPE_TOKENS
  NESTED_TYPE_TOKENS
  ENUM_TYPE_TOKENS
  AGGREGATE_TYPE_TOKENS
  TYPE_TOKENS
  SIGNED_TO_UNSIGNED_TYPE_TOKEN
  SUBQUERY_PREDICATES
  RESERVED_TOKENS
  DB_CREATABLES
  CREATABLES
  INTERVAL_VARS
  ALIAS_TOKENS
  ARRAY_CONSTRUCTORS
  COMMENT_TABLE_ALIAS_TOKENS
  UPDATE_ALIAS_TOKENS
  TRIM_TYPES
  FUNC_TOKENS
  CONJUNCTION
  ASSIGNMENT
  DISJUNCTION
  EQUALITY
  COMPARISON
  BITWISE
  TERM
  FACTOR
  EXPONENT
  TIMES
  TIMESTAMPS
  SET_OPERATIONS
  JOIN_METHODS
  JOIN_SIDES
  JOIN_KINDS
  LAMBDAS
  EXPRESSION_PARSERS
  UNARY_PARSERS
  STRING_PARSERS
  NUMERIC_PARSERS
  PRIMARY_PARSERS
  PLACEHOLDER_PARSERS
  PROPERTY_PARSERS
  CONSTRAINT_PARSERS
  ALTER_PARSERS
  ALTER_ALTER_PARSERS
  SCHEMA_UNNAMED_CONSTRAINTS
  NO_PAREN_FUNCTION_PARSERS
  INVALID_FUNC_NAME_TOKENS
  FUNCTIONS_WITH_ALIASED_ARGS
  KEY_VALUE_DEFINITIONS
  FUNCTION_PARSERS
  SET_PARSERS
  SHOW_PARSERS
  TYPE_LITERAL_PARSERS
  TYPE_CONVERTERS
  DDL_SELECT_TOKENS
  PRE_VOLATILE_TOKENS
  TRANSACTION_KIND
  TRANSACTION_CHARACTERISTICS
  CONFLICT_ACTIONS
  CREATE_SEQUENCE
  ISOLATED_LOADING_OPTIONS
  USABLES
  CAST_ACTIONS
  SCHEMA_BINDING_OPTIONS
  KEY_CONSTRAINT_OPTIONS
  INSERT_ALTERNATIVES
  CLONE_KEYWORDS
  HISTORICAL_DATA_PREFIX
  HISTORICAL_DATA_KIND
  OPCLASS_FOLLOW_KEYWORDS
  OPTYPE_FOLLOW_TOKENS
  TABLE_INDEX_HINT_TOKENS
  VIEW_ATTRIBUTES
  WINDOW_ALIAS_TOKENS
  WINDOW_BEFORE_PAREN_TOKENS
  WINDOW_SIDES
  JSON_KEY_VALUE_SEPARATOR_TOKENS
  FETCH_TOKENS
  ADD_CONSTRAINT_TOKENS
  DISTINCT_TOKENS
  NULL_TOKENS
  UNNEST_OFFSET_ALIAS_TOKENS
  SELECT_START_TOKENS
  COPY_INTO_VARLEN_OPTIONS
  STRICT_CAST
  PREFIXED_PIVOT_COLUMNS
  IDENTIFY_PIVOT_STRINGS
  TABLESAMPLE_CSV
  DEFAULT_SAMPLING_METHOD
  TRIM_PATTERN_FIRST
  MODIFIERS_ATTACHED_TO_SET_OP
  SET_OP_MODIFIERS
  JSON_ARROWS_REQUIRE_JSON_TYPE
  COLON_IS_VARIANT_EXTRACT
  VALUES_FOLLOWED_BY_PAREN
  SUPPORTS_IMPLICIT_UNNEST
  INTERVAL_SPANS
  SUPPORTS_PARTITION_SELECTION
  error_level
  error_message_context
  max_errors
  dialect
  reset
  parse
  parse_into
  check_errors
  raise_error
  expression
  validate_expression
  errors
  sql

class TSQL.Generator(sqlglot.generator.Generator): View Source

     782    class Generator(generator.Generator):
     783        LIMIT_IS_TOP = True
     784        QUERY_HINTS = False
     785        RETURNING_END = False
     786        NVL2_SUPPORTED = False
     787        ALTER_TABLE_INCLUDE_COLUMN_KEYWORD = False
     788        LIMIT_FETCH = "FETCH"
     789        COMPUTED_COLUMN_WITH_TYPE = False
     790        CTE_RECURSIVE_KEYWORD_REQUIRED = False
     791        ENSURE_BOOLS = True
     792        NULL_ORDERING_SUPPORTED = None
     793        SUPPORTS_SINGLE_ARG_CONCAT = False
     794        TABLESAMPLE_SEED_KEYWORD = "REPEATABLE"
     795        SUPPORTS_SELECT_INTO = True
     796        JSON_PATH_BRACKETED_KEY_SUPPORTED = False
     797        SUPPORTS_TO_NUMBER = False
     798        SET_OP_MODIFIERS = False
     799        COPY_PARAMS_EQ_REQUIRED = True
     800        PARSE_JSON_NAME = None
     801
     802        EXPRESSIONS_WITHOUT_NESTED_CTES = {
     803            exp.Create,
     804            exp.Delete,
     805            exp.Insert,
     806            exp.Intersect,
     807            exp.Except,
     808            exp.Merge,
     809            exp.Select,
     810            exp.Subquery,
     811            exp.Union,
     812            exp.Update,
     813        }
     814
     815        SUPPORTED_JSON_PATH_PARTS = {
     816            exp.JSONPathKey,
     817            exp.JSONPathRoot,
     818            exp.JSONPathSubscript,
     819        }
     820
     821        TYPE_MAPPING = {
     822            **generator.Generator.TYPE_MAPPING,
     823            exp.DataType.Type.BOOLEAN: "BIT",
     824            exp.DataType.Type.DECIMAL: "NUMERIC",
     825            exp.DataType.Type.DATETIME: "DATETIME2",
     826            exp.DataType.Type.DOUBLE: "FLOAT",
     827            exp.DataType.Type.INT: "INTEGER",
     828            exp.DataType.Type.ROWVERSION: "ROWVERSION",
     829            exp.DataType.Type.TEXT: "VARCHAR(MAX)",
     830            exp.DataType.Type.TIMESTAMP: "DATETIME2",
     831            exp.DataType.Type.TIMESTAMPTZ: "DATETIMEOFFSET",
     832            exp.DataType.Type.UTINYINT: "TINYINT",
     833            exp.DataType.Type.VARIANT: "SQL_VARIANT",
     834        }
     835
     836        TYPE_MAPPING.pop(exp.DataType.Type.NCHAR)
     837        TYPE_MAPPING.pop(exp.DataType.Type.NVARCHAR)
     838
     839        TRANSFORMS = {
     840            **generator.Generator.TRANSFORMS,
     841            exp.AnyValue: any_value_to_max_sql,
     842            exp.ArrayToString: rename_func("STRING_AGG"),
     843            exp.AutoIncrementColumnConstraint: lambda *_: "IDENTITY",
     844            exp.DateAdd: date_delta_sql("DATEADD"),
     845            exp.DateDiff: date_delta_sql("DATEDIFF"),
     846            exp.CTE: transforms.preprocess([qualify_derived_table_outputs]),
     847            exp.CurrentDate: rename_func("GETDATE"),
     848            exp.CurrentTimestamp: rename_func("GETDATE"),
     849            exp.DateStrToDate: datestrtodate_sql,
     850            exp.Extract: rename_func("DATEPART"),
     851            exp.GeneratedAsIdentityColumnConstraint: generatedasidentitycolumnconstraint_sql,
     852            exp.GroupConcat: _string_agg_sql,
     853            exp.If: rename_func("IIF"),
     854            exp.JSONExtract: _json_extract_sql,
     855            exp.JSONExtractScalar: _json_extract_sql,
     856            exp.LastDay: lambda self, e: self.func("EOMONTH", e.this),
     857            exp.Max: max_or_greatest,
     858            exp.MD5: lambda self, e: self.func("HASHBYTES", exp.Literal.string("MD5"), e.this),
     859            exp.Min: min_or_least,
     860            exp.NumberToStr: _format_sql,
     861            exp.Repeat: rename_func("REPLICATE"),
     862            exp.Select: transforms.preprocess(
     863                [
     864                    transforms.eliminate_distinct_on,
     865                    transforms.eliminate_semi_and_anti_joins,
     866                    transforms.eliminate_qualify,
     867                ]
     868            ),
     869            exp.Stddev: rename_func("STDEV"),
     870            exp.StrPosition: lambda self, e: self.func(
     871                "CHARINDEX", e.args.get("substr"), e.this, e.args.get("position")
     872            ),
     873            exp.Subquery: transforms.preprocess([qualify_derived_table_outputs]),
     874            exp.SHA: lambda self, e: self.func("HASHBYTES", exp.Literal.string("SHA1"), e.this),
     875            exp.SHA2: lambda self, e: self.func(
     876                "HASHBYTES", exp.Literal.string(f"SHA2_{e.args.get('length', 256)}"), e.this
     877            ),
     878            exp.TemporaryProperty: lambda self, e: "",
     879            exp.TimeStrToTime: lambda self, e: self.sql(
     880                exp.cast(e.this, exp.DataType.Type.DATETIME)
     881            ),
     882            exp.TimeToStr: _format_sql,
     883            exp.Trim: trim_sql,
     884            exp.TsOrDsAdd: date_delta_sql("DATEADD", cast=True),
     885            exp.TsOrDsDiff: date_delta_sql("DATEDIFF"),
     886        }
     887
     888        TRANSFORMS.pop(exp.ReturnsProperty)
     889
     890        PROPERTIES_LOCATION = {
     891            **generator.Generator.PROPERTIES_LOCATION,
     892            exp.VolatileProperty: exp.Properties.Location.UNSUPPORTED,
     893        }
     894
     895        def scope_resolution(self, rhs: str, scope_name: str) -> str:
     896            return f"{scope_name}::{rhs}"
     897
     898        def select_sql(self, expression: exp.Select) -> str:
     899            if expression.args.get("offset"):
     900                if not expression.args.get("order"):
     901                    # ORDER BY is required in order to use OFFSET in a query, so we use
     902                    # a noop order by, since we don't really care about the order.
     903                    # See: https://www.microsoftpressstore.com/articles/article.aspx?p=2314819
     904                    expression.order_by(exp.select(exp.null()).subquery(), copy=False)
     905
     906                limit = expression.args.get("limit")
     907                if isinstance(limit, exp.Limit):
     908                    # TOP and OFFSET can't be combined, we need use FETCH instead of TOP
     909                    # we replace here because otherwise TOP would be generated in select_sql
     910                    limit.replace(exp.Fetch(direction="FIRST", count=limit.expression))
     911
     912            return super().select_sql(expression)
     913
     914        def convert_sql(self, expression: exp.Convert) -> str:
     915            name = "TRY_CONVERT" if expression.args.get("safe") else "CONVERT"
     916            return self.func(
     917                name, expression.this, expression.expression, expression.args.get("style")
     918            )
     919
     920        def queryoption_sql(self, expression: exp.QueryOption) -> str:
     921            option = self.sql(expression, "this")
     922            value = self.sql(expression, "expression")
     923            if value:
     924                optional_equal_sign = "= " if option in OPTIONS_THAT_REQUIRE_EQUAL else ""
     925                return f"{option} {optional_equal_sign}{value}"
     926            return option
     927
     928        def lateral_op(self, expression: exp.Lateral) -> str:
     929            cross_apply = expression.args.get("cross_apply")
     930            if cross_apply is True:
     931                return "CROSS APPLY"
     932            if cross_apply is False:
     933                return "OUTER APPLY"
     934
     935            # TODO: perhaps we can check if the parent is a Join and transpile it appropriately
     936            self.unsupported("LATERAL clause is not supported.")
     937            return "LATERAL"
     938
     939        def timefromparts_sql(self, expression: exp.TimeFromParts) -> str:
     940            nano = expression.args.get("nano")
     941            if nano is not None:
     942                nano.pop()
     943                self.unsupported("Specifying nanoseconds is not supported in TIMEFROMPARTS.")
     944
     945            if expression.args.get("fractions") is None:
     946                expression.set("fractions", exp.Literal.number(0))
     947            if expression.args.get("precision") is None:
     948                expression.set("precision", exp.Literal.number(0))
     949
     950            return rename_func("TIMEFROMPARTS")(self, expression)
     951
     952        def timestampfromparts_sql(self, expression: exp.TimestampFromParts) -> str:
     953            zone = expression.args.get("zone")
     954            if zone is not None:
     955                zone.pop()
     956                self.unsupported("Time zone is not supported in DATETIMEFROMPARTS.")
     957
     958            nano = expression.args.get("nano")
     959            if nano is not None:
     960                nano.pop()
     961                self.unsupported("Specifying nanoseconds is not supported in DATETIMEFROMPARTS.")
     962
     963            if expression.args.get("milli") is None:
     964                expression.set("milli", exp.Literal.number(0))
     965
     966            return rename_func("DATETIMEFROMPARTS")(self, expression)
     967
     968        def setitem_sql(self, expression: exp.SetItem) -> str:
     969            this = expression.this
     970            if isinstance(this, exp.EQ) and not isinstance(this.left, exp.Parameter):
     971                # T-SQL does not use '=' in SET command, except when the LHS is a variable.
     972                return f"{self.sql(this.left)} {self.sql(this.right)}"
     973
     974            return super().setitem_sql(expression)
     975
     976        def boolean_sql(self, expression: exp.Boolean) -> str:
     977            if type(expression.parent) in BIT_TYPES:
     978                return "1" if expression.this else "0"
     979
     980            return "(1 = 1)" if expression.this else "(1 = 0)"
     981
     982        def is_sql(self, expression: exp.Is) -> str:
     983            if isinstance(expression.expression, exp.Boolean):
     984                return self.binary(expression, "=")
     985            return self.binary(expression, "IS")
     986
     987        def createable_sql(self, expression: exp.Create, locations: t.DefaultDict) -> str:
     988            sql = self.sql(expression, "this")
     989            properties = expression.args.get("properties")
     990
     991            if sql[:1] != "#" and any(
     992                isinstance(prop, exp.TemporaryProperty)
     993                for prop in (properties.expressions if properties else [])
     994            ):
     995                sql = f"[#{sql[1:]}" if sql.startswith("[") else f"#{sql}"
     996
     997            return sql
     998
     999        def create_sql(self, expression: exp.Create) -> str:
    1000            kind = expression.kind
    1001            exists = expression.args.pop("exists", None)
    1002
    1003            like_property = expression.find(exp.LikeProperty)
    1004            if like_property:
    1005                ctas_expression = like_property.this
    1006            else:
    1007                ctas_expression = expression.expression
    1008
    1009            if kind == "VIEW":
    1010                expression.this.set("catalog", None)
    1011                with_ = expression.args.get("with")
    1012                if ctas_expression and with_:
    1013                    # We've already preprocessed the Create expression to bubble up any nested CTEs,
    1014                    # but CREATE VIEW actually requires the WITH clause to come after it so we need
    1015                    # to amend the AST by moving the CTEs to the CREATE VIEW statement's query.
    1016                    ctas_expression.set("with", with_.pop())
    1017
    1018            sql = super().create_sql(expression)
    1019
    1020            table = expression.find(exp.Table)
    1021
    1022            # Convert CTAS statement to SELECT .. INTO ..
    1023            if kind == "TABLE" and ctas_expression:
    1024                if isinstance(ctas_expression, exp.UNWRAPPED_QUERIES):
    1025                    ctas_expression = ctas_expression.subquery()
    1026
    1027                select_into = exp.select("*").from_(exp.alias_(ctas_expression, "temp", table=True))
    1028                select_into.set("into", exp.Into(this=table))
    1029
    1030                if like_property:
    1031                    select_into.limit(0, copy=False)
    1032
    1033                sql = self.sql(select_into)
    1034
    1035            if exists:
    1036                identifier = self.sql(exp.Literal.string(exp.table_name(table) if table else ""))
    1037                sql = self.sql(exp.Literal.string(sql))
    1038                if kind == "SCHEMA":
    1039                    sql = f"""IF NOT EXISTS (SELECT * FROM information_schema.schemata WHERE schema_name = {identifier}) EXEC({sql})"""
    1040                elif kind == "TABLE":
    1041                    assert table
    1042                    where = exp.and_(
    1043                        exp.column("table_name").eq(table.name),
    1044                        exp.column("table_schema").eq(table.db) if table.db else None,
    1045                        exp.column("table_catalog").eq(table.catalog) if table.catalog else None,
    1046                    )
    1047                    sql = f"""IF NOT EXISTS (SELECT * FROM information_schema.tables WHERE {where}) EXEC({sql})"""
    1048                elif kind == "INDEX":
    1049                    index = self.sql(exp.Literal.string(expression.this.text("this")))
    1050                    sql = f"""IF NOT EXISTS (SELECT * FROM sys.indexes WHERE object_id = object_id({identifier}) AND name = {index}) EXEC({sql})"""
    1051            elif expression.args.get("replace"):
    1052                sql = sql.replace("CREATE OR REPLACE ", "CREATE OR ALTER ", 1)
    1053
    1054            return self.prepend_ctes(expression, sql)
    1055
    1056        def offset_sql(self, expression: exp.Offset) -> str:
    1057            return f"{super().offset_sql(expression)} ROWS"
    1058
    1059        def version_sql(self, expression: exp.Version) -> str:
    1060            name = "SYSTEM_TIME" if expression.name == "TIMESTAMP" else expression.name
    1061            this = f"FOR {name}"
    1062            expr = expression.expression
    1063            kind = expression.text("kind")
    1064            if kind in ("FROM", "BETWEEN"):
    1065                args = expr.expressions
    1066                sep = "TO" if kind == "FROM" else "AND"
    1067                expr_sql = f"{self.sql(seq_get(args, 0))} {sep} {self.sql(seq_get(args, 1))}"
    1068            else:
    1069                expr_sql = self.sql(expr)
    1070
    1071            expr_sql = f" {expr_sql}" if expr_sql else ""
    1072            return f"{this} {kind}{expr_sql}"
    1073
    1074        def returnsproperty_sql(self, expression: exp.ReturnsProperty) -> str:
    1075            table = expression.args.get("table")
    1076            table = f"{table} " if table else ""
    1077            return f"RETURNS {table}{self.sql(expression, 'this')}"
    1078
    1079        def returning_sql(self, expression: exp.Returning) -> str:
    1080            into = self.sql(expression, "into")
    1081            into = self.seg(f"INTO {into}") if into else ""
    1082            return f"{self.seg('OUTPUT')} {self.expressions(expression, flat=True)}{into}"
    1083
    1084        def transaction_sql(self, expression: exp.Transaction) -> str:
    1085            this = self.sql(expression, "this")
    1086            this = f" {this}" if this else ""
    1087            mark = self.sql(expression, "mark")
    1088            mark = f" WITH MARK {mark}" if mark else ""
    1089            return f"BEGIN TRANSACTION{this}{mark}"
    1090
    1091        def commit_sql(self, expression: exp.Commit) -> str:
    1092            this = self.sql(expression, "this")
    1093            this = f" {this}" if this else ""
    1094            durability = expression.args.get("durability")
    1095            durability = (
    1096                f" WITH (DELAYED_DURABILITY = {'ON' if durability else 'OFF'})"
    1097                if durability is not None
    1098                else ""
    1099            )
    1100            return f"COMMIT TRANSACTION{this}{durability}"
    1101
    1102        def rollback_sql(self, expression: exp.Rollback) -> str:
    1103            this = self.sql(expression, "this")
    1104            this = f" {this}" if this else ""
    1105            return f"ROLLBACK TRANSACTION{this}"
    1106
    1107        def identifier_sql(self, expression: exp.Identifier) -> str:
    1108            identifier = super().identifier_sql(expression)
    1109
    1110            if expression.args.get("global"):
    1111                identifier = f"##{identifier}"
    1112            elif expression.args.get("temporary"):
    1113                identifier = f"#{identifier}"
    1114
    1115            return identifier
    1116
    1117        def constraint_sql(self, expression: exp.Constraint) -> str:
    1118            this = self.sql(expression, "this")
    1119            expressions = self.expressions(expression, flat=True, sep=" ")
    1120            return f"CONSTRAINT {this} {expressions}"
    1121
    1122        def length_sql(self, expression: exp.Length) -> str:
    1123            return self._uncast_text(expression, "LEN")
    1124
    1125        def right_sql(self, expression: exp.Right) -> str:
    1126            return self._uncast_text(expression, "RIGHT")
    1127
    1128        def left_sql(self, expression: exp.Left) -> str:
    1129            return self._uncast_text(expression, "LEFT")
    1130
    1131        def _uncast_text(self, expression: exp.Expression, name: str) -> str:
    1132            this = expression.this
    1133            if isinstance(this, exp.Cast) and this.is_type(exp.DataType.Type.TEXT):
    1134                this_sql = self.sql(this, "this")
    1135            else:
    1136                this_sql = self.sql(this)
    1137            expression_sql = self.sql(expression, "expression")
    1138            return self.func(name, this_sql, expression_sql if expression_sql else None)
    1139
    1140        def partition_sql(self, expression: exp.Partition) -> str:
    1141            return f"WITH (PARTITIONS({self.expressions(expression, flat=True)}))"
    1142
    1143        def altertable_sql(self, expression: exp.AlterTable) -> str:
    1144            action = seq_get(expression.args.get("actions") or [], 0)
    1145            if isinstance(action, exp.RenameTable):
    1146                return f"EXEC sp_rename '{self.sql(expression.this)}', '{action.this.name}'"
    1147            return super().altertable_sql(expression)
    1148
    1149        def drop_sql(self, expression: exp.Drop) -> str:
    1150            if expression.args["kind"] == "VIEW":
    1151                expression.this.set("catalog", None)
    1152            return super().drop_sql(expression)
    1153
    1154        def declare_sql(self, expression: exp.Declare) -> str:
    1155            return f"DECLARE {self.expressions(expression, flat=True)}"
    1156
    1157        def declareitem_sql(self, expression: exp.DeclareItem) -> str:
    1158            variable = self.sql(expression, "this")
    1159            default = self.sql(expression, "default")
    1160            default = f" = {default}" if default else ""
    1161
    1162            kind = self.sql(expression, "kind")
    1163            if isinstance(expression.args.get("kind"), exp.Schema):
    1164                kind = f"TABLE {kind}"
    1165
    1166            return f"{variable} AS {kind}{default}"
    1167
    1168        def options_modifier(self, expression: exp.Expression) -> str:
    1169            options = self.expressions(expression, key="options")
    1170            return f" OPTION{self.wrap(options)}" if options else ""

Generator converts a given syntax tree to the corresponding SQL string.

###### Arguments:

- pretty: Whether to format the produced SQL string. Default: False.
- identify: Determines when an identifier should be quoted. Possible values are: False (default): Never quote, except in cases where it's mandatory by the dialect. True or 'always': Always quote. 'safe': Only quote identifiers that are case insensitive.
- normalize: Whether to normalize identifiers to lowercase. Default: False.
- pad: The pad size in a formatted string. For example, this affects the indentation of a projection in a query, relative to its nesting level. Default: 2.
- indent: The indentation size in a formatted string. For example, this affects the indentation of subqueries and filters under a `WHERE` clause. Default: 2.
- normalize_functions: How to normalize function names. Possible values are: "upper" or True (default): Convert names to uppercase. "lower": Convert names to lowercase. False: Disables function name normalization.
- unsupported_level: Determines the generator's behavior when it encounters unsupported expressions. Default ErrorLevel.WARN.
- max_unsupported: Maximum number of unsupported messages to include in a raised UnsupportedError. This is only relevant if unsupported_level is ErrorLevel.RAISE. Default: 3
- leading_comma: Whether the comma is leading or trailing in select expressions. This is only relevant when generating in pretty mode. Default: False
- max_text_width: The max number of characters in a segment before creating new lines in pretty mode. The default is on the smaller end because the length only represents a segment and not the true line length. Default: 80
- comments: Whether to preserve comments in the output SQL code. Default: True

LIMIT_IS_TOP = True
QUERY_HINTS = False
RETURNING_END = False
NVL2_SUPPORTED = False
ALTER_TABLE_INCLUDE_COLUMN_KEYWORD = False
LIMIT_FETCH = 'FETCH'
COMPUTED_COLUMN_WITH_TYPE = False
CTE_RECURSIVE_KEYWORD_REQUIRED = False
ENSURE_BOOLS = True
NULL_ORDERING_SUPPORTED = None
SUPPORTS_SINGLE_ARG_CONCAT = False
TABLESAMPLE_SEED_KEYWORD = 'REPEATABLE'
SUPPORTS_SELECT_INTO = True
JSON_PATH_BRACKETED_KEY_SUPPORTED = False
SUPPORTS_TO_NUMBER = False
SET_OP_MODIFIERS = False
COPY_PARAMS_EQ_REQUIRED = True
PARSE_JSON_NAME = None
EXPRESSIONS_WITHOUT_NESTED_CTES = {<class 'sqlglot.expressions.Insert'>, <class 'sqlglot.expressions.Update'>, <class 'sqlglot.expressions.Intersect'>, <class 'sqlglot.expressions.Except'>, <class 'sqlglot.expressions.Union'>, <class 'sqlglot.expressions.Create'>, <class 'sqlglot.expressions.Select'>, <class 'sqlglot.expressions.Merge'>, <class 'sqlglot.expressions.Subquery'>, <class 'sqlglot.expressions.Delete'>}
SUPPORTED_JSON_PATH_PARTS = {<class 'sqlglot.expressions.JSONPathKey'>, <class 'sqlglot.expressions.JSONPathSubscript'>, <class 'sqlglot.expressions.JSONPathRoot'>}
TYPE_MAPPING = {<Type.MEDIUMTEXT: 'MEDIUMTEXT'>: 'TEXT', <Type.LONGTEXT: 'LONGTEXT'>: 'TEXT', <Type.TINYTEXT: 'TINYTEXT'>: 'TEXT', <Type.MEDIUMBLOB: 'MEDIUMBLOB'>: 'BLOB', <Type.LONGBLOB: 'LONGBLOB'>: 'BLOB', <Type.TINYBLOB: 'TINYBLOB'>: 'BLOB', <Type.INET: 'INET'>: 'INET', <Type.ROWVERSION: 'ROWVERSION'>: 'ROWVERSION', <Type.BOOLEAN: 'BOOLEAN'>: 'BIT', <Type.DECIMAL: 'DECIMAL'>: 'NUMERIC', <Type.DATETIME: 'DATETIME'>: 'DATETIME2', <Type.DOUBLE: 'DOUBLE'>: 'FLOAT', <Type.INT: 'INT'>: 'INTEGER', <Type.TEXT: 'TEXT'>: 'VARCHAR(MAX)', <Type.TIMESTAMP: 'TIMESTAMP'>: 'DATETIME2', <Type.TIMESTAMPTZ: 'TIMESTAMPTZ'>: 'DATETIMEOFFSET', <Type.UTINYINT: 'UTINYINT'>: 'TINYINT', <Type.VARIANT: 'VARIANT'>: 'SQL_VARIANT'}
TRANSFORMS = {<class 'sqlglot.expressions.JSONPathKey'>: <function <lambda>>, <class 'sqlglot.expressions.JSONPathRoot'>: <function <lambda>>, <class 'sqlglot.expressions.JSONPathSubscript'>: <function <lambda>>, <class 'sqlglot.expressions.AllowedValuesProperty'>: <function Generator.<lambda>>, <class 'sqlglot.expressions.AutoRefreshProperty'>: <function Generator.<lambda>>, <class 'sqlglot.expressions.BackupProperty'>: <function Generator.<lambda>>, <class 'sqlglot.expressions.CaseSpecificColumnConstraint'>: <function Generator.<lambda>>, <class 'sqlglot.expressions.CharacterSetColumnConstraint'>: <function Generator.<lambda>>, <class 'sqlglot.expressions.CharacterSetProperty'>: <function Generator.<lambda>>, <class 'sqlglot.expressions.ClusteredColumnConstraint'>: <function Generator.<lambda>>, <class 'sqlglot.expressions.CollateColumnConstraint'>: <function Generator.<lambda>>, <class 'sqlglot.expressions.CommentColumnConstraint'>: <function Generator.<lambda>>, <class 'sqlglot.expressions.ConnectByRoot'>: <function Generator.<lambda>>, <class 'sqlglot.expressions.CopyGrantsProperty'>: <function Generator.<lambda>>, <class 'sqlglot.expressions.DateFormatColumnConstraint'>: <function Generator.<lambda>>, <class 'sqlglot.expressions.DefaultColumnConstraint'>: <function Generator.<lambda>>, <class 'sqlglot.expressions.DynamicProperty'>: <function Generator.<lambda>>, <class 'sqlglot.expressions.EncodeColumnConstraint'>: <function Generator.<lambda>>, <class 'sqlglot.expressions.EphemeralColumnConstraint'>: <function Generator.<lambda>>, <class 'sqlglot.expressions.ExcludeColumnConstraint'>: <function Generator.<lambda>>, <class 'sqlglot.expressions.ExecuteAsProperty'>: <function Generator.<lambda>>, <class 'sqlglot.expressions.ExternalProperty'>: <function Generator.<lambda>>, <class 'sqlglot.expressions.GlobalProperty'>: <function Generator.<lambda>>, <class 'sqlglot.expressions.HeapProperty'>: <function Generator.<lambda>>, <class 'sqlglot.expressions.IcebergProperty'>: <function Generator.<lambda>>, <class 'sqlglot.expressions.InheritsProperty'>: <function Generator.<lambda>>, <class 'sqlglot.expressions.InlineLengthColumnConstraint'>: <function Generator.<lambda>>, <class 'sqlglot.expressions.InputModelProperty'>: <function Generator.<lambda>>, <class 'sqlglot.expressions.IntervalSpan'>: <function Generator.<lambda>>, <class 'sqlglot.expressions.LanguageProperty'>: <function Generator.<lambda>>, <class 'sqlglot.expressions.LocationProperty'>: <function Generator.<lambda>>, <class 'sqlglot.expressions.LogProperty'>: <function Generator.<lambda>>, <class 'sqlglot.expressions.MaterializedProperty'>: <function Generator.<lambda>>, <class 'sqlglot.expressions.NonClusteredColumnConstraint'>: <function Generator.<lambda>>, <class 'sqlglot.expressions.NoPrimaryIndexProperty'>: <function Generator.<lambda>>, <class 'sqlglot.expressions.NotForReplicationColumnConstraint'>: <function Generator.<lambda>>, <class 'sqlglot.expressions.OnCommitProperty'>: <function Generator.<lambda>>, <class 'sqlglot.expressions.OnProperty'>: <function Generator.<lambda>>, <class 'sqlglot.expressions.OnUpdateColumnConstraint'>: <function Generator.<lambda>>, <class 'sqlglot.expressions.OutputModelProperty'>: <function Generator.<lambda>>, <class 'sqlglot.expressions.PathColumnConstraint'>: <function Generator.<lambda>>, <class 'sqlglot.expressions.ProjectionPolicyColumnConstraint'>: <function Generator.<lambda>>, <class 'sqlglot.expressions.RemoteWithConnectionModelProperty'>: <function Generator.<lambda>>, <class 'sqlglot.expressions.SampleProperty'>: <function Generator.<lambda>>, <class 'sqlglot.expressions.SecureProperty'>: <function Generator.<lambda>>, <class 'sqlglot.expressions.SetConfigProperty'>: <function Generator.<lambda>>, <class 'sqlglot.expressions.SetProperty'>: <function Generator.<lambda>>, <class 'sqlglot.expressions.SettingsProperty'>: <function Generator.<lambda>>, <class 'sqlglot.expressions.SharingProperty'>: <function Generator.<lambda>>, <class 'sqlglot.expressions.SqlReadWriteProperty'>: <function Generator.<lambda>>, <class 'sqlglot.expressions.SqlSecurityProperty'>: <function Generator.<lambda>>, <class 'sqlglot.expressions.StabilityProperty'>: <function Generator.<lambda>>, <class 'sqlglot.expressions.StrictProperty'>: <function Generator.<lambda>>, <class 'sqlglot.expressions.TemporaryProperty'>: <function TSQL.Generator.<lambda>>, <class 'sqlglot.expressions.TagColumnConstraint'>: <function Generator.<lambda>>, <class 'sqlglot.expressions.TitleColumnConstraint'>: <function Generator.<lambda>>, <class 'sqlglot.expressions.ToMap'>: <function Generator.<lambda>>, <class 'sqlglot.expressions.ToTableProperty'>: <function Generator.<lambda>>, <class 'sqlglot.expressions.TransformModelProperty'>: <function Generator.<lambda>>, <class 'sqlglot.expressions.TransientProperty'>: <function Generator.<lambda>>, <class 'sqlglot.expressions.UppercaseColumnConstraint'>: <function Generator.<lambda>>, <class 'sqlglot.expressions.UnloggedProperty'>: <function Generator.<lambda>>, <class 'sqlglot.expressions.VarMap'>: <function Generator.<lambda>>, <class 'sqlglot.expressions.ViewAttributeProperty'>: <function Generator.<lambda>>, <class 'sqlglot.expressions.VolatileProperty'>: <function Generator.<lambda>>, <class 'sqlglot.expressions.WithJournalTableProperty'>: <function Generator.<lambda>>, <class 'sqlglot.expressions.WithSchemaBindingProperty'>: <function Generator.<lambda>>, <class 'sqlglot.expressions.WithOperator'>: <function Generator.<lambda>>, <class 'sqlglot.expressions.AnyValue'>: <function any_value_to_max_sql>, <class 'sqlglot.expressions.ArrayToString'>: <function rename_func.<locals>.<lambda>>, <class 'sqlglot.expressions.AutoIncrementColumnConstraint'>: <function TSQL.Generator.<lambda>>, <class 'sqlglot.expressions.DateAdd'>: <function date_delta_sql.<locals>._delta_sql>, <class 'sqlglot.expressions.DateDiff'>: <function date_delta_sql.<locals>._delta_sql>, <class 'sqlglot.expressions.CTE'>: <function preprocess.<locals>._to_sql>, <class 'sqlglot.expressions.CurrentDate'>: <function rename_func.<locals>.<lambda>>, <class 'sqlglot.expressions.CurrentTimestamp'>: <function rename_func.<locals>.<lambda>>, <class 'sqlglot.expressions.DateStrToDate'>: <function datestrtodate_sql>, <class 'sqlglot.expressions.Extract'>: <function rename_func.<locals>.<lambda>>, <class 'sqlglot.expressions.GeneratedAsIdentityColumnConstraint'>: <function generatedasidentitycolumnconstraint_sql>, <class 'sqlglot.expressions.GroupConcat'>: <function _string_agg_sql>, <class 'sqlglot.expressions.If'>: <function rename_func.<locals>.<lambda>>, <class 'sqlglot.expressions.JSONExtract'>: <function _json_extract_sql>, <class 'sqlglot.expressions.JSONExtractScalar'>: <function _json_extract_sql>, <class 'sqlglot.expressions.LastDay'>: <function TSQL.Generator.<lambda>>, <class 'sqlglot.expressions.Max'>: <function max_or_greatest>, <class 'sqlglot.expressions.MD5'>: <function TSQL.Generator.<lambda>>, <class 'sqlglot.expressions.Min'>: <function min_or_least>, <class 'sqlglot.expressions.NumberToStr'>: <function _format_sql>, <class 'sqlglot.expressions.Repeat'>: <function rename_func.<locals>.<lambda>>, <class 'sqlglot.expressions.Select'>: <function preprocess.<locals>._to_sql>, <class 'sqlglot.expressions.Stddev'>: <function rename_func.<locals>.<lambda>>, <class 'sqlglot.expressions.StrPosition'>: <function TSQL.Generator.<lambda>>, <class 'sqlglot.expressions.Subquery'>: <function preprocess.<locals>._to_sql>, <class 'sqlglot.expressions.SHA'>: <function TSQL.Generator.<lambda>>, <class 'sqlglot.expressions.SHA2'>: <function TSQL.Generator.<lambda>>, <class 'sqlglot.expressions.TimeStrToTime'>: <function TSQL.Generator.<lambda>>, <class 'sqlglot.expressions.TimeToStr'>: <function _format_sql>, <class 'sqlglot.expressions.Trim'>: <function trim_sql>, <class 'sqlglot.expressions.TsOrDsAdd'>: <function date_delta_sql.<locals>._delta_sql>, <class 'sqlglot.expressions.TsOrDsDiff'>: <function date_delta_sql.<locals>._delta_sql>}
PROPERTIES_LOCATION = {<class 'sqlglot.expressions.AllowedValuesProperty'>: <Location.POST_SCHEMA: 'POST_SCHEMA'>, <class 'sqlglot.expressions.AlgorithmProperty'>: <Location.POST_CREATE: 'POST_CREATE'>, <class 'sqlglot.expressions.AutoIncrementProperty'>: <Location.POST_SCHEMA: 'POST_SCHEMA'>, <class 'sqlglot.expressions.AutoRefreshProperty'>: <Location.POST_SCHEMA: 'POST_SCHEMA'>, <class 'sqlglot.expressions.BackupProperty'>: <Location.POST_SCHEMA: 'POST_SCHEMA'>, <class 'sqlglot.expressions.BlockCompressionProperty'>: <Location.POST_NAME: 'POST_NAME'>, <class 'sqlglot.expressions.CharacterSetProperty'>: <Location.POST_SCHEMA: 'POST_SCHEMA'>, <class 'sqlglot.expressions.ChecksumProperty'>: <Location.POST_NAME: 'POST_NAME'>, <class 'sqlglot.expressions.CollateProperty'>: <Location.POST_SCHEMA: 'POST_SCHEMA'>, <class 'sqlglot.expressions.CopyGrantsProperty'>: <Location.POST_SCHEMA: 'POST_SCHEMA'>, <class 'sqlglot.expressions.Cluster'>: <Location.POST_SCHEMA: 'POST_SCHEMA'>, <class 'sqlglot.expressions.ClusteredByProperty'>: <Location.POST_SCHEMA: 'POST_SCHEMA'>, <class 'sqlglot.expressions.DataBlocksizeProperty'>: <Location.POST_NAME: 'POST_NAME'>, <class 'sqlglot.expressions.DataDeletionProperty'>: <Location.POST_SCHEMA: 'POST_SCHEMA'>, <class 'sqlglot.expressions.DefinerProperty'>: <Location.POST_CREATE: 'POST_CREATE'>, <class 'sqlglot.expressions.DictRange'>: <Location.POST_SCHEMA: 'POST_SCHEMA'>, <class 'sqlglot.expressions.DictProperty'>: <Location.POST_SCHEMA: 'POST_SCHEMA'>, <class 'sqlglot.expressions.DynamicProperty'>: <Location.POST_CREATE: 'POST_CREATE'>, <class 'sqlglot.expressions.DistKeyProperty'>: <Location.POST_SCHEMA: 'POST_SCHEMA'>, <class 'sqlglot.expressions.DistStyleProperty'>: <Location.POST_SCHEMA: 'POST_SCHEMA'>, <class 'sqlglot.expressions.EngineProperty'>: <Location.POST_SCHEMA: 'POST_SCHEMA'>, <class 'sqlglot.expressions.ExecuteAsProperty'>: <Location.POST_SCHEMA: 'POST_SCHEMA'>, <class 'sqlglot.expressions.ExternalProperty'>: <Location.POST_CREATE: 'POST_CREATE'>, <class 'sqlglot.expressions.FallbackProperty'>: <Location.POST_NAME: 'POST_NAME'>, <class 'sqlglot.expressions.FileFormatProperty'>: <Location.POST_WITH: 'POST_WITH'>, <class 'sqlglot.expressions.FreespaceProperty'>: <Location.POST_NAME: 'POST_NAME'>, <class 'sqlglot.expressions.GlobalProperty'>: <Location.POST_CREATE: 'POST_CREATE'>, <class 'sqlglot.expressions.HeapProperty'>: <Location.POST_WITH: 'POST_WITH'>, <class 'sqlglot.expressions.InheritsProperty'>: <Location.POST_SCHEMA: 'POST_SCHEMA'>, <class 'sqlglot.expressions.IcebergProperty'>: <Location.POST_CREATE: 'POST_CREATE'>, <class 'sqlglot.expressions.InputModelProperty'>: <Location.POST_SCHEMA: 'POST_SCHEMA'>, <class 'sqlglot.expressions.IsolatedLoadingProperty'>: <Location.POST_NAME: 'POST_NAME'>, <class 'sqlglot.expressions.JournalProperty'>: <Location.POST_NAME: 'POST_NAME'>, <class 'sqlglot.expressions.LanguageProperty'>: <Location.POST_SCHEMA: 'POST_SCHEMA'>, <class 'sqlglot.expressions.LikeProperty'>: <Location.POST_SCHEMA: 'POST_SCHEMA'>, <class 'sqlglot.expressions.LocationProperty'>: <Location.POST_SCHEMA: 'POST_SCHEMA'>, <class 'sqlglot.expressions.LockProperty'>: <Location.POST_SCHEMA: 'POST_SCHEMA'>, <class 'sqlglot.expressions.LockingProperty'>: <Location.POST_ALIAS: 'POST_ALIAS'>, <class 'sqlglot.expressions.LogProperty'>: <Location.POST_NAME: 'POST_NAME'>, <class 'sqlglot.expressions.MaterializedProperty'>: <Location.POST_CREATE: 'POST_CREATE'>, <class 'sqlglot.expressions.MergeBlockRatioProperty'>: <Location.POST_NAME: 'POST_NAME'>, <class 'sqlglot.expressions.NoPrimaryIndexProperty'>: <Location.POST_EXPRESSION: 'POST_EXPRESSION'>, <class 'sqlglot.expressions.OnProperty'>: <Location.POST_SCHEMA: 'POST_SCHEMA'>, <class 'sqlglot.expressions.OnCommitProperty'>: <Location.POST_EXPRESSION: 'POST_EXPRESSION'>, <class 'sqlglot.expressions.Order'>: <Location.POST_SCHEMA: 'POST_SCHEMA'>, <class 'sqlglot.expressions.OutputModelProperty'>: <Location.POST_SCHEMA: 'POST_SCHEMA'>, <class 'sqlglot.expressions.PartitionedByProperty'>: <Location.POST_WITH: 'POST_WITH'>, <class 'sqlglot.expressions.PartitionedOfProperty'>: <Location.POST_SCHEMA: 'POST_SCHEMA'>, <class 'sqlglot.expressions.PrimaryKey'>: <Location.POST_SCHEMA: 'POST_SCHEMA'>, <class 'sqlglot.expressions.Property'>: <Location.POST_WITH: 'POST_WITH'>, <class 'sqlglot.expressions.RemoteWithConnectionModelProperty'>: <Location.POST_SCHEMA: 'POST_SCHEMA'>, <class 'sqlglot.expressions.ReturnsProperty'>: <Location.POST_SCHEMA: 'POST_SCHEMA'>, <class 'sqlglot.expressions.RowFormatProperty'>: <Location.POST_SCHEMA: 'POST_SCHEMA'>, <class 'sqlglot.expressions.RowFormatDelimitedProperty'>: <Location.POST_SCHEMA: 'POST_SCHEMA'>, <class 'sqlglot.expressions.RowFormatSerdeProperty'>: <Location.POST_SCHEMA: 'POST_SCHEMA'>, <class 'sqlglot.expressions.SampleProperty'>: <Location.POST_SCHEMA: 'POST_SCHEMA'>, <class 'sqlglot.expressions.SchemaCommentProperty'>: <Location.POST_SCHEMA: 'POST_SCHEMA'>, <class 'sqlglot.expressions.SecureProperty'>: <Location.POST_CREATE: 'POST_CREATE'>, <class 'sqlglot.expressions.SerdeProperties'>: <Location.POST_SCHEMA: 'POST_SCHEMA'>, <class 'sqlglot.expressions.Set'>: <Location.POST_SCHEMA: 'POST_SCHEMA'>, <class 'sqlglot.expressions.SettingsProperty'>: <Location.POST_SCHEMA: 'POST_SCHEMA'>, <class 'sqlglot.expressions.SetProperty'>: <Location.POST_CREATE: 'POST_CREATE'>, <class 'sqlglot.expressions.SetConfigProperty'>: <Location.POST_SCHEMA: 'POST_SCHEMA'>, <class 'sqlglot.expressions.SharingProperty'>: <Location.POST_EXPRESSION: 'POST_EXPRESSION'>, <class 'sqlglot.expressions.SequenceProperties'>: <Location.POST_EXPRESSION: 'POST_EXPRESSION'>, <class 'sqlglot.expressions.SortKeyProperty'>: <Location.POST_SCHEMA: 'POST_SCHEMA'>, <class 'sqlglot.expressions.SqlReadWriteProperty'>: <Location.POST_SCHEMA: 'POST_SCHEMA'>, <class 'sqlglot.expressions.SqlSecurityProperty'>: <Location.POST_CREATE: 'POST_CREATE'>, <class 'sqlglot.expressions.StabilityProperty'>: <Location.POST_SCHEMA: 'POST_SCHEMA'>, <class 'sqlglot.expressions.StrictProperty'>: <Location.POST_SCHEMA: 'POST_SCHEMA'>, <class 'sqlglot.expressions.TemporaryProperty'>: <Location.POST_CREATE: 'POST_CREATE'>, <class 'sqlglot.expressions.ToTableProperty'>: <Location.POST_SCHEMA: 'POST_SCHEMA'>, <class 'sqlglot.expressions.TransientProperty'>: <Location.POST_CREATE: 'POST_CREATE'>, <class 'sqlglot.expressions.TransformModelProperty'>: <Location.POST_SCHEMA: 'POST_SCHEMA'>, <class 'sqlglot.expressions.MergeTreeTTL'>: <Location.POST_SCHEMA: 'POST_SCHEMA'>, <class 'sqlglot.expressions.UnloggedProperty'>: <Location.POST_CREATE: 'POST_CREATE'>, <class 'sqlglot.expressions.ViewAttributeProperty'>: <Location.POST_SCHEMA: 'POST_SCHEMA'>, <class 'sqlglot.expressions.VolatileProperty'>: <Location.UNSUPPORTED: 'UNSUPPORTED'>, <class 'sqlglot.expressions.WithDataProperty'>: <Location.POST_EXPRESSION: 'POST_EXPRESSION'>, <class 'sqlglot.expressions.WithJournalTableProperty'>: <Location.POST_NAME: 'POST_NAME'>, <class 'sqlglot.expressions.WithSchemaBindingProperty'>: <Location.POST_SCHEMA: 'POST_SCHEMA'>, <class 'sqlglot.expressions.WithSystemVersioningProperty'>: <Location.POST_SCHEMA: 'POST_SCHEMA'>}
def scope_resolution(self, rhs: str, scope_name: str) -> str: View Source

    895        def scope_resolution(self, rhs: str, scope_name: str) -> str:
    896            return f"{scope_name}::{rhs}"

def select_sql(self, expression: sqlglot.expressions.Select) -> str: View Source

    898        def select_sql(self, expression: exp.Select) -> str:
    899            if expression.args.get("offset"):
    900                if not expression.args.get("order"):
    901                    # ORDER BY is required in order to use OFFSET in a query, so we use
    902                    # a noop order by, since we don't really care about the order.
    903                    # See: https://www.microsoftpressstore.com/articles/article.aspx?p=2314819
    904                    expression.order_by(exp.select(exp.null()).subquery(), copy=False)
    905
    906                limit = expression.args.get("limit")
    907                if isinstance(limit, exp.Limit):
    908                    # TOP and OFFSET can't be combined, we need use FETCH instead of TOP
    909                    # we replace here because otherwise TOP would be generated in select_sql
    910                    limit.replace(exp.Fetch(direction="FIRST", count=limit.expression))
    911
    912            return super().select_sql(expression)

def convert_sql(self, expression: sqlglot.expressions.Convert) -> str: View Source

    914        def convert_sql(self, expression: exp.Convert) -> str:
    915            name = "TRY_CONVERT" if expression.args.get("safe") else "CONVERT"
    916            return self.func(
    917                name, expression.this, expression.expression, expression.args.get("style")
    918            )

def queryoption_sql(self, expression: sqlglot.expressions.QueryOption) -> str: View Source

    920        def queryoption_sql(self, expression: exp.QueryOption) -> str:
    921            option = self.sql(expression, "this")
    922            value = self.sql(expression, "expression")
    923            if value:
    924                optional_equal_sign = "= " if option in OPTIONS_THAT_REQUIRE_EQUAL else ""
    925                return f"{option} {optional_equal_sign}{value}"
    926            return option

def lateral_op(self, expression: sqlglot.expressions.Lateral) -> str: View Source

    928        def lateral_op(self, expression: exp.Lateral) -> str:
    929            cross_apply = expression.args.get("cross_apply")
    930            if cross_apply is True:
    931                return "CROSS APPLY"
    932            if cross_apply is False:
    933                return "OUTER APPLY"
    934
    935            # TODO: perhaps we can check if the parent is a Join and transpile it appropriately
    936            self.unsupported("LATERAL clause is not supported.")
    937            return "LATERAL"

def timefromparts_sql(self, expression: sqlglot.expressions.TimeFromParts) -> str: View Source

    939        def timefromparts_sql(self, expression: exp.TimeFromParts) -> str:
    940            nano = expression.args.get("nano")
    941            if nano is not None:
    942                nano.pop()
    943                self.unsupported("Specifying nanoseconds is not supported in TIMEFROMPARTS.")
    944
    945            if expression.args.get("fractions") is None:
    946                expression.set("fractions", exp.Literal.number(0))
    947            if expression.args.get("precision") is None:
    948                expression.set("precision", exp.Literal.number(0))
    949
    950            return rename_func("TIMEFROMPARTS")(self, expression)

def timestampfromparts_sql(self, expression: sqlglot.expressions.TimestampFromParts) -> str: View Source

    952        def timestampfromparts_sql(self, expression: exp.TimestampFromParts) -> str:
    953            zone = expression.args.get("zone")
    954            if zone is not None:
    955                zone.pop()
    956                self.unsupported("Time zone is not supported in DATETIMEFROMPARTS.")
    957
    958            nano = expression.args.get("nano")
    959            if nano is not None:
    960                nano.pop()
    961                self.unsupported("Specifying nanoseconds is not supported in DATETIMEFROMPARTS.")
    962
    963            if expression.args.get("milli") is None:
    964                expression.set("milli", exp.Literal.number(0))
    965
    966            return rename_func("DATETIMEFROMPARTS")(self, expression)

def setitem_sql(self, expression: sqlglot.expressions.SetItem) -> str: View Source

    968        def setitem_sql(self, expression: exp.SetItem) -> str:
    969            this = expression.this
    970            if isinstance(this, exp.EQ) and not isinstance(this.left, exp.Parameter):
    971                # T-SQL does not use '=' in SET command, except when the LHS is a variable.
    972                return f"{self.sql(this.left)} {self.sql(this.right)}"
    973
    974            return super().setitem_sql(expression)

def boolean_sql(self, expression: sqlglot.expressions.Boolean) -> str: View Source

    976        def boolean_sql(self, expression: exp.Boolean) -> str:
    977            if type(expression.parent) in BIT_TYPES:
    978                return "1" if expression.this else "0"
    979
    980            return "(1 = 1)" if expression.this else "(1 = 0)"

def is_sql(self, expression: sqlglot.expressions.Is) -> str: View Source

    982        def is_sql(self, expression: exp.Is) -> str:
    983            if isinstance(expression.expression, exp.Boolean):
    984                return self.binary(expression, "=")
    985            return self.binary(expression, "IS")

def createable_sql( self, expression: sqlglot.expressions.Create, locations: DefaultDict) -> str: View Source

    987        def createable_sql(self, expression: exp.Create, locations: t.DefaultDict) -> str:
    988            sql = self.sql(expression, "this")
    989            properties = expression.args.get("properties")
    990
    991            if sql[:1] != "#" and any(
    992                isinstance(prop, exp.TemporaryProperty)
    993                for prop in (properties.expressions if properties else [])
    994            ):
    995                sql = f"[#{sql[1:]}" if sql.startswith("[") else f"#{sql}"
    996
    997            return sql

def create_sql(self, expression: sqlglot.expressions.Create) -> str: View Source

     999        def create_sql(self, expression: exp.Create) -> str:
    1000            kind = expression.kind
    1001            exists = expression.args.pop("exists", None)
    1002
    1003            like_property = expression.find(exp.LikeProperty)
    1004            if like_property:
    1005                ctas_expression = like_property.this
    1006            else:
    1007                ctas_expression = expression.expression
    1008
    1009            if kind == "VIEW":
    1010                expression.this.set("catalog", None)
    1011                with_ = expression.args.get("with")
    1012                if ctas_expression and with_:
    1013                    # We've already preprocessed the Create expression to bubble up any nested CTEs,
    1014                    # but CREATE VIEW actually requires the WITH clause to come after it so we need
    1015                    # to amend the AST by moving the CTEs to the CREATE VIEW statement's query.
    1016                    ctas_expression.set("with", with_.pop())
    1017
    1018            sql = super().create_sql(expression)
    1019
    1020            table = expression.find(exp.Table)
    1021
    1022            # Convert CTAS statement to SELECT .. INTO ..
    1023            if kind == "TABLE" and ctas_expression:
    1024                if isinstance(ctas_expression, exp.UNWRAPPED_QUERIES):
    1025                    ctas_expression = ctas_expression.subquery()
    1026
    1027                select_into = exp.select("*").from_(exp.alias_(ctas_expression, "temp", table=True))
    1028                select_into.set("into", exp.Into(this=table))
    1029
    1030                if like_property:
    1031                    select_into.limit(0, copy=False)
    1032
    1033                sql = self.sql(select_into)
    1034
    1035            if exists:
    1036                identifier = self.sql(exp.Literal.string(exp.table_name(table) if table else ""))
    1037                sql = self.sql(exp.Literal.string(sql))
    1038                if kind == "SCHEMA":
    1039                    sql = f"""IF NOT EXISTS (SELECT * FROM information_schema.schemata WHERE schema_name = {identifier}) EXEC({sql})"""
    1040                elif kind == "TABLE":
    1041                    assert table
    1042                    where = exp.and_(
    1043                        exp.column("table_name").eq(table.name),
    1044                        exp.column("table_schema").eq(table.db) if table.db else None,
    1045                        exp.column("table_catalog").eq(table.catalog) if table.catalog else None,
    1046                    )
    1047                    sql = f"""IF NOT EXISTS (SELECT * FROM information_schema.tables WHERE {where}) EXEC({sql})"""
    1048                elif kind == "INDEX":
    1049                    index = self.sql(exp.Literal.string(expression.this.text("this")))
    1050                    sql = f"""IF NOT EXISTS (SELECT * FROM sys.indexes WHERE object_id = object_id({identifier}) AND name = {index}) EXEC({sql})"""
    1051            elif expression.args.get("replace"):
    1052                sql = sql.replace("CREATE OR REPLACE ", "CREATE OR ALTER ", 1)
    1053
    1054            return self.prepend_ctes(expression, sql)

def offset_sql(self, expression: sqlglot.expressions.Offset) -> str: View Source

    1056        def offset_sql(self, expression: exp.Offset) -> str:
    1057            return f"{super().offset_sql(expression)} ROWS"

def version_sql(self, expression: sqlglot.expressions.Version) -> str: View Source

    1059        def version_sql(self, expression: exp.Version) -> str:
    1060            name = "SYSTEM_TIME" if expression.name == "TIMESTAMP" else expression.name
    1061            this = f"FOR {name}"
    1062            expr = expression.expression
    1063            kind = expression.text("kind")
    1064            if kind in ("FROM", "BETWEEN"):
    1065                args = expr.expressions
    1066                sep = "TO" if kind == "FROM" else "AND"
    1067                expr_sql = f"{self.sql(seq_get(args, 0))} {sep} {self.sql(seq_get(args, 1))}"
    1068            else:
    1069                expr_sql = self.sql(expr)
    1070
    1071            expr_sql = f" {expr_sql}" if expr_sql else ""
    1072            return f"{this} {kind}{expr_sql}"

def returnsproperty_sql(self, expression: sqlglot.expressions.ReturnsProperty) -> str: View Source

    1074        def returnsproperty_sql(self, expression: exp.ReturnsProperty) -> str:
    1075            table = expression.args.get("table")
    1076            table = f"{table} " if table else ""
    1077            return f"RETURNS {table}{self.sql(expression, 'this')}"

def returning_sql(self, expression: sqlglot.expressions.Returning) -> str: View Source

    1079        def returning_sql(self, expression: exp.Returning) -> str:
    1080            into = self.sql(expression, "into")
    1081            into = self.seg(f"INTO {into}") if into else ""
    1082            return f"{self.seg('OUTPUT')} {self.expressions(expression, flat=True)}{into}"

def transaction_sql(self, expression: sqlglot.expressions.Transaction) -> str: View Source

    1084        def transaction_sql(self, expression: exp.Transaction) -> str:
    1085            this = self.sql(expression, "this")
    1086            this = f" {this}" if this else ""
    1087            mark = self.sql(expression, "mark")
    1088            mark = f" WITH MARK {mark}" if mark else ""
    1089            return f"BEGIN TRANSACTION{this}{mark}"

def commit_sql(self, expression: sqlglot.expressions.Commit) -> str: View Source

    1091        def commit_sql(self, expression: exp.Commit) -> str:
    1092            this = self.sql(expression, "this")
    1093            this = f" {this}" if this else ""
    1094            durability = expression.args.get("durability")
    1095            durability = (
    1096                f" WITH (DELAYED_DURABILITY = {'ON' if durability else 'OFF'})"
    1097                if durability is not None
    1098                else ""
    1099            )
    1100            return f"COMMIT TRANSACTION{this}{durability}"

def rollback_sql(self, expression: sqlglot.expressions.Rollback) -> str: View Source

    1102        def rollback_sql(self, expression: exp.Rollback) -> str:
    1103            this = self.sql(expression, "this")
    1104            this = f" {this}" if this else ""
    1105            return f"ROLLBACK TRANSACTION{this}"

def identifier_sql(self, expression: sqlglot.expressions.Identifier) -> str: View Source

    1107        def identifier_sql(self, expression: exp.Identifier) -> str:
    1108            identifier = super().identifier_sql(expression)
    1109
    1110            if expression.args.get("global"):
    1111                identifier = f"##{identifier}"
    1112            elif expression.args.get("temporary"):
    1113                identifier = f"#{identifier}"
    1114
    1115            return identifier

def constraint_sql(self, expression: sqlglot.expressions.Constraint) -> str: View Source

    1117        def constraint_sql(self, expression: exp.Constraint) -> str:
    1118            this = self.sql(expression, "this")
    1119            expressions = self.expressions(expression, flat=True, sep=" ")
    1120            return f"CONSTRAINT {this} {expressions}"

def length_sql(self, expression: sqlglot.expressions.Length) -> str: View Source

    1122        def length_sql(self, expression: exp.Length) -> str:
    1123            return self._uncast_text(expression, "LEN")

def right_sql(self, expression: sqlglot.expressions.Right) -> str: View Source

    1125        def right_sql(self, expression: exp.Right) -> str:
    1126            return self._uncast_text(expression, "RIGHT")

def left_sql(self, expression: sqlglot.expressions.Left) -> str: View Source

    1128        def left_sql(self, expression: exp.Left) -> str:
    1129            return self._uncast_text(expression, "LEFT")

def partition_sql(self, expression: sqlglot.expressions.Partition) -> str: View Source

    1140        def partition_sql(self, expression: exp.Partition) -> str:
    1141            return f"WITH (PARTITIONS({self.expressions(expression, flat=True)}))"

def altertable_sql(self, expression: sqlglot.expressions.AlterTable) -> str: View Source

    1143        def altertable_sql(self, expression: exp.AlterTable) -> str:
    1144            action = seq_get(expression.args.get("actions") or [], 0)
    1145            if isinstance(action, exp.RenameTable):
    1146                return f"EXEC sp_rename '{self.sql(expression.this)}', '{action.this.name}'"
    1147            return super().altertable_sql(expression)

def drop_sql(self, expression: sqlglot.expressions.Drop) -> str: View Source

    1149        def drop_sql(self, expression: exp.Drop) -> str:
    1150            if expression.args["kind"] == "VIEW":
    1151                expression.this.set("catalog", None)
    1152            return super().drop_sql(expression)

def declare_sql(self, expression: sqlglot.expressions.Declare) -> str: View Source

    1154        def declare_sql(self, expression: exp.Declare) -> str:
    1155            return f"DECLARE {self.expressions(expression, flat=True)}"

def declareitem_sql(self, expression: sqlglot.expressions.DeclareItem) -> str: View Source

    1157        def declareitem_sql(self, expression: exp.DeclareItem) -> str:
    1158            variable = self.sql(expression, "this")
    1159            default = self.sql(expression, "default")
    1160            default = f" = {default}" if default else ""
    1161
    1162            kind = self.sql(expression, "kind")
    1163            if isinstance(expression.args.get("kind"), exp.Schema):
    1164                kind = f"TABLE {kind}"
    1165
    1166            return f"{variable} AS {kind}{default}"

def options_modifier(self, expression: sqlglot.expressions.Expression) -> str: View Source

    1168        def options_modifier(self, expression: exp.Expression) -> str:
    1169            options = self.expressions(expression, key="options")
    1170            return f" OPTION{self.wrap(options)}" if options else ""

SELECT_KINDS: Tuple[str, ...] = ()
TRY_SUPPORTED = False
SUPPORTS_UESCAPE = False
AFTER_HAVING_MODIFIER_TRANSFORMS = {'windows': <function Generator.<lambda>>, 'qualify': <function Generator.<lambda>>}

##### Inherited Members

sqlglot.generator.Generator
  Generator
  IGNORE_NULLS_IN_FUNC
  LOCKING_READS_SUPPORTED
  EXPLICIT_SET_OP
  WRAP_DERIVED_VALUES
  CREATE_FUNCTION_RETURN_AS
  MATCHED_BY_SOURCE
  SINGLE_STRING_INTERVAL
  INTERVAL_ALLOWS_PLURAL_FORM
  LIMIT_ONLY_LITERALS
  RENAME_TABLE_WITH_DB
  GROUPINGS_SEP
  INDEX_ON
  JOIN_HINTS
  TABLE_HINTS
  QUERY_HINT_SEP
  IS_BOOL_ALLOWED
  DUPLICATE_KEY_UPDATE_WITH_SET
  EXTRACT_ALLOWS_QUOTES
  TZ_TO_WITH_TIME_ZONE
  VALUES_AS_TABLE
  UNNEST_WITH_ORDINALITY
  AGGREGATE_FILTER_SUPPORTED
  SEMI_ANTI_JOIN_WITH_SIDE
  SUPPORTS_TABLE_COPY
  TABLESAMPLE_REQUIRES_PARENS
  TABLESAMPLE_SIZE_IS_ROWS
  TABLESAMPLE_KEYWORDS
  TABLESAMPLE_WITH_METHOD
  COLLATE_IS_FUNC
  DATA_TYPE_SPECIFIERS_ALLOWED
  LAST_DAY_SUPPORTS_DATE_PART
  SUPPORTS_TABLE_ALIAS_COLUMNS
  UNPIVOT_ALIASES_ARE_IDENTIFIERS
  JSON_KEY_VALUE_PAIR_SEP
  INSERT_OVERWRITE
  SUPPORTS_UNLOGGED_TABLES
  SUPPORTS_CREATE_TABLE_LIKE
  LIKE_PROPERTY_INSIDE_SCHEMA
  MULTI_ARG_DISTINCT
  JSON_TYPE_REQUIRED_FOR_EXTRACTION
  JSON_PATH_SINGLE_QUOTE_ESCAPE
  CAN_IMPLEMENT_ARRAY_ANY
  COPY_PARAMS_ARE_WRAPPED
  COPY_HAS_INTO_KEYWORD
  STAR_EXCEPT
  HEX_FUNC
  WITH_PROPERTIES_PREFIX
  QUOTE_JSON_PATH
  PAD_FILL_PATTERN_IS_REQUIRED
  TIME_PART_SINGULARS
  TOKEN_MAPPING
  STRUCT_DELIMITER
  PARAMETER_TOKEN
  NAMED_PLACEHOLDER_TOKEN
  RESERVED_KEYWORDS
  WITH_SEPARATED_COMMENTS
  EXCLUDE_COMMENTS
  UNWRAPPED_INTERVAL_VALUES
  PARAMETERIZABLE_TEXT_TYPES
  SENTINEL_LINE_BREAK
  pretty
  identify
  normalize
  pad
  unsupported_level
  max_unsupported
  leading_comma
  max_text_width
  comments
  dialect
  normalize_functions
  unsupported_messages
  generate
  preprocess
  unsupported
  sep
  seg
  pad_comment
  maybe_comment
  wrap
  no_identify
  normalize_func
  indent
  sql
  uncache_sql
  cache_sql
  characterset_sql
  column_parts
  column_sql
  columnposition_sql
  columndef_sql
  columnconstraint_sql
  computedcolumnconstraint_sql
  autoincrementcolumnconstraint_sql
  compresscolumnconstraint_sql
  generatedasidentitycolumnconstraint_sql
  generatedasrowcolumnconstraint_sql
  periodforsystemtimeconstraint_sql
  notnullcolumnconstraint_sql
  transformcolumnconstraint_sql
  primarykeycolumnconstraint_sql
  uniquecolumnconstraint_sql
  sequenceproperties_sql
  clone_sql
  describe_sql
  heredoc_sql
  prepend_ctes
  with_sql
  cte_sql
  tablealias_sql
  bitstring_sql
  hexstring_sql
  bytestring_sql
  unicodestring_sql
  rawstring_sql
  datatypeparam_sql
  datatype_sql
  directory_sql
  delete_sql
  except_sql
  except_op
  fetch_sql
  filter_sql
  hint_sql
  indexparameters_sql
  index_sql
  hex_sql
  lowerhex_sql
  inputoutputformat_sql
  national_sql
  properties_sql
  root_properties
  properties
  with_properties
  locate_properties
  property_name
  property_sql
  likeproperty_sql
  fallbackproperty_sql
  journalproperty_sql
  freespaceproperty_sql
  checksumproperty_sql
  mergeblockratioproperty_sql
  datablocksizeproperty_sql
  blockcompressionproperty_sql
  isolatedloadingproperty_sql
  partitionboundspec_sql
  partitionedofproperty_sql
  lockingproperty_sql
  withdataproperty_sql
  withsystemversioningproperty_sql
  insert_sql
  intersect_sql
  intersect_op
  introducer_sql
  kill_sql
  pseudotype_sql
  objectidentifier_sql
  onconflict_sql
  rowformatdelimitedproperty_sql
  withtablehint_sql
  indextablehint_sql
  historicaldata_sql
  table_parts
  table_sql
  tablesample_sql
  pivot_sql
  tuple_sql
  update_sql
  values_sql
  var_sql
  into_sql
  from_sql
  group_sql
  having_sql
  connect_sql
  prior_sql
  join_sql
  lambda_sql
  lateral_sql
  limit_sql
  set_sql
  pragma_sql
  lock_sql
  literal_sql
  escape_str
  loaddata_sql
  null_sql
  order_sql
  withfill_sql
  cluster_sql
  distribute_sql
  sort_sql
  ordered_sql
  matchrecognizemeasure_sql
  matchrecognize_sql
  query_modifiers
  offset_limit_modifiers
  after_limit_modifiers
  schema_sql
  schema_columns_sql
  star_sql
  parameter_sql
  sessionparameter_sql
  placeholder_sql
  subquery_sql
  qualify_sql
  set_operations
  union_sql
  union_op
  unnest_sql
  prewhere_sql
  where_sql
  window_sql
  partition_by_sql
  windowspec_sql
  withingroup_sql
  between_sql
  bracket_offset_expressions
  bracket_sql
  all_sql
  any_sql
  exists_sql
  case_sql
  nextvaluefor_sql
  extract_sql
  trim_sql
  convert_concat_args
  concat_sql
  concatws_sql
  check_sql
  foreignkey_sql
  primarykey_sql
  if_sql
  matchagainst_sql
  jsonkeyvalue_sql
  jsonpath_sql
  json_path_part
  formatjson_sql
  jsonobject_sql
  jsonobjectagg_sql
  jsonarray_sql
  jsonarrayagg_sql
  jsoncolumndef_sql
  jsonschema_sql
  jsontable_sql
  openjsoncolumndef_sql
  openjson_sql
  in_sql
  in_unnest_op
  interval_sql
  return_sql
  reference_sql
  anonymous_sql
  paren_sql
  neg_sql
  not_sql
  alias_sql
  pivotalias_sql
  aliases_sql
  atindex_sql
  attimezone_sql
  fromtimezone_sql
  add_sql
  and_sql
  or_sql
  xor_sql
  connector_sql
  bitwiseand_sql
  bitwiseleftshift_sql
  bitwisenot_sql
  bitwiseor_sql
  bitwiserightshift_sql
  bitwisexor_sql
  cast_sql
  currentdate_sql
  collate_sql
  command_sql
  comment_sql
  mergetreettlaction_sql
  mergetreettl_sql
  altercolumn_sql
  alterdiststyle_sql
  altersortkey_sql
  renametable_sql
  renamecolumn_sql
  alterset_sql
  add_column_sql
  droppartition_sql
  addconstraint_sql
  distinct_sql
  ignorenulls_sql
  respectnulls_sql
  havingmax_sql
  intdiv_sql
  dpipe_sql
  div_sql
  overlaps_sql
  distance_sql
  dot_sql
  eq_sql
  propertyeq_sql
  escape_sql
  glob_sql
  gt_sql
  gte_sql
  ilike_sql
  ilikeany_sql
  like_sql
  likeany_sql
  similarto_sql
  lt_sql
  lte_sql
  mod_sql
  mul_sql
  neq_sql
  nullsafeeq_sql
  nullsafeneq_sql
  slice_sql
  sub_sql
  trycast_sql
  try_sql
  log_sql
  use_sql
  binary
  function_fallback_sql
  func
  format_args
  too_wide
  format_time
  expressions
  op_expressions
  naked_property
  tag_sql
  token_sql
  userdefinedfunction_sql
  joinhint_sql
  kwarg_sql
  when_sql
  merge_sql
  tochar_sql
  tonumber_sql
  dictproperty_sql
  dictrange_sql
  dictsubproperty_sql
  oncluster_sql
  clusteredbyproperty_sql
  anyvalue_sql
  querytransform_sql
  indexconstraintoption_sql
  checkcolumnconstraint_sql
  indexcolumnconstraint_sql
  nvl2_sql
  comprehension_sql
  columnprefix_sql
  opclass_sql
  predict_sql
  forin_sql
  refresh_sql
  operator_sql
  toarray_sql
  tsordstotime_sql
  tsordstotimestamp_sql
  tsordstodate_sql
  unixdate_sql
  lastday_sql
  dateadd_sql
  arrayany_sql
  struct_sql
  partitionrange_sql
  truncatetable_sql
  copyparameter_sql
  credentials_sql
  copy_sql
  semicolon_sql
  datadeletionproperty_sql
  maskingpolicycolumnconstraint_sql
  gapfill_sql
  scoperesolution_sql
  parsejson_sql
  rand_sql
  changes_sql
  pad_sql
  summarize_sql